use std::{
    collections::VecDeque,
    time::{Duration, Instant},
};

use asciivation_duty_renderer::{
    part::{Part, PartRunner},
    render::Color,
};
use rand::{thread_rng, Rng};
use uianimator::{default_animator_f64_quadratic::DefaultAnimatorF64Quadratic, Animator};

use crate::state::GlobalState;

/// a cross-screen consistent background, stored in `GlobalState`.
/// to use this, call the `draw` function (found in the `background` module).
pub struct Background {
    last_star_added: Instant,
    seconds_per_star_per_pixel: f64,
    seconds_per_down_movement_min: f64,
    seconds_per_down_movement_max: f64,
    transparency_change_speed: f64,
    star_adding_delay: Duration,
    size: (usize, usize),
    stars: VecDeque<Star>,
    transparency: DefaultAnimatorF64Quadratic,
}

impl Background {
    pub fn new(
        seconds_per_star_per_pixel: f64,
        seconds_per_down_movement_min: f64,
        seconds_per_down_movement_max: f64,
        transparency_change_speed: f64,
        size: (usize, usize),
    ) -> Self {
        Self {
            last_star_added: Instant::now(),
            seconds_per_star_per_pixel,
            seconds_per_down_movement_min,
            seconds_per_down_movement_max,
            transparency_change_speed,
            star_adding_delay: Duration::from_secs_f64(seconds_per_star_per_pixel / size.0 as f64),
            size,
            stars: VecDeque::new(),
            transparency: DefaultAnimatorF64Quadratic::new(1.0, transparency_change_speed),
        }
    }
    /// Set the target transparency for the stars in the background.
    /// `transparency` should be within `0.0..=1.0`,
    /// and the actual transparency will smoothly fade between
    /// requested transparency values, meaning the
    /// visual change will not be applied immediately.
    pub fn set_transparency(&mut self, transparency: f64) {
        self.transparency
            .set_target(transparency.clamp(0.0, 1.0), Instant::now());
    }
    pub fn set_transparency_now(&mut self, transparency: f64) {
        self.transparency =
            DefaultAnimatorF64Quadratic::new(transparency, self.transparency_change_speed);
    }
}

/// the function used to draw a cross-screen consistent background, to be called from all parts' `draw` handlers.
///
/// this only works if the background has no visible chars set on its pixels, i.e.
/// the background may only be drawn using the `bg` color of pixels, since
/// stars will use `fg` and `char` to render themselves.
///
/// should be called in `draw`, **after** the code that handles
/// clearing the screen (if any) if `runner.draw_reset` is true,
/// but **before** any code that draws things on the screen,
/// to avoid stars overwriting ui elements.
pub fn draw<P: Part<State = GlobalState>>(runner: &mut PartRunner<P>, now: Instant) {
    if !runner.screen_buf.is_empty() {
        let height = runner.screen_buf.len();
        let width = runner.screen_buf[0].len();

        // check if the screen was resized
        if (width, height) != runner.state.background.size {
            // add random stars if the width/height is increased
            if width > runner.state.background.size.0 || height > runner.state.background.size.1 {
                let stars_per_pixel = runner.state.background.stars.len() as f64
                    / (runner.state.background.size.0 * runner.state.background.size.1) as f64;
                let mut rand = thread_rng();
                // fill the space on the right (excluding the space at the bottom right)
                for x in runner.state.background.size.0..width {
                    for y in 0..runner.state.background.size.1 {
                        if rand.gen_bool(stars_per_pixel) {
                            runner.state.background.add_star_at_randtime(
                                (width, height),
                                (x, y),
                                now,
                            );
                        }
                    }
                }
                // fill the space at the bottom (including the space at the bottom right)
                for x in 0..width {
                    for y in runner.state.background.size.1..height {
                        if rand.gen_bool(stars_per_pixel) {
                            runner.state.background.add_star_at_randtime(
                                (width, height),
                                (x, y),
                                now,
                            );
                        }
                    }
                }
            }
            runner.state.background.size = (width, height);
            runner.state.background.star_adding_delay = Duration::from_secs_f64(
                runner.state.background.seconds_per_star_per_pixel / width as f64,
            );
        }

        // periodically add another star
        while now - runner.state.background.last_star_added
            > runner.state.background.star_adding_delay
        {
            runner.state.background.last_star_added += runner.state.background.star_adding_delay;
            runner.state.background.add_star((width, height), now);
        }

        let mut star_removal_indices = vec![];
        // draw stars
        let transparency =
            (runner.state.background.transparency.get_value(now) * 255.0).clamp(0.0, 255.0) as u8;
        // move stars and clear pixels if a star moves away from a pixel
        for star in runner.state.background.stars.iter_mut() {
            if let Some((prev_x, prev_y)) = star.moved_down(now) {
                // star moved, clear its previous pixel
                if let Some(p) = runner
                    .screen_buf
                    .get_mut(prev_y)
                    .and_then(|row| row.get_mut(prev_x))
                {
                    p.char = ' ';
                }
            }
        }
        for (star_index, star) in runner.state.background.stars.iter_mut().enumerate() {
            // draw the star
            if let Some(p) = runner
                .screen_buf
                .get_mut(star.pos.1)
                .and_then(|row| row.get_mut(star.pos.0))
            {
                // set the pixel color to the star's color merged with the background's color.
                // where transparency is how transparent the star should appear.
                let mut fg_merge = p.bg;
                fg_merge.a = transparency;
                p.fg = star.fg.merge(fg_merge);
                p.char = star.char;
            } else {
                // out of screen
                star_removal_indices.push(star_index);
            }
        }
        // remove stars that are not on screen
        for star_index in star_removal_indices.into_iter().rev() {
            runner.state.background.stars.remove(star_index);
        }
    }
}

impl Background {
    /// add a star at the top of the screen, at a random x-coordinate
    fn add_star(&mut self, buf_size: (usize, usize), now: Instant) {
        self.add_star_at_norandtime(buf_size, (thread_rng().gen_range(0..buf_size.0), 0), now);
    }
    /// add a star at a specified position, with its move-down timer elapsed randomly (the star might move down directly after being added)
    fn add_star_at_randtime(
        &mut self,
        buf_size: (usize, usize),
        pos: (usize, usize),
        now: Instant,
    ) {
        self.add_star_at_movedownspeed(buf_size, pos, now, thread_rng().gen_range(0.0..=1.0))
    }
    /// add a star at a specified position, with its move-down timer set to zero. this is used when adding new stars at the top of the screen
    fn add_star_at_norandtime(
        &mut self,
        buf_size: (usize, usize),
        pos: (usize, usize),
        now: Instant,
    ) {
        let mut rng = thread_rng();
        let move_down_speed = rng.gen_range(0.0..=1.0);
        self.add_star_at_movedownspeed(
            buf_size,
            pos,
            now - self
                .move_down_time_from_move_down_speed(move_down_speed)
                .mul_f64(rng.gen_range(0.0..1.0)),
            move_down_speed,
        )
    }
    /// add a star at a specified position with a specified move_down_speed (0 -> minimum, 1 -> maximum)
    fn add_star_at_movedownspeed(
        &mut self,
        _buf_size: (usize, usize),
        pos: (usize, usize),
        now: Instant,
        move_down_speed: f64,
    ) {
        let move_down_time = self.move_down_time_from_move_down_speed(move_down_speed);
        let slower_stars = self
            .stars
            .iter()
            .take_while(|star| star.move_down_time > move_down_time)
            .count();
        // minimum brightness of the r, g, or b color channel of the stars color
        let mut rng = thread_rng();
        let b = rng.gen_range(150..=205);
        let g = b;
        let r = rng.gen_range(b..=b + 50);
        let a = u8::MAX;
        // uphold order: slower stars first, faster stars later -> faster stars are "closer" to the camera, so they hide slower stars if two stars have to share a pixel
        self.stars.insert(
            slower_stars,
            Star {
                pos,
                fg: Color { r, g, b, a },
                // draw different star symbols depending on the star's speed (note: slower -> further away)
                char: if move_down_speed > 0.95 {
                    '|'
                } else if move_down_speed > 0.67 {
                    '°'
                } else {
                    '.'
                },
                last_moved_down: now,
                move_down_time,
            },
        )
    }

    fn move_down_time_from_move_down_speed(&self, move_down_speed: f64) -> Duration {
        Duration::from_secs_f64(
            self.seconds_per_down_movement_max
                - move_down_speed
                    * move_down_speed
                    * (self.seconds_per_down_movement_max - self.seconds_per_down_movement_min),
        )
    }
}

#[derive(Clone, Copy)]
pub struct Star {
    pos: (usize, usize),
    fg: Color,
    char: char,
    last_moved_down: Instant,
    move_down_time: Duration,
}
impl Star {
    /// check if the star has moved down on this frame. if yes,
    /// this function sets the stars position and returns the position
    /// the star had before this function was called.
    pub fn moved_down(&mut self, now: Instant) -> Option<(usize, usize)> {
        let mut prev_pos = None;
        while now - self.last_moved_down > self.move_down_time {
            prev_pos.get_or_insert(self.pos);
            if thread_rng().gen_bool(0.02) {
                if thread_rng().gen_bool(0.5) {
                    self.pos.0 = self.pos.0.wrapping_add(1);
                } else {
                    self.pos.0 = self.pos.0.wrapping_sub(1);
                }
            }
            self.last_moved_down += self.move_down_time;
            self.pos.1 += 1;
        }
        prev_pos
    }
}
