use serde::{Deserialize, Serialize};

/// Collision bounds for objects that are more complicated
/// than a single rectangle, made up of multiple rectangles.
#[derive(Clone, PartialEq)]
pub struct CollisionBounds {
    rects: Vec<Rect>,
}

impl CollisionBounds {
    pub fn new(rects: Vec<Rect>) -> Self {
        Self { rects }
    }
}

/// A rectangle used for collision detection.
#[derive(Clone, Copy, PartialEq)]
pub struct Rect {
    pub left: f32,
    pub right: f32,
    pub top: f32,
    pub bottom: f32,
}

impl CollisionBounds {
    /// Returns the rectangles that make up `self`
    pub fn rects(&self) -> &[Rect] {
        &self.rects
    }
    /// Checks if any point in `other` is contained in `self`
    pub fn intersects(&self, other: &Self) -> bool {
        other.rects.iter().any(|o| self.intersects_rect(*o))
    }
    /// Checks if any point in `rect` is contained in `self`
    pub fn intersects_rect(&self, rect: Rect) -> bool {
        self.rects.iter().any(|s| s.intersects(rect))
    }
}

impl Rect {
    pub fn new_xyxy(left: f32, top: f32, right: f32, bottom: f32) -> Self {
        Self {
            left,
            right,
            top,
            bottom,
        }
    }
    pub fn new_xxyy(left: f32, right: f32, top: f32, bottom: f32) -> Self {
        Self::new_xyxy(left, top, right, bottom)
    }
    pub fn new_xywh(x: f32, y: f32, w: f32, h: f32) -> Self {
        Self {
            left: x,
            right: x + w,
            top: y,
            bottom: y + h,
        }
    }
    pub fn new_xwyh(x: f32, w: f32, y: f32, h: f32) -> Self {
        Self::new_xywh(x, y, w, h)
    }
    /// Returns the smaller of the two `x` coordinates: either `left` or `right`
    pub fn x1(self) -> f32 {
        self.left.min(self.right)
    }
    /// Returns the larger of the two `x` coordinates: either `left` or `right`
    pub fn x2(self) -> f32 {
        self.left.max(self.right)
    }
    /// Returns the smaller of the two `y` coordinates: either `top` or `bottom`
    pub fn y1(self) -> f32 {
        self.top.min(self.bottom)
    }
    /// Returns the larger of the two `y` coordinates: either `top` or `bottom`
    pub fn y2(self) -> f32 {
        self.top.max(self.bottom)
    }
    /// Checks if the given point is contained in `self`.
    /// If the point is on the edge of `self`, `true` is returned.
    pub fn contains_point(self, x: f32, y: f32) -> bool {
        self.x1() <= x && x <= self.x2() && self.y1() <= y && y <= self.y2()
    }
    /// Checks if any point on the vertical line `(x,y1)->(x,y2)` is contained in `self`
    pub fn intersects_vertical_line(self, x: f32, y1: f32, y2: f32) -> bool {
        let (y1, y2) = (y1.min(y2), y1.max(y2));
        let x_intersecs = x >= self.x1() && x <= self.x2();
        let y_intersecs = y2 >= self.y1() && y1 <= self.y2();
        x_intersecs && y_intersecs
    }
    /// Checks if any point in `other` is contained in `self`
    pub fn intersects(self, other: Self) -> bool {
        let x_intersecs = other.x2() >= self.x1() && other.x1() <= self.x2();
        let y_intersecs = other.y2() >= self.y1() && other.y1() <= self.y2();
        x_intersecs && y_intersecs
    }
}

#[test]
fn test_rect_new() {
    let rect_ez = Rect::new_xyxy(-1.0, -2.0, 1.0, 2.0);
    assert!(rect_ez.left == -1.0);
    assert!(rect_ez.top == -2.0);
    assert!(rect_ez.right == 1.0);
    assert!(rect_ez.bottom == 2.0);
    assert!(rect_ez.x1() == -1.0);
    assert!(rect_ez.y1() == -2.0);
    assert!(rect_ez.x2() == 1.0);
    assert!(rect_ez.y2() == 2.0);

    let rect_inv = Rect::new_xyxy(1.0, 2.0, -1.0, -2.0);
    assert!(rect_inv.left == 1.0);
    assert!(rect_inv.top == 2.0);
    assert!(rect_inv.right == -1.0);
    assert!(rect_inv.bottom == -2.0);
    assert!(rect_inv.x1() == -1.0);
    assert!(rect_inv.y1() == -2.0);
    assert!(rect_inv.x2() == 1.0);
    assert!(rect_inv.y2() == 2.0);
}

#[test]
fn test_rect_intersects() {
    assert!(Rect::new_xyxy(1.0, 2.0, 3.0, 4.0).intersects(Rect::new_xyxy(1.0, 2.0, 3.0, 4.0)));
    assert!(Rect::new_xyxy(1.0, 2.0, 3.0, 4.0).intersects(Rect::new_xyxy(0.0, 1.0, 2.0, 3.0)));
    assert!(Rect::new_xyxy(1.0, 2.0, 3.0, 4.0).intersects(Rect::new_xyxy(2.0, 1.0, 4.0, 3.0)));
    assert!(!Rect::new_xyxy(1.0, 2.0, 3.0, 4.0).intersects(Rect::new_xyxy(4.0, 2.0, 5.0, 4.0)));
}

impl Serialize for CollisionBounds {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        self.rects.serialize(serializer)
    }
}
impl<'de> Deserialize<'de> for CollisionBounds {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        Ok(Self::new(Deserialize::deserialize(deserializer)?))
    }
}

impl Serialize for Rect {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        (self.left, self.top, self.right, self.bottom).serialize(serializer)
    }
}
impl<'de> Deserialize<'de> for Rect {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        let (left, top, right, bottom) = Deserialize::deserialize(deserializer)?;
        Ok(Self {
            left,
            right,
            top,
            bottom,
        })
    }
}
