//
// Config structs have a S-prefixed equivalent
// with `#[derive(Serialize, Deserialize)]`
// these are used internally for parsing and formatting,
// but are converted to/from the non-S-structs when
// the `config` module is used from elsewhere.
//

use std::{
    fmt::Display,
    path::{Path, PathBuf},
};

use asciivation_duty_renderer::world::{
    Assets, ElemInOrder, Object, ObjectInWorld, SObjRef, SObject, SWorld, SortBy, World,
};
use serde::{Deserialize, Serialize};

use crate::{
    collision::CollisionBounds,
    popups::{Popup, Popups},
    websocket::WsColor,
};

/// The struct used to represent the configuration file's contents
pub struct Config {
    instructor_disabled: bool,
    rel_height: f32,
    fps_limit: f32,
    lives: usize,
    websocket: Option<String>,
    damage_effect: DamageEffect,
    controls_key_mode: ControlsKeyMode,
    selected_spaceship: usize,
    /// non-empty
    spaceships: Vec<Spaceship>,

    pub(crate) path: PathBuf,

    pub(crate) minmax_health: (f32, f32),
    pub(crate) minmax_shot_power: (f32, f32),
    pub(crate) minmax_shot_rate: (f32, f32),
    pub(crate) minmax_speed: (f32, f32),

    changed: bool,
}
/// A one-to-one representation of the config file.
/// Defines how the config file is read and written.
#[derive(Serialize, Deserialize)]
struct SConfig {
    #[serde(default, skip_serializing_if = "std::ops::Not::not")]
    instructor_disabled: bool,
    #[serde(default = "rel_height_default")]
    rel_height: f32,
    #[serde(default = "fps_limit_default")]
    fps_limit: f32,
    lives: usize,
    #[serde(default, skip_serializing_if = "Option::is_none")]
    websocket: Option<String>,
    #[serde(default)]
    damage_effect: DamageEffect,
    #[serde(default)]
    controls_key_mode: ControlsKeyMode,
    #[serde(default)]
    selected_spaceship: usize,
    spaceships: Vec<SSpaceship>,
}
fn rel_height_default() -> f32 {
    2.0
}
fn fps_limit_default() -> f32 {
    60.0
}

/// The setting that determines what, if any, visual effect
/// should be displayed when the spaceship takes damage
#[derive(Clone, Copy, PartialEq, Eq, Serialize, Deserialize, Default)]
pub enum DamageEffect {
    None,
    TextOnly,
    #[default]
    FlashBlack,
    FlashRed,
}

/// The setting that determines how keyboard inputs
/// are interpreted while in-game
#[derive(Clone, Copy, PartialEq, Eq, Serialize, Deserialize, Default)]
pub enum ControlsKeyMode {
    Hold,
    #[default]
    ToggleIfUnsupportedTerminal,
    Toggle,
}

/// The spaceship type used by the application at runtime.
/// Contains more information than is saved in the config file.
pub struct Spaceship {
    pub name: String,
    /// at most 24-char long phrase describing the spaceship,
    /// or just stating something more or less funny
    pub phrase: String,
    pub description: String,
    pub health: f32,
    pub shot_power: f32,
    pub shot_rate: f32,
    pub speed: f32,
    pub shot_spawn_pos: Vec<(f32, f32, f32)>,
    /// the 3d model for this spaceship
    pub model: MaybeColored<Object>,
    model_path: PathBuf,
    pub bounds: CollisionBounds,
    /// an empty world containing the 3d model.
    /// used in the spaceship selection screen.
    pub preview_world: World,
    pub model_in_preview_world: ObjectInWorld,
    pub model_in_preview_world_order: ElemInOrder,
    /// contains models to which the game switches once a shot is fired,
    /// until the shot cooldown has elapsed.
    /// the float is until what point in the countdown the object should be displayed.
    /// 0.0 -> will be displayed for at most one frame, 1.0 -> will be displayed in the end.
    /// if there was one value in this list, with the float set to `0.5`, the spaceship would
    /// transition to that object and then back to the default spaceship every time a shot is fired.
    pub models_if_shooting: MaybeColored<Vec<(Object, f32)>>,
}
/// Contains at least one `T`, the `no_color` variant, and optionally more `T`s,
/// which are associated with one of the 9 supported spaceship colors.
pub struct MaybeColored<T> {
    pub no_color: T,
    pub colored: [Option<T>; 9],
}
impl<T> MaybeColored<T> {
    pub fn no(t: T) -> Self {
        Self {
            no_color: t,
            colored: [None, None, None, None, None, None, None, None, None],
        }
    }
    pub fn get(&self, color: WsColor) -> &T {
        self.colored[match color {
            WsColor::Red => 0,
            WsColor::Blue => 1,
            WsColor::Yellow => 2,
            WsColor::Orange => 3,
            WsColor::Green => 4,
            WsColor::Violet => 5,
            WsColor::White => 6,
            WsColor::Black => 7,
            WsColor::Transparent => 8,
        }]
        .as_ref()
        .unwrap_or(&self.no_color)
    }
}
/// A spaceship as defined in the config file
#[derive(Serialize, Deserialize)]
struct SSpaceship {
    name: String,
    phrase: String,
    description: String,
    health: f32,
    shot_power: f32,
    shot_rate: f32,
    speed: f32,
    shot_spawn_pos: Vec<(f32, f32, f32)>,
    model: PathBuf,
    bounds: CollisionBounds,
}
/// A spaceship's (effective) stats, used by the highscores system
#[derive(Clone, PartialEq, Serialize, Deserialize)]
pub struct SpaceshipStats {
    pub name: String,
    pub health: usize,
    pub shot_power: f32,
    pub shot_rate: f32,
    pub speed: f32,
    pub shot_spawn_pos: Vec<(f32, f32, f32)>,
    pub bounds: CollisionBounds,
}

impl Config {
    /// Try to load the config from the specified file path
    pub fn from_json(path: &(impl AsRef<Path> + ?Sized)) -> Result<Self, ConfigFromJsonError> {
        let path = path.as_ref();
        // read the file
        let json = std::fs::read_to_string(path)
            .map_err(|e| ConfigFromJsonError::CouldNotReadConfigFile(path.to_owned(), e))?;
        Self::from_json_internal(&json, path)
    }
    /// Try to load the config from the specified string, as if it was loaded from a file.
    /// Since the file path the config was loaded from is unknown when constructing a `Config`
    /// from a string, the config's `.save()` method will never perform any I/O operations.
    pub fn from_json_str(json: &str) -> Result<Self, ConfigFromJsonError> {
        Self::from_json_internal(json, Path::new(""))
    }
    fn from_json_internal(json: &str, path: &Path) -> Result<Self, ConfigFromJsonError> {
        // parse json into required format
        let json = serde_json::from_str::<SConfig>(json)
            .map_err(|e| ConfigFromJsonError::InvalidJson(path.to_owned(), e))?;

        // check if there is at least one spaceship
        if json.spaceships.is_empty() {
            return Err(ConfigFromJsonError::NoSpaceshipsDefined(path.to_owned()));
        }

        let mut spaceships = Vec::with_capacity(json.spaceships.len());
        for spaceship in json.spaceships {
            let parent = path.parent().map(|p| p.to_owned()).unwrap_or_default();
            let model_path = parent.join(&spaceship.model);
            let model_str = std::fs::read_to_string(&model_path).map_err(|e| {
                ConfigFromJsonError::CouldNotReadSpaceshipModelFile(model_path.clone(), e)
            })?;
            let model_world = toml::from_str::<SWorld>(&model_str)
                .map_err(|e| ConfigFromJsonError::InvalidModelToml(model_path.clone(), e))?;
            let model_world = model_world.optimize(&mut Assets::new_in(parent.to_owned()));
            if model_world.load_err {
                return Err(ConfigFromJsonError::InvalidModelBadWorld(
                    model_path.to_owned(),
                ));
            }
            let mut model = MaybeColored::no(
                model_world
                    .find_object("Ship")
                    .map(|v| model_world.object(v).clone())
                    .ok_or_else(|| ConfigFromJsonError::InvalidModelNoObject(model_path.clone()))?,
            );
            // search for and load alternative models, if any
            let mut models_if_shooting: MaybeColored<Vec<(Object, f32)>> = MaybeColored::no(vec![]);
            for obj in model_world.objects() {
                if obj.name.starts_with("Ship") {
                    let mut name_cond = &obj.name[4..];
                    // if it continues with "Color(Red/Blue/...)", we are working on a colored spaceship
                    let color = if name_cond.starts_with("Color") {
                        name_cond = &name_cond[5..];
                        let (color, len) = if name_cond.starts_with("Red") {
                            (0, 3)
                        } else if name_cond.starts_with("Blue") {
                            (1, 4)
                        } else if name_cond.starts_with("Yellow") {
                            (2, 6)
                        } else if name_cond.starts_with("Orange") {
                            (3, 6)
                        } else if name_cond.starts_with("Green") {
                            (4, 5)
                        } else if name_cond.starts_with("Violet") {
                            (5, 6)
                        } else if name_cond.starts_with("White") {
                            (6, 5)
                        } else if name_cond.starts_with("Black") {
                            (7, 5)
                        } else if name_cond.starts_with("Transparent") {
                            (8, 11)
                        } else {
                            return Err(ConfigFromJsonError::InvalidShipColor(
                                model_path.clone(),
                                obj.name.clone(),
                            ));
                        };
                        name_cond = &name_cond[len..];
                        Some(color)
                    } else {
                        None
                    };
                    if let Some(args) = name_cond.strip_prefix("IfShooting") {
                        if let Ok(p) = args.trim().parse::<f32>() {
                            if !p.is_finite() {
                                return Err(ConfigFromJsonError::InvalidShipConditioned(
                                    model_path.clone(),
                                    "IfShooting",
                                    args.to_owned(),
                                    "number must be finite".to_owned(),
                                ));
                            }
                            if !(0.0..=1.0).contains(&p) {
                                return Err(ConfigFromJsonError::InvalidShipConditioned(
                                    model_path.clone(),
                                    "IfShooting",
                                    args.to_owned(),
                                    "number must be between 0 and 1 (both inclusive)".to_owned(),
                                ));
                            }
                            if let Some(color) = color {
                                models_if_shooting.colored[color]
                                    .get_or_insert_with(Vec::new)
                                    .push((obj.clone(), p));
                            } else {
                                models_if_shooting.no_color.push((obj.clone(), p));
                            }
                        } else {
                            return Err(ConfigFromJsonError::InvalidShipConditioned(model_path.clone(), "IfShooting", args.to_owned(), "usage: ShipIfShooting<p>, where <p> is a number with 0 <= <p> <= 1".to_owned()));
                        }
                    } else if let Some(color) = color {
                        model.colored[color] = Some(obj.clone());
                    }
                }
            }
            for m in [&mut models_if_shooting.no_color].into_iter().chain(
                models_if_shooting
                    .colored
                    .iter_mut()
                    .filter_map(|v| v.as_mut()),
            ) {
                m.sort_by(|a, b| a.1.total_cmp(&b.1));
            }

            let mut preview_world = SWorld {
                objects: vec![SObject {
                    name: "Spaceship".to_owned(),
                    import: vec![],
                    parts: vec![],
                }],
                order_by: SortBy::Order,
                order: vec![SObjRef {
                    obj: Some("Spaceship".to_owned()),
                    id: Some("Spaceship".to_owned()),
                    condition: None,
                    min: Some((-0.5, -0.5, -0.5)),
                    max: Some((0.5, 0.5, 0.5)),
                    rotation: None,
                }],
                conditions: vec![],
            }
            .optimize(&mut Assets::new_empty());
            let model_in_preview_world = preview_world.find_object("Spaceship").unwrap();
            let model_in_preview_world_order = preview_world.find_in_order("Spaceship").unwrap();
            *preview_world.object_mut(model_in_preview_world) = model.no_color.clone();
            spaceships.push(Spaceship {
                name: spaceship.name,
                phrase: spaceship.phrase,
                description: spaceship.description,
                health: spaceship.health,
                shot_power: spaceship.shot_power,
                shot_rate: spaceship.shot_rate,
                speed: spaceship.speed,
                shot_spawn_pos: spaceship.shot_spawn_pos,
                model,
                model_path: spaceship.model,
                bounds: spaceship.bounds,
                preview_world,
                model_in_preview_world,
                model_in_preview_world_order,
                models_if_shooting,
            });
        }

        // return converted config
        fn sminmax(spaceships: &[Spaceship], map: impl Fn(&Spaceship) -> f32) -> (f32, f32) {
            (
                spaceships
                    .iter()
                    .map(&map)
                    .min_by(f32::total_cmp)
                    .unwrap()
                    .min(0.0),
                spaceships
                    .iter()
                    .map(&map)
                    .max_by(f32::total_cmp)
                    .unwrap()
                    .max(0.0),
            )
        }
        Ok(Config {
            instructor_disabled: json.instructor_disabled,
            lives: json.lives,
            rel_height: json.rel_height,
            fps_limit: json.fps_limit,
            selected_spaceship: json.selected_spaceship.min(spaceships.len() - 1),

            path: path.to_path_buf(),

            minmax_health: sminmax(&spaceships, |s| s.health),
            minmax_shot_power: sminmax(&spaceships, |s| s.shot_power),
            minmax_shot_rate: sminmax(&spaceships, |s| s.shot_rate),
            minmax_speed: sminmax(&spaceships, |s| s.speed),

            websocket: json.websocket,

            damage_effect: json.damage_effect,

            controls_key_mode: json.controls_key_mode,

            spaceships,

            changed: false,
        })
    }

    /// Save the config to the file from which it was loaded, if any, if it has changed.
    /// Errors are reported through `popups`.
    pub fn save(&mut self, popups: &mut Popups) {
        if !self.path.as_os_str().is_empty() && self.changed {
            if let Ok(str) = serde_json::to_string_pretty(&SConfig {
                instructor_disabled: self.instructor_disabled,
                rel_height: self.rel_height,
                fps_limit: self.fps_limit,
                lives: self.lives,
                websocket: self.websocket.clone(),
                damage_effect: self.damage_effect,
                controls_key_mode: self.controls_key_mode,
                selected_spaceship: self.selected_spaceship,
                spaceships: self
                    .spaceships
                    .iter()
                    .map(|s| SSpaceship {
                        name: s.name.clone(),
                        phrase: s.phrase.clone(),
                        description: s.description.clone(),
                        health: s.health,
                        shot_power: s.shot_power,
                        shot_rate: s.shot_rate,
                        speed: s.speed,
                        shot_spawn_pos: s.shot_spawn_pos.clone(),
                        model: s.model_path.clone(),
                        bounds: s.bounds.clone(),
                    })
                    .collect(),
            }) {
                if let Err(e) = std::fs::write(&self.path, str) {
                    popups.show_error(Popup::new(
                        "failed to save config",
                        format!(
                            "the config was changed, but it could not be saved to {:?}.\n{e}",
                            self.path
                        ),
                    ));
                } else {
                    self.changed = false;
                }
            } else {
                popups.show_error(Popup::new(
                    "failed to save config",
                    "the config was changed, but it could not be saved (internal error).",
                ));
            }
        }
    }

    pub fn set_instructor_disabled(&mut self, instructor_disabled: bool) {
        if self.instructor_disabled != instructor_disabled {
            self.instructor_disabled = instructor_disabled;
            self.changed = true;
        }
    }
    pub fn instructor_disabled(&self) -> bool {
        self.instructor_disabled
    }

    pub fn set_rel_height(&mut self, rel_height: f32) {
        if rel_height != self.rel_height {
            self.rel_height = rel_height;
            self.changed = true;
        }
    }
    pub fn rel_height(&self) -> f32 {
        self.rel_height
    }
    pub fn set_fps_limit(&mut self, fps_limit: f32) {
        if fps_limit != self.fps_limit {
            self.fps_limit = fps_limit;
            self.changed = true;
        }
    }
    pub fn fps_limit(&self) -> f32 {
        self.fps_limit
    }

    pub fn lives(&self) -> usize {
        self.lives
    }

    pub fn websocket(&self) -> Option<&str> {
        self.websocket.as_deref()
    }

    pub fn damage_effect(&self) -> DamageEffect {
        self.damage_effect
    }
    pub fn set_damage_effect(&mut self, damage_effect: DamageEffect) {
        self.changed = true;
        self.damage_effect = damage_effect;
    }

    pub fn controls_key_mode(&self) -> ControlsKeyMode {
        self.controls_key_mode
    }
    pub fn set_controls_key_mode(&mut self, controls_key_mode: ControlsKeyMode) {
        self.changed = true;
        self.controls_key_mode = controls_key_mode;
    }

    /// Returns the available spaceships as an iterator.
    /// This is guaranteed to have at least one element.
    pub fn spaceships(&self) -> impl Iterator<Item = &Spaceship> {
        self.spaceships.iter()
    }
    /// How many spaceships are available.
    /// This is guaranteed to be at least `1`, so you can safely do `config.spaceships_count() - 1`.
    pub fn spaceships_count(&self) -> usize {
        self.spaceships.len()
    }
    /// Returns mutable references to the available spaceships as an iterator.
    /// This is guaranteed to have at least one element.
    pub fn spaceships_mut(&mut self) -> impl Iterator<Item = &mut Spaceship> {
        self.spaceships.iter_mut()
    }
    /// Returns the spaceship with the given index,
    /// or the last spaceship if the index is out of bounds.
    pub fn spaceship(&self, index: usize) -> &Spaceship {
        self.spaceships
            .get(index)
            .unwrap_or_else(|| self.spaceships.last().unwrap())
    }
    /// Equivalent to `config.spaceship(config.selected_spaceship_index())`
    pub fn selected_spaceship(&self) -> &Spaceship {
        &self.spaceships[self.selected_spaceship]
    }
    pub fn selected_spaceship_index(&self) -> usize {
        self.selected_spaceship
    }
    /// Select the spaceship with the given index,
    /// or the last spaceship if the index is out of bounds
    pub fn select_spaceship(&mut self, index: usize) {
        self.selected_spaceship = index.min(self.spaceships.len() - 1);
        self.changed = true;
    }

    pub(crate) fn selected_spaceships_stats(&self) -> SpaceshipStats {
        self.spaceships_stats(self.selected_spaceship_index())
    }
    pub(crate) fn spaceships_stats(&self, index: usize) -> SpaceshipStats {
        let spaceship = self.spaceship(index);
        SpaceshipStats {
            name: spaceship.name.clone(),
            health: (self.lives() as f32 * spaceship.health).round() as _,
            shot_power: spaceship.shot_power,
            shot_rate: spaceship.shot_rate,
            speed: spaceship.speed,
            shot_spawn_pos: spaceship.shot_spawn_pos.clone(),
            bounds: spaceship.bounds.clone(),
        }
    }
}

#[derive(Debug)]
pub enum ConfigFromJsonError {
    CouldNotReadConfigFile(PathBuf, std::io::Error),
    /// serde_json failed to parse the json string into the required format
    InvalidJson(PathBuf, serde_json::Error),
    CouldNotReadSpaceshipModelFile(PathBuf, std::io::Error),
    /// the spaceships list was empty. at least one spaceship must exist for the config to be valid
    NoSpaceshipsDefined(PathBuf),
    InvalidModelToml(PathBuf, toml::de::Error),
    InvalidModelNoObject(PathBuf),
    InvalidModelBadWorld(PathBuf),
    InvalidShipConditioned(PathBuf, &'static str, String, String),
    InvalidShipColor(PathBuf, String),
}

impl std::error::Error for ConfigFromJsonError {}
impl Display for ConfigFromJsonError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::CouldNotReadConfigFile(path, e) => {
                write!(f, "could not read config file {path:?}: {e}")
            }
            Self::InvalidJson(path, e) => {
                write!(f, "could not parse config file {path:?}: {e}")
            }
            Self::CouldNotReadSpaceshipModelFile(path, e) => {
                write!(f, "could not read spaceship model file {path:?}: {e}")
            }
            Self::NoSpaceshipsDefined(path) => {
                write!(f, "no spaceships are defined in {path:?} (only an empty list, which is not allowed - at least one spaceship must be defined)")
            }
            Self::InvalidModelToml(path, e) => {
                write!(f, "invalid model {path:?}: {e}")
            }
            Self::InvalidModelNoObject(path) => {
                write!(f, "model {path:?} defines no object named \"Ship\"")
            }
            Self::InvalidModelBadWorld(path) => {
                write!(f, "model {path:?} is invalid")
            }
            Self::InvalidShipConditioned(path, cond_type, args, msg) => {
                write!(f, "invalid conditioned ship model with condition type {cond_type} and args {args} in {path:?}: {msg}")
            }
            Self::InvalidShipColor(path, color) => {
                write!(f, "invalid ship color in {path:?} in: {color}")
            }
        }
    }
}
