/// defines the application's command line interface
pub mod args;
/// the background visible on most of the application's screens
pub mod background;
pub mod collision;
/// The configuration file and its contents.
/// Types prefixed with `S` are only used for Serialization/Deserialization.
pub mod config;
/// The different screens of the application
pub mod parts;
/// Popups, used to alert the user if an error occurs
pub mod popups;
/// The highscores system
pub mod scores;
/// State shared between the applications' screens
pub mod state;
#[cfg(test)]
pub(crate) mod tests;
/// Handles getting the spaceship color from the websocket.
pub mod websocket;

use std::{io::stdout, path::Path, process::ExitCode, time::Duration};

use asciivation_duty_renderer::{
    init_undo,
    part::{DrawMode, Settings},
};
use config::Config;
use parts::{main_menu::MainMenu, AnyPart, AnyPartRunner};
use state::GlobalState;

fn main() -> ExitCode {
    // parse CLI arguments
    let args = args::parse();

    // load config
    let config = match Config::from_json(
        args.config
            .as_deref()
            .unwrap_or(Path::new("starfighter_alliance_config.json")),
    ) {
        Ok(config) => config,
        Err(e) => {
            eprintln!("Error in config: {e}");
            return ExitCode::FAILURE;
        }
    };

    // query the terminal for some information required by the renderer
    let res = match asciivation_duty_renderer::init_check() {
        Ok(res) => res,
        Err(err) => {
            eprintln!("Error while checking terminal support: {err}");
            return ExitCode::FAILURE;
        }
    };
    // initialize the game's global state and settings
    let mut settings = Settings::new(&res, DrawMode::DiffOnly);
    settings.rel_height = config.rel_height();
    settings.frame_time = if !config.fps_limit().is_finite() || config.fps_limit() < 4.0 {
        Duration::ZERO
    } else {
        Duration::from_secs_f32(1.0 / config.fps_limit())
    };

    let mut global_state = GlobalState::new(&res, config);
    global_state.better_obstacle_spawning = !args.disallow_fun;

    // query color from websocket if one is specified in config
    if let Some(websocket) = global_state.config.websocket() {
        websocket::start_thread(websocket.to_owned(), &mut global_state);
    }

    // initialize the main menu
    let part = MainMenu::new();

    // lock stdout for better performance.
    // this is the program's communication channel with the terminal
    let mut stdout = stdout().lock();
    // initialize the terminal and get a PartRunner on success
    let mut part_runner =
        match asciivation_duty_renderer::init_do(part, global_state, &res, settings, &mut stdout) {
            Ok(part_runner) => AnyPartRunner::MainMenu(part_runner),
            Err(err) => {
                eprintln!("Error while initializing terminal: {err}");
                // try to uninitialize the terminal, ignoring errors if they occur
                // to not distract users with non-relevant errors
                let _ = init_undo(res, stdout);
                return ExitCode::FAILURE;
            }
        };
    // terminal is initialized, run game parts
    loop {
        part_runner = match part_runner {
            AnyPartRunner::MainMenu(mut part_runner) => match part_runner.run(&mut stdout) {
                None => break,
                Some(next_part) => AnyPartRunner::new(next_part, part_runner),
            },
            AnyPartRunner::GameState(mut part_runner) => {
                AnyPartRunner::new(part_runner.run(&mut stdout), part_runner)
            }
            AnyPartRunner::EndScreen(mut part_runner) => {
                AnyPartRunner::new(part_runner.run(&mut stdout), part_runner)
            }
            AnyPartRunner::SpaceshipSelectionScreen(mut part_runner) => {
                AnyPartRunner::new(part_runner.run(&mut stdout), part_runner)
            }
            AnyPartRunner::OptionsScreen(mut part_runner) => {
                AnyPartRunner::new(AnyPart::MainMenu(part_runner.run(&mut stdout)), part_runner)
            }
        };
    }
    // program should exit, so we need to reset the terminal to its previous state
    let errs = init_undo(res, stdout);
    // if there are errors, show them
    for err in errs {
        eprintln!("Error while uninitializing terminal: {err}");
    }
    // exit the program
    ExitCode::SUCCESS
}
