use std::time::{Duration, Instant};

use asciivation_duty_renderer::{
    deps::crossterm::event::{Event, KeyCode, KeyEvent, KeyEventKind, KeyModifiers},
    part::{Part, PartRunner},
    render::{clear_pre_render, render, Color, CLR_WHITE, PX_BLACK},
    world::{ElemInOrder, World},
};
use uianimator::{default_animator_f64_quadratic::DefaultAnimatorF64Quadratic, Animator};

use crate::{background, popups::Popups, state::GlobalState};

use super::{fov_scale, main_menu::MainMenu, AnyPart};

pub struct EndScreen {
    world: World,
    spaceship_in_order: ElemInOrder,
    spaceship_size: (f32, f32, f32),
    spaceship_pos: (
        DefaultAnimatorF64Quadratic,
        DefaultAnimatorF64Quadratic,
        DefaultAnimatorF64Quadratic,
    ),
    score: usize,
    loaded: Instant,
    changed_scores: Instant,
    scores_type: u8,
    ending: Option<(Instant, bool)>,
}

impl EndScreen {
    pub fn new(world: World, spaceship_in_order: ElemInOrder, score: usize) -> Self {
        let speed = 1.0;
        let spaceship = world.elem_in_order(spaceship_in_order);
        let spaceship_size = (
            spaceship.max.0 - spaceship.min.0,
            spaceship.max.1 - spaceship.min.1,
            spaceship.max.2 - spaceship.min.2,
        );
        let mut spaceship_pos = (
            DefaultAnimatorF64Quadratic::new(spaceship.min.0 as f64, speed),
            DefaultAnimatorF64Quadratic::new(spaceship.min.1 as f64, speed * 2.0),
            DefaultAnimatorF64Quadratic::new(spaceship.min.2 as f64, speed),
        );
        let now = Instant::now();
        spaceship_pos.0.set_target(-spaceship_size.2 as f64, now);
        spaceship_pos.1.set_target(5.0, now);
        spaceship_pos
            .2
            .set_target(-spaceship_size.2 as f64 / 2.0, now);
        Self {
            world,
            spaceship_in_order,
            spaceship_size,
            spaceship_pos,
            score,
            loaded: now,
            changed_scores: now,
            scores_type: 0,
            ending: None,
        }
    }
}

const MAX_SCORES_TYPE: u8 = 4;

impl Part for EndScreen {
    type End = AnyPart;
    type State = GlobalState;
    fn draw(runner: &mut PartRunner<Self>, now: Instant) {
        if runner.draw_reset {
            // make star background opaque
            runner.state.background.set_transparency(0.0);
        }

        let elapsed_since_loaded = now - runner.part.loaded;

        // move spaceship smoothly
        {
            let spaceship = runner
                .part
                .world
                .elem_in_order_mut(runner.part.spaceship_in_order);
            spaceship.min = (
                runner.part.spaceship_pos.0.get_value(now) as _,
                runner.part.spaceship_pos.1.get_value(now) as _,
                runner.part.spaceship_pos.2.get_value(now) as _,
            );
            spaceship.max = (
                spaceship.min.0 + runner.part.spaceship_size.0,
                spaceship.min.1 + runner.part.spaceship_size.1,
                spaceship.min.2 + runner.part.spaceship_size.2,
            );
        }

        if !runner.screen_buf.is_empty() {
            // fill entire screen buffer with black pixels
            clear_pre_render(&mut runner.screen_buf, PX_BLACK);

            background::draw(runner, now);

            // show current spaceship
            render(
                &runner.part.world,
                (0.0, 10.0, 0.0),
                (0.0, std::f32::consts::FRAC_PI_2, 0.0),
                runner.settings.rel_height,
                fov_scale(
                    runner.screen_buf[0].len(),
                    runner.screen_buf.len(),
                    runner.settings.rel_height,
                ),
                1.0,
                &mut runner.screen_buf,
            );

            // show highscores
            let highscores_type = match runner.part.scores_type {
                0 => "all highscores >> ".to_owned(),
                1 => "<< same-ship highscores >> ".to_owned(),
                2 => format!(
                    "<< {}'s highscores >> ",
                    runner
                        .state
                        .highscores
                        .score_unchecked(runner.part.score)
                        .player_name
                ),
                3 => format!(
                    "<< {}'s same-ship highscores >> ",
                    runner
                        .state
                        .highscores
                        .score_unchecked(runner.part.score)
                        .player_name
                ),
                _ => String::new(),
            };
            let elapsed_since_changed_scores =
                (now - runner.part.changed_scores).min(elapsed_since_loaded / 4);
            let chars_count_factor = elapsed_since_changed_scores.as_secs_f32().min(0.25) * 4.0;
            let chars_count_factor = 1.0 - (1.0 - chars_count_factor) * (1.0 - chars_count_factor);
            let chars_count =
                (highscores_type.chars().count() as f32 * chars_count_factor) as usize;
            let skip = runner.screen_buf[0].len().saturating_sub(chars_count);
            for (px, char) in runner.screen_buf[0]
                .iter_mut()
                .skip(skip)
                .zip(highscores_type.chars())
            {
                px.fg = CLR_WHITE;
                px.char = char;
            }
            for ((i, (si, score)), row) in runner
                .state
                .highscores
                .scores()
                .iter()
                .enumerate()
                .filter({
                    let scores_type = runner.part.scores_type;
                    let my_score_index = runner.part.score;
                    let my_score = runner.state.highscores.score_unchecked(my_score_index);
                    move |(i, score)| {
                        score
                            .obstacle_spawn_mode
                            .same_mode(&my_score.obstacle_spawn_mode)
                            && (*i == my_score_index
                                || match scores_type {
                                    0 => true,
                                    1 => score.spaceship == my_score.spaceship,
                                    2 => score.player_name == my_score.player_name,
                                    3 => {
                                        score.player_name == my_score.player_name
                                            && score.spaceship == my_score.spaceship
                                    }
                                    _ => false,
                                })
                    }
                })
                .enumerate()
                .zip(runner.screen_buf.iter_mut().skip(2))
            {
                let mut score_fmt = format!("{}  |  {}", score.player_name, score.points)
                    .chars()
                    .collect::<Vec<_>>();
                let seconds = elapsed_since_changed_scores.as_secs_f32();
                let p_seconds = if seconds > 15.0 {
                    runner.part.changed_scores += Duration::from_secs(15);
                    if runner.part.scores_type >= MAX_SCORES_TYPE {
                        runner.part.scores_type = 0;
                    } else {
                        runner.part.scores_type += 1;
                    }
                    seconds % 15.0
                } else {
                    seconds
                };
                let p = (p_seconds * 4.0 - 0.25 * i as f32).clamp(0.0, 1.0);
                score_fmt.truncate(score_fmt.len().saturating_sub(
                    (score_fmt.len().max(10) as f32 * (1.0 - p) * (1.0 - p) * (1.0 - p)).round()
                        as _,
                ));
                let skip = row.len().saturating_sub(score_fmt.len());
                for (char, px) in score_fmt.into_iter().zip(row.iter_mut().skip(skip)) {
                    px.fg = px.fg.merge(Color::rgba(
                        180,
                        if si == runner.part.score { 255 } else { 180 },
                        180,
                        (255.0 / 1.05f32.powi(i as _)).round() as _,
                    ));
                    px.char = char;
                }
            }

            let score = runner.state.highscores.score_unchecked(runner.part.score);

            for (i, header_text) in [
                format!("You scored {} points!", score.points),
                String::new(),
                format!("Spaceship: {}", score.spaceship.name),
                if score.shots_fired == 0 {
                    "you didn't fire a single shot this round".to_owned()
                } else if score.shots_hit + score.shots_missed == 0 {
                    format!("{} shots were fired this round", score.shots_fired)
                } else {
                    format!(
                        "{} shots were fired this round, {}% hit something",
                        score.shots_fired,
                        (100.0 * score.shots_hit as f32
                            / (score.shots_hit + score.shots_missed) as f32)
                            .round() as u8
                    )
                },
                format!("Mode: {}", score.obstacle_spawn_mode),
                String::new(),
                "Your Name:".to_owned(),
                format!(
                    "> {}{}",
                    score.player_name,
                    if elapsed_since_loaded.subsec_millis() < 500 {
                        " "
                    } else {
                        "_"
                    },
                ),
            ]
            .into_iter()
            .enumerate()
            {
                // above the spaceship
                let height = runner.screen_buf.len();
                let header_text_row = &mut runner.screen_buf[(height / 4).saturating_sub(4) + i];
                let header_text_chars = header_text.len();
                let margin_left = header_text_row.len().saturating_sub(header_text_chars) / 2;
                for (px, char) in header_text_row[margin_left..]
                    .iter_mut()
                    .zip(header_text.chars())
                {
                    let clr = (elapsed_since_loaded.as_millis().min(255 * 4) / 4) as u8;
                    px.fg = Color::rgb(clr, clr, clr);
                    px.char = char;
                }
            }

            // instructor
            if !runner.state.config.instructor_disabled() {
                let clr = Color::rgb(100, 100, 100);
                let lines = [
                    "To save your score, enter your name and press Enter.",
                    "If you do not want your score to be saved, press Esc.",
                ];
                for (i, line) in lines.into_iter().rev().enumerate() {
                    let y = runner.screen_buf.len().saturating_sub(i + 1);
                    if let Some(row) = runner.screen_buf.get_mut(y) {
                        let skip = row.len().saturating_sub(line.len()) / 2;
                        for (px, ch) in row.iter_mut().skip(skip).zip(line.chars()) {
                            px.fg = px.fg.merge(clr);
                            px.char = ch;
                        }
                    }
                }
            }
        }

        Popups::draw(&mut runner.screen_buf, &mut runner.state, now);

        if let Some((ending, delete)) = &mut runner.part.ending {
            let elapsed = ((now - *ending).as_secs_f32() * 2.0).min(1.0);
            let clr = Color::rgba(0, 0, 0, (255.0 * elapsed) as u8);
            for px in runner.screen_buf.iter_mut().flat_map(|v| v.iter_mut()) {
                px.bg = px.bg.merge(clr);
                px.fg = px.fg.merge(clr);
            }
            if elapsed >= 1.0 {
                if *delete {
                    *delete = false;
                    runner
                        .state
                        .highscores
                        .remove_score_unchecked(runner.part.score);
                }
                runner.state.background.set_transparency_now(1.0);
                runner.end = Some(AnyPart::MainMenu(MainMenu::new()));
            }
        }
    }
    fn event(runner: &mut PartRunner<Self>, e: Event) {
        if super::event(runner, &e) || runner.state.popups.event(&e) {
            return;
        }
        match e {
            Event::Key(KeyEvent {
                code: KeyCode::Char(ch),
                modifiers: KeyModifiers::NONE | KeyModifiers::SHIFT,
                kind: KeyEventKind::Press | KeyEventKind::Repeat,
                state: _,
            }) if !ch.is_control() => {
                runner
                    .state
                    .highscores
                    .score_mut_unchecked(runner.part.score)
                    .player_name
                    .push(ch);
            }
            Event::Key(KeyEvent {
                code: KeyCode::Backspace,
                modifiers: KeyModifiers::NONE | KeyModifiers::SHIFT,
                kind: KeyEventKind::Press | KeyEventKind::Repeat,
                state: _,
            }) => {
                runner
                    .state
                    .highscores
                    .score_mut_unchecked(runner.part.score)
                    .player_name
                    .pop();
            }
            Event::Key(KeyEvent {
                code: KeyCode::Left | KeyCode::Up,
                kind: KeyEventKind::Press,
                ..
            }) => {
                if runner.part.scores_type > 0 {
                    runner.part.scores_type -= 1;
                    runner.part.changed_scores = Instant::now();
                }
            }
            Event::Key(KeyEvent {
                code: KeyCode::Right | KeyCode::Down,
                kind: KeyEventKind::Press,
                ..
            }) => {
                if runner.part.scores_type < MAX_SCORES_TYPE {
                    runner.part.scores_type += 1;
                    runner.part.changed_scores = Instant::now();
                }
            }
            // save score and go back to main menu
            Event::Key(KeyEvent {
                kind: KeyEventKind::Press,
                code: KeyCode::Enter,
                ..
            }) => {
                if runner.part.ending.is_none()
                    && !runner
                        .state
                        .highscores
                        .score_unchecked(runner.part.score)
                        .player_name
                        .trim()
                        .is_empty()
                {
                    runner.part.ending = Some((Instant::now(), false));
                }
            }
            // do not save score and go back to main menu
            Event::Key(KeyEvent {
                kind: KeyEventKind::Press,
                code: KeyCode::Esc,
                ..
            }) => {
                if let Some((ending, _)) = &mut runner.part.ending {
                    *ending -= Duration::from_secs(1);
                } else {
                    runner.part.ending = Some((Instant::now(), true));
                }
            }
            _ => (),
        }
    }
}
