#[cfg(test)]
pub mod tests;

use std::{
    collections::VecDeque,
    fmt::Display,
    time::{Duration, Instant},
};

use asciivation_duty_renderer::{
    deps::crossterm::event::{
        Event, KeyCode, KeyEvent, KeyEventKind, MouseButton, MouseEvent, MouseEventKind,
    },
    part::{Part, PartRunner, Settings},
    render::{clear_pre_render, render, Color, CLR_BLACK, CLR_WHITE, PX_BLACK},
    world::{
        Assets, ElemInOrder, ObjRef, Object, ObjectInWorld, SObjRef, SObject, SShape, SWorld,
        SortBy, World,
    },
};
use large_numbers::{
    LARGE_NUM_0, LARGE_NUM_1, LARGE_NUM_2, LARGE_NUM_3, LARGE_NUM_4, LARGE_NUM_5, LARGE_NUM_6,
    LARGE_NUM_7, LARGE_NUM_8, LARGE_NUM_9, LARGE_NUM_DOT, LARGE_NUM_MINUS,
};
use rand::{seq::SliceRandom, Rng};
use serde::{Deserialize, Serialize};
use uianimator::{default_animator_f64_quadratic::DefaultAnimatorF64Quadratic, Animator};

use crate::{
    background,
    collision::Rect,
    config::{ControlsKeyMode, DamageEffect},
    scores::Score,
    state::GlobalState,
};

use super::{end_screen::EndScreen, fov_scale, AnyPart};

pub type SpaceshipPosOverrideFunction =
    Box<dyn FnMut(&mut ObjRef, f64, Instant, f32, f32, usize, usize, f32, f32) -> (f32, f32)>;

pub struct GameState {
    world: World,
    border_world: World,
    spaceship_in_world: ObjectInWorld,
    spaceship_in_order: ElemInOrder,
    loaded: Instant,
    last_frame: Instant,
    speed_multiplier: f32,
    speed_multiplier_og: f32,
    spaceship_pos: (f64, f64),
    spaceship_pos_override: Option<SpaceshipPosOverrideFunction>,
    cam_rot: (
        DefaultAnimatorF64Quadratic,
        DefaultAnimatorF64Quadratic,
        DefaultAnimatorF64Quadratic,
    ),
    mouse_pos: (u16, u16),
    left_mouse_down: bool,
    movement_keys: (bool, bool, bool, bool),
    paused: Option<Instant>,
    paused_selected_option: usize,

    obstacle_objects: Vec<Object>,
    spawned_obstacles: VecDeque<Obstacle>,

    shooting: bool,
    last_shot: Instant,
    shot_objects_smaller: Vec<Object>,
    shot_objects_larger: Vec<Object>,
    shots: Vec<Obstacle>,
    shots_fired: u32,
    shots_hit: u32,
    shots_missed: u32,

    waveless_obstacle_spawning_target_count: f32,
    wave_stats: Vec<u8>,

    obstacle_spawn_mode: ObstacleSpawnMode,

    game_over: bool,

    score: usize,
    players_health: usize,
    /// if, within this many seconds, the spaceship takes damage twice, it becomes immune for another period of this many seconds.
    live_loose_immunity_time: f32,
    last_live_lost_time: Option<(Instant, bool)>,
}
struct Obstacle {
    right: f64,
    up: f64,
    ymin: f32,
    ymax: f32,
    height_half: f64,
    width_half: f64,
    speed: f64,
    world: World,
    elem_in_world: ElemInOrder,
    destroyed: f64,
    wave: usize,
    spawn_time: f64,
    destroy_time: f64,
}

#[derive(Clone, Copy, PartialEq, Eq, Serialize, Deserialize)]
pub enum ObstacleSpawnMode {
    Waveless,
    Lastenheft { wave: usize },
}
impl ObstacleSpawnMode {
    pub fn same_mode(&self, other: &Self) -> bool {
        match (self, other) {
            (Self::Waveless, Self::Waveless) => true,
            (Self::Waveless, _) => false,
            (Self::Lastenheft { .. }, Self::Lastenheft { .. }) => true,
            (Self::Lastenheft { .. }, _) => false,
        }
    }
}

const CAM_ROT_ANIMATOR_SPEED: f64 = 1.0;

impl GameState {
    pub fn new(
        state: &GlobalState,
        spaceship_pos_override_during_countdown: Option<SpaceshipPosOverrideFunction>,
    ) -> Self {
        // world with an empty object
        let world = SWorld {
            objects: vec![SObject {
                name: "Spaceship".to_owned(),
                import: vec![],
                parts: vec![],
            }],
            order_by: SortBy::Order,
            order: vec![SObjRef {
                obj: Some("Spaceship".to_owned()),
                id: Some("Spaceship".to_owned()),
                condition: None,
                min: Some((-0.5, -0.5, -0.5)),
                max: Some((0.5, 0.5, 0.5)),
                rotation: None,
            }],
            conditions: vec![],
        }
        .optimize(&mut Assets::new_empty());
        let spaceship_in_world = world.find_object("Spaceship").unwrap();
        let spaceship_in_order = world.find_in_order("Spaceship").unwrap();
        let now = Instant::now();
        Self {
            world,
            border_world: SWorld {
                objects: vec![
                    SObject {
                        name: "BorderT".to_owned(),
                        import: vec![],
                        parts: vec![SShape::Quad(
                            (0.0, 0.0, 0.0),
                            (0.0, 0.0, 1.0),
                            (1.0, 0.0, 1.0),
                            (1.0, 0.0, 0.0),
                            Color::rgb(20, 20, 20),
                        )],
                    },
                    SObject {
                        name: "BorderB".to_owned(),
                        import: vec![],
                        parts: vec![SShape::Quad(
                            (0.0, 0.0, 0.0),
                            (0.0, 0.0, 1.0),
                            (1.0, 0.0, 1.0),
                            (1.0, 0.0, 0.0),
                            Color::rgb(20, 20, 20),
                        )],
                    },
                    SObject {
                        name: "BorderL".to_owned(),
                        import: vec![],
                        parts: vec![SShape::Quad(
                            (0.0, 0.0, 0.0),
                            (0.0, 0.0, 1.0),
                            (1.0, 0.0, 1.0),
                            (1.0, 0.0, 0.0),
                            Color::rgb(20, 20, 20),
                        )],
                    },
                    SObject {
                        name: "BorderR".to_owned(),
                        import: vec![],
                        parts: vec![SShape::Quad(
                            (0.0, 0.0, 0.0),
                            (0.0, 0.0, 1.0),
                            (1.0, 0.0, 1.0),
                            (1.0, 0.0, 0.0),
                            Color::rgb(20, 20, 20),
                        )],
                    },
                ],
                order_by: SortBy::Order,
                order: vec![
                    SObjRef {
                        obj: Some("BorderT".to_owned()),
                        id: None,
                        condition: None,
                        min: Some((3.0, 0.0, -99.0)),
                        max: Some((99.0, 0.0, 99.0)),
                        rotation: None,
                    },
                    SObjRef {
                        obj: Some("BorderB".to_owned()),
                        id: None,
                        condition: None,
                        min: Some((-99.0, 0.0, -99.0)),
                        max: Some((-3.0, 0.0, 99.0)),
                        rotation: None,
                    },
                    SObjRef {
                        obj: Some("BorderL".to_owned()),
                        id: None,
                        condition: None,
                        min: Some((-99.0, 0.0, -99.0)),
                        max: Some((99.0, 0.0, -4.0)),
                        rotation: None,
                    },
                    SObjRef {
                        obj: Some("BorderR".to_owned()),
                        id: None,
                        condition: None,
                        min: Some((-99.0, 0.0, 4.0)),
                        max: Some((99.0, 0.0, 99.0)),
                        rotation: None,
                    },
                ],
                conditions: vec![],
            }
            .optimize(&mut Assets::new_empty()),
            spaceship_in_world,
            spaceship_in_order,
            loaded: now,
            last_frame: now,
            speed_multiplier: 1.5,
            speed_multiplier_og: 1.5,
            spaceship_pos: (0.0, -1.0),
            left_mouse_down: false,
            spaceship_pos_override: spaceship_pos_override_during_countdown,
            cam_rot: (
                DefaultAnimatorF64Quadratic::new(0.0, CAM_ROT_ANIMATOR_SPEED),
                DefaultAnimatorF64Quadratic::new(
                    std::f64::consts::FRAC_PI_2,
                    CAM_ROT_ANIMATOR_SPEED,
                ),
                DefaultAnimatorF64Quadratic::new(0.0, CAM_ROT_ANIMATOR_SPEED),
            ),
            mouse_pos: (u16::MAX, u16::MAX),
            movement_keys: (false, false, false, false),
            paused: None,
            paused_selected_option: 0,
            obstacle_objects: SWorld {
                objects: vec![
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (1.0, 0.0, 0.0),
                                (1.0, 0.0, 1.0),
                                (0.0, 0.0, 1.0),
                                Color::rgb(200, 200, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.0, 0.5, 0.0),
                                (0.0, 0.5, 1.0),
                                (0.0, 0.0, 1.0),
                                Color::rgb(200, 200, 200),
                            ),
                            SShape::Quad(
                                (1.0, 0.0, 0.0),
                                (1.0, 0.5, 0.0),
                                (1.0, 0.5, 1.0),
                                (1.0, 0.0, 1.0),
                                Color::rgb(200, 200, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.0, 0.5, 0.0),
                                (1.0, 0.5, 0.0),
                                (1.0, 0.0, 0.0),
                                Color::rgb(200, 200, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 1.0),
                                (0.0, 0.5, 1.0),
                                (1.0, 0.5, 1.0),
                                (1.0, 0.0, 1.0),
                                Color::rgb(200, 200, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.5, 0.0),
                                (1.0, 0.5, 0.0),
                                (1.0, 0.5, 1.0),
                                (0.0, 0.5, 1.0),
                                Color::rgb(255, 255, 255),
                            ),
                        ],
                    },
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (1.0, 0.0, 0.0),
                                (1.0, 0.0, 1.0),
                                (0.0, 0.0, 1.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.0, 0.5, 0.0),
                                (0.0, 0.5, 1.0),
                                (0.0, 0.0, 1.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (1.0, 0.0, 0.0),
                                (1.0, 0.5, 0.0),
                                (1.0, 0.5, 1.0),
                                (1.0, 0.0, 1.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.0, 0.5, 0.0),
                                (1.0, 0.5, 0.0),
                                (1.0, 0.0, 0.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 1.0),
                                (0.0, 0.5, 1.0),
                                (1.0, 0.5, 1.0),
                                (1.0, 0.0, 1.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.5, 0.0),
                                (1.0, 0.5, 0.0),
                                (1.0, 0.5, 1.0),
                                (0.0, 0.5, 1.0),
                                Color::rgb(191, 191, 255),
                            ),
                        ],
                    },
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (1.0, 0.0, 0.0),
                                (1.0, 0.0, 1.0),
                                (0.0, 0.0, 1.0),
                                Color::rgb(150, 120, 120),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.5, 0.5, 0.0),
                                (0.0, 0.5, 0.5),
                                (0.0, 0.0, 1.0),
                                Color::rgb(150, 120, 120),
                            ),
                            SShape::Quad(
                                (1.0, 0.0, 0.0),
                                (1.0, 0.5, 0.5),
                                (0.5, 0.5, 1.0),
                                (1.0, 0.0, 1.0),
                                Color::rgb(150, 120, 120),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.5, 0.5, 0.0),
                                (1.0, 0.5, 0.5),
                                (1.0, 0.0, 0.0),
                                Color::rgb(150, 120, 120),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 1.0),
                                (0.0, 0.5, 0.5),
                                (0.5, 0.5, 1.0),
                                (1.0, 0.0, 1.0),
                                Color::rgb(150, 120, 120),
                            ),
                            SShape::Quad(
                                (0.5, 0.5, 0.0),
                                (1.0, 0.5, 0.5),
                                (0.5, 0.5, 1.0),
                                (0.0, 0.5, 0.5),
                                Color::rgb(180, 150, 150),
                            ),
                        ],
                    },
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (1.0, 0.0, 0.0),
                                (1.0, 0.0, 1.0),
                                (0.0, 0.0, 1.0),
                                Color::rgb(180, 200, 180),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.25, 0.5, 0.25),
                                (0.25, 0.5, 0.75),
                                (0.0, 0.0, 1.0),
                                Color::rgb(180, 200, 180),
                            ),
                            SShape::Quad(
                                (1.0, 0.0, 0.0),
                                (0.75, 0.5, 0.25),
                                (0.75, 0.5, 0.75),
                                (1.0, 0.0, 1.0),
                                Color::rgb(180, 200, 180),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.25, 0.5, 0.25),
                                (0.75, 0.5, 0.25),
                                (1.0, 0.0, 0.0),
                                Color::rgb(180, 200, 180),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 1.0),
                                (0.25, 0.5, 0.75),
                                (0.75, 0.5, 0.75),
                                (1.0, 0.0, 1.0),
                                Color::rgb(180, 200, 180),
                            ),
                            SShape::Quad(
                                (0.25, 0.5, 0.25),
                                (0.75, 0.5, 0.25),
                                (0.75, 0.5, 0.75),
                                (0.25, 0.5, 0.75),
                                Color::rgb(210, 255, 210),
                            ),
                        ],
                    },
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (1.0, 0.0, 0.0),
                                (1.0, 0.0, 1.0),
                                (0.0, 0.0, 1.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.0, 0.5, 0.0),
                                (0.0, 0.5, 1.0),
                                (0.0, 0.0, 1.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (1.0, 0.0, 0.0),
                                (1.0, 0.5, 0.0),
                                (1.0, 0.5, 1.0),
                                (1.0, 0.0, 1.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 0.0),
                                (0.0, 0.5, 0.0),
                                (1.0, 0.5, 0.0),
                                (1.0, 0.0, 0.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.0, 1.0),
                                (0.0, 0.5, 1.0),
                                (1.0, 0.5, 1.0),
                                (1.0, 0.0, 1.0),
                                Color::rgb(150, 150, 200),
                            ),
                            SShape::Quad(
                                (0.0, 0.5, 0.0),
                                (1.0, 0.5, 0.0),
                                (1.0, 0.5, 1.0),
                                (0.0, 0.5, 1.0),
                                Color::rgb(191, 191, 255),
                            ),
                        ],
                    },
                ],
                order_by: SortBy::Order,
                order: vec![],
                conditions: vec![],
            }
            .optimize(&mut Assets::new_empty())
            .objects()
            .cloned()
            .collect(),
            spawned_obstacles: VecDeque::new(),
            shooting: false,
            last_shot: now,
            shot_objects_smaller: SWorld {
                objects: vec![
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![SShape::Line(
                            (0.0, 0.5, 0.5),
                            (1.0, 0.5, 0.5),
                            Color::rgb(200, 64, 64),
                        )],
                    },
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![SShape::Line(
                            (0.0, 0.5, 0.5),
                            (1.0, 0.5, 0.5),
                            Color::rgb(230, 64, 64),
                        )],
                    },
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![SShape::Line(
                            (0.0, 0.5, 0.5),
                            (1.0, 0.5, 0.5),
                            Color::rgb(230, 100, 100),
                        )],
                    },
                ],
                order_by: SortBy::Order,
                order: vec![],
                conditions: vec![],
            }
            .optimize(&mut Assets::new_empty())
            .objects()
            .cloned()
            .collect(),
            shot_objects_larger: SWorld {
                objects: vec![
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![SShape::Quad(
                            (0.0, 0.5, 0.0),
                            (0.0, 0.5, 1.0),
                            (1.0, 0.5, 1.0),
                            (1.0, 0.5, 0.0),
                            Color::rgb(200, 64, 64),
                        )],
                    },
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![SShape::Quad(
                            (0.0, 0.5, 0.0),
                            (0.0, 0.5, 1.0),
                            (1.0, 0.5, 1.0),
                            (1.0, 0.5, 0.0),
                            Color::rgb(230, 64, 64),
                        )],
                    },
                    SObject {
                        name: String::new(),
                        import: vec![],
                        parts: vec![SShape::Quad(
                            (0.0, 0.5, 0.0),
                            (0.0, 0.5, 1.0),
                            (1.0, 0.5, 1.0),
                            (1.0, 0.5, 0.0),
                            Color::rgb(230, 100, 100),
                        )],
                    },
                ],
                order_by: SortBy::Order,
                order: vec![],
                conditions: vec![],
            }
            .optimize(&mut Assets::new_empty())
            .objects()
            .cloned()
            .collect(),
            shots: vec![],
            shots_fired: 0,
            shots_hit: 0,
            shots_missed: 0,
            obstacle_spawn_mode: if state.better_obstacle_spawning {
                ObstacleSpawnMode::Waveless
            } else {
                ObstacleSpawnMode::Lastenheft { wave: 0 }
            },
            waveless_obstacle_spawning_target_count: 1.0,
            wave_stats: vec![],
            game_over: false,

            score: 0,
            players_health: (state.config.lives() as f32 * state.config.selected_spaceship().health)
                .round() as _,
            live_loose_immunity_time: 0.5,
            last_live_lost_time: None,
        }
    }

    fn toggle_pause(&mut self) {
        if let Some(paused) = self.paused.take() {
            let elapsed = (Instant::now() - paused).saturating_sub(Duration::from_secs_f64(0.25));
            self.loaded += elapsed;
            self.last_frame += elapsed;
            if let Some((t, _)) = &mut self.last_live_lost_time {
                *t += elapsed;
            }
        } else if self.players_health > 0 {
            self.paused = Some(Instant::now());
            self.paused_selected_option = 0;
        }
    }

    fn lose(runner: &mut PartRunner<Self>) {
        let score = runner.state.highscores.add(Score::new(
            runner.part.score,
            runner.part.obstacle_spawn_mode,
            runner.state.config.selected_spaceships_stats(),
            runner.part.shots_fired,
            runner.part.shots_hit,
            runner.part.shots_missed,
        ));
        runner.end = Some(AnyPart::EndScreen(EndScreen::new(
            runner.part.world.clone(),
            runner.part.spaceship_in_order,
            score,
        )));
    }
}

const COUNTDOWN_DURATION: f64 = 2.0;
impl Part for GameState {
    type End = AnyPart;
    type State = GlobalState;
    fn draw(runner: &mut PartRunner<Self>, real_now: Instant) {
        let now = if let Some(paused) = runner.part.paused {
            let paused_for = ((real_now - paused).as_secs_f64() / 0.5).min(1.0);
            paused + Duration::from_secs_f64(0.25 - (1.0 - paused_for) * (1.0 - paused_for) / 4.0)
        } else {
            real_now
        };
        if runner.draw_reset || runner.state.draw_reset {
            runner.state.draw_reset = false;
            // put spaceship from config into world
            *runner.part.world.object_mut(runner.part.spaceship_in_world) = runner
                .state
                .config
                .selected_spaceship()
                .model
                .get(runner.state.spaceship_color)
                .clone();

            // make star background semitransparent
            runner.state.background.set_transparency(0.5);
        }

        if runner.part.players_health == 0 && !runner.part.game_over {
            runner.part.game_over = true;
            for o in &mut runner.part.spawned_obstacles {
                o.destroy();
            }
            runner.part.cam_rot.0.set_target(0.0, real_now);
            runner
                .part
                .cam_rot
                .1
                .set_target(std::f64::consts::FRAC_PI_2, real_now);
            runner.part.cam_rot.2.set_target(0.0, real_now);
        }

        // fill entire screen buffer with black pixels
        clear_pre_render(&mut runner.screen_buf, PX_BLACK);

        let mut rng = rand::thread_rng();
        let delta_time = (now - runner.part.last_frame).as_secs_f64();
        runner.part.last_frame = now;

        runner.part.waveless_obstacle_spawning_target_count +=
            delta_time as f32 / runner.part.waveless_obstacle_spawning_target_count.sqrt();

        let elapsed_time = (now - runner.part.loaded).as_secs_f64();
        let elapsed_time_rel = elapsed_time / COUNTDOWN_DURATION;
        let countdown = elapsed_time_rel <= 1.0;

        // spaceship movement
        if !countdown && runner.part.paused.is_none() {
            if runner.part.left_mouse_down {
                // move towards mouse
                if !runner.screen_buf.is_empty() && !runner.screen_buf[0].is_empty() {
                    let width = runner.screen_buf[0].len();
                    let height = runner.screen_buf.len();
                    let (angle_x, angle_y) = asciivation_duty_renderer::render::angles_from_2d(
                        (
                            runner.part.mouse_pos.0 as f32,
                            runner.part.mouse_pos.1 as f32,
                        ),
                        (width as f32, height as f32),
                        runner.settings.rel_height,
                        fov_scale(width, height, runner.settings.rel_height),
                    );
                    let now = Instant::now();
                    let cam_rot = (
                        runner.part.cam_rot.0.get_value(now),
                        runner.part.cam_rot.1.get_value(now),
                        runner.part.cam_rot.2.get_value(now),
                    );
                    let (cam_y, cam_height, cam_x) = (
                        -10.0 * cam_rot.1.cos() * cam_rot.0.cos(),
                        10.0 * cam_rot.1.sin(),
                        -10.0 * cam_rot.1.cos() * cam_rot.0.sin(),
                    );
                    let angle_y = angle_y as f64 + cam_rot.1;
                    let pointer_y = cam_height / angle_y.tan();
                    let pointer_x = cam_height * (angle_x as f64).tan() / cam_rot.1.sin();
                    let (rot_sin, rot_cos) = cam_rot.0.sin_cos();
                    let (pointer_x, pointer_y) = (
                        cam_x + pointer_x * rot_cos + rot_sin * pointer_y,
                        cam_y + pointer_y * rot_cos - rot_sin * pointer_x,
                    );
                    let desired_position = (pointer_x * 2.0 / 3.5, pointer_y * 1.5 / 2.5);
                    let desired_movement_vector = (
                        desired_position.0 - runner.part.spaceship_pos.0,
                        desired_position.1 - runner.part.spaceship_pos.1,
                    );
                    let movement_vector_len_sq = desired_movement_vector.0
                        * desired_movement_vector.0
                        + desired_movement_vector.1 * desired_movement_vector.1;
                    if movement_vector_len_sq.is_finite() && movement_vector_len_sq > 0.0 {
                        let factor = (runner.part.speed_multiplier as f64
                            * runner.state.config.selected_spaceship().speed as f64
                            * delta_time
                            / movement_vector_len_sq.sqrt())
                        .min(1.0);
                        runner.part.spaceship_pos.0 = (runner.part.spaceship_pos.0
                            + factor * desired_movement_vector.0)
                            .clamp(-2.0, 2.0);
                        runner.part.spaceship_pos.1 = (runner.part.spaceship_pos.1
                            + factor * desired_movement_vector.1)
                            .clamp(-1.5, 1.5);
                    }
                }
            } else {
                // move by keyboard input
                let delta_movement = if runner.part.movement_keys.0 != runner.part.movement_keys.1
                    && runner.part.movement_keys.2 != runner.part.movement_keys.3
                {
                    runner.part.speed_multiplier as f64 * std::f64::consts::FRAC_1_SQRT_2
                } else {
                    runner.part.speed_multiplier as f64
                } * delta_time;
                let speed = runner.state.config.selected_spaceship().speed as f64;
                if runner.part.movement_keys.0 != runner.part.movement_keys.1 {
                    runner.part.spaceship_pos.1 = (runner.part.spaceship_pos.1
                        + delta_movement
                            * if runner.part.movement_keys.0 {
                                speed
                            } else {
                                -speed
                            })
                    .clamp(-1.5, 1.5);
                }
                if runner.part.movement_keys.2 != runner.part.movement_keys.3 {
                    runner.part.spaceship_pos.0 = (runner.part.spaceship_pos.0
                        + delta_movement
                            * if runner.part.movement_keys.2 {
                                speed
                            } else {
                                -speed
                            })
                    .clamp(-2.0, 2.0);
                }
            }
        }

        let fov_scale = fov_scale(
            runner.screen_buf[0].len(),
            runner.screen_buf.len(),
            runner.settings.rel_height,
        );

        let spaceship_elem = runner
            .part
            .world
            .elem_in_order_mut(runner.part.spaceship_in_order);

        // set spaceship position
        if !runner.screen_buf.is_empty() {
            // x = -2 --> min.x = -3.8
            // x = +2 --> max.x = +3.8 -> min.x = +2.8
            // y = -1.5 --> min.y = -2.85
            // y = +1.5 --> max.y = +2.85 -> min.y = 1.85
            // max.x = min.x + 1, max.y = min.y + 1
            let p1 = -runner.part.spaceship_pos.0 / 4.0 + 0.5;
            let p2 = -runner.part.spaceship_pos.1 / 3.0 + 0.5;
            let (mut x, mut z) = (
                (-3.0 * p2 + 2.0 * (1.0 - p2)) as _,
                (-4.0 * p1 + 3.0 * (1.0 - p1)) as _,
            );

            if countdown && runner.part.spaceship_pos_override.is_some() {
                let (x2, z2) = (runner.part.spaceship_pos_override.as_mut().unwrap())(
                    spaceship_elem,
                    elapsed_time_rel,
                    now,
                    x,
                    z,
                    runner.screen_buf[0].len(),
                    runner.screen_buf.len(),
                    runner.settings.rel_height,
                    fov_scale,
                );
                x = x2;
                z = z2;
            }

            spaceship_elem.min = (x, -0.5, z);
            spaceship_elem.max = (
                spaceship_elem.min.0 + 1.0,
                spaceship_elem.min.1 + 1.0,
                spaceship_elem.min.2 + 1.0,
            );
        }

        // move shots and remove ones that are no longer visible or hit an obstacle
        let mut remove_shots = vec![];
        let mut shot_bottom_edges = vec![];
        for (i, shot) in runner.part.shots.iter_mut().enumerate() {
            let shot_bottom_edge = shot.bottom_edge();
            shot.move_down(delta_time);
            if shot.bottom_edge() > 3.0 && shot.destroy() {
                runner.part.shots_missed += 1;
            }
            if shot.destroyed >= 1.0 {
                remove_shots.push(i);
            } else {
                shot_bottom_edges.push(shot_bottom_edge);
            }
        }
        for i in remove_shots.into_iter().rev() {
            runner.part.shots.remove(i);
        }

        // move obstacles and remove ones that are no longer visible or were hit by shots
        let mut remove_obstacles = vec![];
        for (i, obstacle) in runner.part.spawned_obstacles.iter_mut().enumerate() {
            let obstacle_top_edge = obstacle.top_edge();

            obstacle.move_down(delta_time);

            let mut obstacle_was_shot = false;
            if runner.part.players_health > 0 && obstacle.destroyed < 0.0 {
                for (i, shot) in runner.part.shots.iter_mut().enumerate() {
                    if shot.destroyed < 0.0
                        && Rect::new_xyxy(
                            obstacle.left_edge() as _,
                            obstacle_top_edge as _,
                            obstacle.right_edge() as _,
                            obstacle.bottom_edge() as _,
                        )
                        .intersects_vertical_line(
                            shot.right as _,
                            shot_bottom_edges[i] as _,
                            shot.top_edge() as _,
                        )
                    {
                        if shot.destroy() {
                            runner.part.shots_hit += 1;
                        }
                        obstacle_was_shot = true;
                        break;
                    }
                }
            }

            let obstacle_intersects_with_spaceship = runner.part.players_health > 0
                && obstacle.destroyed < 0.0
                && runner
                    .state
                    .config
                    .selected_spaceship()
                    .bounds
                    .intersects_rect(Rect::new_xyxy(
                        obstacle.left_edge() as f32 - spaceship_elem.min.2,
                        obstacle_top_edge as f32 - spaceship_elem.min.0,
                        obstacle.right_edge() as f32 - spaceship_elem.min.2,
                        obstacle.bottom_edge() as f32 - spaceship_elem.min.0,
                    ));

            if obstacle_intersects_with_spaceship
                && runner
                    .part
                    .last_live_lost_time
                    .as_ref()
                    .is_none_or(|(_, immunity)| !*immunity)
            {
                runner.part.players_health = runner.part.players_health.saturating_sub(1);
                runner.part.last_live_lost_time =
                    Some((now, runner.part.last_live_lost_time.is_some()));
            }

            if obstacle_intersects_with_spaceship || obstacle_was_shot {
                if obstacle.destroy() {
                    match runner.part.obstacle_spawn_mode {
                        ObstacleSpawnMode::Waveless => (),
                        ObstacleSpawnMode::Lastenheft { .. } => {
                            runner.part.wave_stats[obstacle.wave] |= 2;
                        }
                    }
                }
            } else if obstacle.top_edge() < -3.0 && obstacle.destroy() {
                match runner.part.obstacle_spawn_mode {
                    ObstacleSpawnMode::Waveless => runner.part.score += 1,
                    ObstacleSpawnMode::Lastenheft { .. } => {
                        runner.part.wave_stats[obstacle.wave] |= 4
                    }
                }
            }
            if obstacle.destroyed >= 1.0 {
                remove_obstacles.push(i);
            }
        }
        for i in remove_obstacles.into_iter().rev() {
            if let Some(obstacle) = runner.part.spawned_obstacles.remove(i) {
                if runner.part.players_health > 0 {
                    if let ObstacleSpawnMode::Lastenheft { wave } =
                        &mut runner.part.obstacle_spawn_mode
                    {
                        if runner
                            .part
                            .spawned_obstacles
                            .iter()
                            .all(|o| o.wave != obstacle.wave)
                        {
                            // this was the last obstacle of that wave
                            if runner.part.wave_stats[obstacle.wave] | 3 == 3 {
                                // no obstacle of that wave was missed (missed = 4u8, destroyed by spaceship = 2u8)
                                runner.part.score += (obstacle.wave + 1) * 3;
                            } else {
                                // wave completed, but not fully destroyed
                                runner.part.score += obstacle.wave + 1;
                            }
                            runner.part.wave_stats[obstacle.wave] |= 128;
                            if *wave < obstacle.wave + 1 {
                                *wave = obstacle.wave + 1;
                            }
                        }
                    }
                }
            }
        }

        // add obstacles / shots
        if !countdown && runner.part.players_health > 0 {
            // spawn shots
            let mut seconds_between_shots = runner.state.config.selected_spaceship().shot_rate;
            if seconds_between_shots > 0.0 {
                seconds_between_shots = 1.0 / seconds_between_shots;
            } else {
                seconds_between_shots = 0.0;
            }
            if runner.part.shooting
                && (now - runner.part.last_shot).as_secs_f32() >= seconds_between_shots
            {
                match (
                    runner.state.config.controls_key_mode(),
                    runner.settings.with_keyboard_enhancement,
                ) {
                    (ControlsKeyMode::Hold, _)
                    | (ControlsKeyMode::ToggleIfUnsupportedTerminal, true) => (),
                    (ControlsKeyMode::Toggle, _)
                    | (ControlsKeyMode::ToggleIfUnsupportedTerminal, false) => {
                        runner.part.shooting = false
                    }
                }
                runner.part.last_shot = (runner.part.last_shot
                    + Duration::from_secs_f32(seconds_between_shots))
                .max(runner.part.last_frame);
                // one pixel to the right from the screen's center -> get 3d width of that pixel
                let width = runner.screen_buf.first().map(|v| v.len()).unwrap_or(0);
                let height = runner.screen_buf.len();
                let pixel_width_in_3d = asciivation_duty_renderer::render::pos_from_2d(
                    (width as f32 / 2.0 + 1.0, height as f32 / 2.0),
                    (width as f32, height as f32),
                    runner.settings.rel_height,
                    fov_scale,
                )
                .0
                .abs()
                    * 10.0 /* camera's distance */;
                let shot_object = if pixel_width_in_3d > 0.035 {
                    &runner.part.shot_objects_smaller
                } else {
                    &runner.part.shot_objects_larger
                }
                .choose(&mut rng)
                .unwrap();
                for shot_spawn_pos in runner
                    .state
                    .config
                    .selected_spaceship()
                    .shot_spawn_pos
                    .iter()
                    .copied()
                {
                    runner.part.shots_fired += 1;
                    runner.part.shots.push(Obstacle::new_shot(
                        (spaceship_elem.min.2
                            + (spaceship_elem.max.2 - spaceship_elem.min.2) * shot_spawn_pos.2)
                            as _,
                        (spaceship_elem.min.0
                            + (spaceship_elem.max.0 - spaceship_elem.min.0) * shot_spawn_pos.0)
                            as _,
                        spaceship_elem.min.1
                            + (spaceship_elem.max.1 - spaceship_elem.min.1) * shot_spawn_pos.1,
                        0.1 + 0.1 * runner.state.config.selected_spaceship().shot_power as f64,
                        4.0 * runner.state.config.selected_spaceship().shot_power as f64,
                        shot_object.clone(),
                        pixel_width_in_3d as f64,
                    ));
                }
            }

            // spawn obstacles
            match &mut runner.part.obstacle_spawn_mode {
                ObstacleSpawnMode::Lastenheft { .. } => {
                    if runner
                        .part
                        .spawned_obstacles
                        .iter()
                        .all(|o| o.top_edge() < 0.0)
                    {
                        // entire wave destroyed or far enough away, spawn a new wave
                        let wave = runner.part.wave_stats.len();
                        runner.part.wave_stats.push(0);
                        // spawn obstacles
                        let speed = 0.6 + 0.03 * wave as f64;
                        // edges are at +-4, spaceship width is 1, so we need 0.75 units of space from the edges to the gap's center
                        let gap_width = 1.5;
                        let gap_width_half = gap_width / 2.0;
                        let gap_pos_min = -4.0 + gap_width_half;
                        let gap_pos_max = 4.0 - gap_width_half;
                        let gap_pos = rng.gen_range(gap_pos_min..=gap_pos_max);
                        let mut spawned_positions = vec![];
                        for _ in 0..=wave {
                            let width = rng.gen_range(0.1..=0.3);
                            let mut posl = rng.gen_range(-4.0..=(4.0 - width - gap_width));
                            if posl > gap_pos - gap_width_half {
                                posl += gap_width + width;
                            }
                            if spawned_positions
                                .iter()
                                .all(|(l, r)| *l > posl + width || *r < posl)
                            {
                                spawned_positions.push((posl, posl + width));
                                runner
                                    .part
                                    .spawned_obstacles
                                    .push_back(Obstacle::new_of_wave(
                                        posl + width / 2.0,
                                        3.5,
                                        width,
                                        width,
                                        speed,
                                        runner
                                            .part
                                            .obstacle_objects
                                            .choose(&mut rng)
                                            .unwrap()
                                            .clone(),
                                        wave,
                                    ));
                            }
                        }
                    }
                }
                ObstacleSpawnMode::Waveless => {
                    while runner.part.spawned_obstacles.len()
                        < runner.part.waveless_obstacle_spawning_target_count as usize
                    {
                        runner
                            .part
                            .spawned_obstacles
                            .push_back(Obstacle::new_waveless(
                                rng.gen_range(-4.0..=4.0),
                                3.5,
                                rng.gen_range(0.1..=0.3),
                                rng.gen_range(0.1..=0.3),
                                rng.gen_range(0.4..=2.0),
                                runner
                                    .part
                                    .obstacle_objects
                                    .choose(&mut rng)
                                    .unwrap()
                                    .clone(),
                            ));
                    }
                }
            }
        }

        if !runner.screen_buf.is_empty() {
            background::draw(runner, real_now);

            // let cam_pos = (0.0, 10.0, 0.0);
            // let cam_rot = (0.0, std::f32::consts::FRAC_PI_2, 0.0);
            let cam_rot = (
                runner.part.cam_rot.0.get_value(real_now) as f32,
                runner.part.cam_rot.1.get_value(real_now) as f32,
                runner.part.cam_rot.2.get_value(real_now) as f32,
            );
            let cam_pos = (
                -10.0 * cam_rot.1.cos() * cam_rot.0.cos(),
                10.0 * cam_rot.1.sin(),
                -10.0 * cam_rot.1.cos() * cam_rot.0.sin(),
            );

            // show shots
            for shot in &mut runner.part.shots {
                render(
                    shot.world(),
                    cam_pos,
                    cam_rot,
                    runner.settings.rel_height,
                    fov_scale,
                    1.0,
                    &mut runner.screen_buf,
                );
            }

            // show obstacles
            for obstacle in &mut runner.part.spawned_obstacles {
                render(
                    obstacle.world(),
                    cam_pos,
                    cam_rot,
                    runner.settings.rel_height,
                    fov_scale,
                    1.0,
                    &mut runner.screen_buf,
                );
            }

            // show current spaceship
            render(
                &runner.part.world,
                cam_pos,
                cam_rot,
                runner.settings.rel_height,
                fov_scale,
                1.0,
                &mut runner.screen_buf,
            );

            // set border color
            let mut f_left = if runner.part.spaceship_pos.0 < -1.0 {
                (runner.part.spaceship_pos.0.abs() - 1.0).min(1.0) as f32
            } else {
                0.0
            };
            let mut f_right = if runner.part.spaceship_pos.0 > 1.0 {
                (runner.part.spaceship_pos.0 - 1.0).min(1.0) as f32
            } else {
                0.0
            };
            let mut f_top = if runner.part.spaceship_pos.1 > 0.5 {
                (runner.part.spaceship_pos.1 - 0.5).min(1.0) as f32
            } else {
                0.0
            };
            let mut f_bottom = if runner.part.spaceship_pos.1 < -0.5 {
                (runner.part.spaceship_pos.1.abs() - 0.5).min(1.0) as f32
            } else {
                0.0
            };
            if runner.part.players_health == 0 {
                let f = 1.0
                    - runner
                        .part
                        .spawned_obstacles
                        .iter()
                        .map(|o| o.destroyed as f32)
                        .filter(|v| 0.0 <= *v && *v <= 1.0)
                        .min_by(f32::total_cmp)
                        .unwrap_or(1.0);
                f_left *= f;
                f_right *= f;
                f_top *= f;
                f_bottom *= f;
                runner.part.speed_multiplier = runner.part.speed_multiplier_og * f;
            }
            for (obj, f) in runner
                .part
                .border_world
                .objects_mut()
                .zip([f_top, f_bottom, f_left, f_right])
            {
                let clr = Color::rgb((65.0 * f * f) as u8, (25.0 * f) as u8, (25.0 * f) as u8);
                for shape in obj
                    .parts
                    .iter_mut()
                    .flat_map(|v| v.iter_mut())
                    .flat_map(|v| v.1.iter_mut())
                {
                    match &mut shape.0 {
                        asciivation_duty_renderer::world::Shape::Line(_, _, c) => *c = clr,
                        asciivation_duty_renderer::world::Shape::Quad(_, _, _, _, c) => *c = clr,
                        asciivation_duty_renderer::world::Shape::RawText(_, _, _, c, _) => *c = clr,
                    }
                }
            }
            render(
                &runner.part.border_world,
                cam_pos,
                cam_rot,
                runner.settings.rel_height,
                if countdown {
                    let p = 1.0 - (elapsed_time_rel * 2.0).min(1.0);
                    fov_scale * (1.0 + p * p) as f32
                } else {
                    fov_scale
                },
                1.0,
                &mut runner.screen_buf,
            );
        }

        // show status row
        if !runner.screen_buf.is_empty() {
            let status_row = match runner.part.obstacle_spawn_mode {
                ObstacleSpawnMode::Waveless => format!(
                    "Score: {}    Health: {}",
                    runner.part.score, runner.part.players_health
                ),
                ObstacleSpawnMode::Lastenheft { wave } => format!(
                    "Score: {}    Wave: {}/{}    Health: {}",
                    runner.part.score,
                    wave,
                    runner.part.wave_stats.len(),
                    runner.part.players_health
                ),
            };
            let row = runner.screen_buf.len() - 1;
            let row = &mut runner.screen_buf[row];
            for (char, px) in status_row.chars().zip(row.iter_mut()) {
                px.fg = if char.is_numeric() {
                    Color::rgb(180, 180, 180)
                } else {
                    Color::rgb(128, 128, 128)
                };
                px.char = char;
            }
        }

        // lost a live, show an effect
        if let Some((last_live_lost_time, _)) =
            runner
                .part
                .last_live_lost_time
                .or(if countdown { Some((now, false)) } else { None })
        {
            let p = 1.0
                - ((now - last_live_lost_time).as_secs_f32()
                    / runner.part.live_loose_immunity_time)
                    .min(1.0);
            if p == 0.0 {
                runner.part.last_live_lost_time = None;
            }
            if runner.state.config.damage_effect() != DamageEffect::None || countdown {
                let clr = match runner.state.config.damage_effect() {
                    DamageEffect::None | DamageEffect::TextOnly => CLR_BLACK,
                    DamageEffect::FlashBlack => Color::rgba(0, 0, 0, (128.0 * p).round() as _),
                    DamageEffect::FlashRed => Color::rgba(96, 0, 0, (128.0 * p).round() as _),
                };
                let text_clr = if countdown {
                    Color::rgba(
                        255,
                        255,
                        255,
                        (if elapsed_time_rel < 0.25 {
                            100.0
                                * (1.0
                                    - (elapsed_time_rel - 0.25) / 0.25 * (elapsed_time_rel - 0.25)
                                        / 0.25)
                        } else {
                            100.0
                                - 100.0 * (elapsed_time_rel - 0.25) / 0.75
                                    * (elapsed_time_rel - 0.25)
                                    / 0.75
                        }) as _,
                    )
                } else {
                    match runner.state.config.damage_effect() {
                        DamageEffect::None => CLR_BLACK,
                        DamageEffect::TextOnly | DamageEffect::FlashBlack => {
                            Color::rgba(128, 0, 0, (192.0 * p).round() as _)
                        }
                        DamageEffect::FlashRed => {
                            Color::rgba(200, 200, 200, (128.0 * p).round() as _)
                        }
                    }
                };
                if !countdown
                    && matches!(
                        runner.state.config.damage_effect(),
                        DamageEffect::FlashBlack | DamageEffect::FlashRed
                    )
                {
                    for row in &mut runner.screen_buf {
                        for px in row {
                            px.bg = px.bg.merge(clr);
                            px.fg = px.fg.merge(clr);
                        }
                    }
                }
                if !runner.screen_buf.is_empty() {
                    let chars = if countdown {
                        format!("{:.0}", (4.0 * (1.0 - elapsed_time_rel)).ceil())
                    } else {
                        format!("{}", runner.part.players_health)
                    }
                    .chars()
                    .filter_map(|v| {
                        Some(match v {
                            '0' => LARGE_NUM_0,
                            '1' => LARGE_NUM_1,
                            '2' => LARGE_NUM_2,
                            '3' => LARGE_NUM_3,
                            '4' => LARGE_NUM_4,
                            '5' => LARGE_NUM_5,
                            '6' => LARGE_NUM_6,
                            '7' => LARGE_NUM_7,
                            '8' => LARGE_NUM_8,
                            '9' => LARGE_NUM_9,
                            '-' => LARGE_NUM_MINUS,
                            '.' => LARGE_NUM_DOT,
                            _ => return None,
                        })
                    })
                    .collect::<Vec<_>>();
                    let char_count = chars.len();
                    let text_width = runner.screen_buf[0].len() / 2;
                    let text_height = runner.screen_buf.len() / 2;
                    let text_width_1 = 6 * char_count - 1;
                    let text_height_1 = 9;
                    let text_scale = (text_width as f32 / text_width_1 as f32).min(
                        text_height as f32 * runner.settings.rel_height / text_height_1 as f32,
                    );
                    let text_scale_x = text_scale as usize;
                    let text_scale_y = (text_scale / runner.settings.rel_height) as usize;
                    let text_width = text_scale_x * text_width_1;
                    let text_height = text_height_1 * text_scale_y;
                    let text_top = (runner.screen_buf.len() - text_height) / 2;
                    let text_left = (runner.screen_buf[0].len() - text_width) / 2;
                    for y in 0..9 {
                        for row in
                            &mut runner.screen_buf[text_top + y * text_scale_y..][..text_scale_y]
                        {
                            for (x, char) in chars.iter().enumerate() {
                                for i in 0..5 {
                                    for px in &mut row[text_left + (6 * x + i) * text_scale_x..]
                                        [..text_scale_x]
                                    {
                                        if char[y][i] {
                                            px.bg = px.bg.merge(text_clr);
                                            px.fg = px.fg.merge(text_clr);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        if let Some(paused) = runner.part.paused {
            if !runner.screen_buf.is_empty() {
                let p = ((real_now - paused).as_secs_f32() / 0.5).min(1.0);
                let p = 1.0 - (1.0 - p) * (1.0 - p);
                let clr = Color::rgba(0, 0, 0, 200);
                let h = runner.screen_buf.len();
                let w = runner.screen_buf[0].len();
                let x = (w as f32 * p) as usize;
                let y = (h as f32 * p) as usize;
                for row in runner.screen_buf.iter_mut().skip((h - y) / 2).take(y) {
                    for px in row.iter_mut().skip((w - x) / 2).take(x) {
                        px.fg = px.fg.merge(clr);
                        px.bg = px.bg.merge(clr);
                    }
                }

                let y = y.min(2);
                let text_continue = "Continue";
                let text_exit = "Exit";
                let x1 = w.min(text_continue.len());
                let x2 = w.min(text_exit.len());
                for (i, (row, (text, x))) in runner
                    .screen_buf
                    .iter_mut()
                    .skip((h - y) / 2)
                    .zip([(text_continue, x1), (text_exit, x2)])
                    .enumerate()
                {
                    let clr = if i == runner.part.paused_selected_option {
                        CLR_WHITE
                    } else {
                        Color::rgb(200, 200, 200)
                    };
                    for (px, char) in row
                        .iter_mut()
                        .skip((w - x) / 2)
                        .zip(text.chars().take((x as f32 * p).round() as _))
                    {
                        px.fg = clr;
                        px.char = char;
                    }
                }
            }
        }

        // instructor
        if !runner.state.config.instructor_disabled() && elapsed_time < 5.0 {
            let a = if elapsed_time < 0.5 {
                (elapsed_time / 0.5 * 255.0).round().clamp(0.0, 255.0) as u8
            } else if elapsed_time > 4.0 {
                ((5.0 - elapsed_time) * 255.0).round().clamp(0.0, 255.0) as u8
            } else {
                255
            };
            let clr = Color::rgba(100, 100, 100, a);
            let lines = [
                "The game can be paused by pressing Esc or P. (try it!)".to_owned(),
                "Use WASD or the arrow keys to move and Spacebar to shoot.".to_owned(),
                "You can also hold down left mouse button and the spaceship will move towards your mouse cursor.".to_owned(),
                format!(
                    "Your spaceship's health is {}, but colliding with an obstacle will decrease it.",
                    runner.part.players_health
                ),
                match runner.part.obstacle_spawn_mode {
                    ObstacleSpawnMode::Waveless =>  "You will be awarded points for surviving/destroying obstacles.".to_owned(),
                    ObstacleSpawnMode::Lastenheft { .. } => "You will be awarded points for surviving/destroying a wave.".to_owned(),
                },
            ];
            for (i, line) in lines.into_iter().rev().enumerate() {
                let y = runner.screen_buf.len().saturating_sub(i + 1);
                if let Some(row) = runner.screen_buf.get_mut(y) {
                    let skip = row.len().saturating_sub(line.len()) / 2;
                    for (px, ch) in row.iter_mut().skip(skip).zip(line.chars()) {
                        px.fg = px.fg.merge(clr);
                        px.char = ch;
                    }
                }
            }
        }

        if runner.part.players_health == 0 && runner.part.spawned_obstacles.is_empty() {
            Self::lose(runner);
        }
    }

    fn event(runner: &mut PartRunner<Self>, e: Event) {
        if super::event(runner, &e) {
            return;
        }
        match e {
            // press esc to exit after losing
            Event::Key(KeyEvent {
                code: KeyCode::Esc,
                kind: KeyEventKind::Press,
                ..
            }) if runner.part.players_health == 0 => {
                // animation "done" -> exit now
                runner.part.spawned_obstacles.clear();
                runner.part.shots.clear();
            }
            // pause screen things
            Event::Key(KeyEvent {
                code: KeyCode::Esc | KeyCode::Char('p' | 'P'),
                kind: KeyEventKind::Press,
                ..
            }) => {
                runner.part.toggle_pause();
            }
            Event::Key(KeyEvent {
                code: KeyCode::Enter,
                kind: KeyEventKind::Press,
                ..
            }) => {
                if runner.part.paused.is_some() {
                    match runner.part.paused_selected_option {
                        0 => (),
                        1 => {
                            // make obstacles not move even after unpausing
                            for o in &mut runner.part.spawned_obstacles {
                                o.speed = 0.0;
                            }
                            // lose the game
                            runner.part.players_health = 0;
                        }
                        _ => (),
                    }
                    runner.part.toggle_pause();
                }
            }
            // movement keys
            Event::Key(KeyEvent {
                code: KeyCode::Up | KeyCode::Char('w' | 'W'),
                modifiers: _,
                kind,
                state: _,
            }) => {
                if runner.part.players_health > 0 {
                    if matches!(kind, KeyEventKind::Press) {
                        runner.part.paused_selected_option =
                            runner.part.paused_selected_option.saturating_sub(1);
                    };
                    kind2bool(
                        kind,
                        &mut runner.part.movement_keys.0,
                        &mut runner.part.movement_keys.1,
                        &mut runner.part.left_mouse_down,
                        &runner.state,
                        &runner.settings,
                    )
                }
            }
            Event::Key(KeyEvent {
                code: KeyCode::Down | KeyCode::Char('s' | 'S'),
                modifiers: _,
                kind,
                state: _,
            }) => {
                if runner.part.players_health > 0 {
                    if matches!(kind, KeyEventKind::Press) {
                        runner.part.paused_selected_option =
                            (runner.part.paused_selected_option + 1).min(1);
                    };
                    kind2bool(
                        kind,
                        &mut runner.part.movement_keys.1,
                        &mut runner.part.movement_keys.0,
                        &mut runner.part.left_mouse_down,
                        &runner.state,
                        &runner.settings,
                    )
                }
            }
            Event::Key(KeyEvent {
                code: KeyCode::Right | KeyCode::Char('d' | 'D'),
                modifiers: _,
                kind,
                state: _,
            }) => {
                if runner.part.players_health > 0 {
                    kind2bool(
                        kind,
                        &mut runner.part.movement_keys.2,
                        &mut runner.part.movement_keys.3,
                        &mut runner.part.left_mouse_down,
                        &runner.state,
                        &runner.settings,
                    )
                }
            }
            Event::Key(KeyEvent {
                code: KeyCode::Left | KeyCode::Char('a' | 'A'),
                modifiers: _,
                kind,
                state: _,
            }) => {
                if runner.part.players_health > 0 {
                    kind2bool(
                        kind,
                        &mut runner.part.movement_keys.3,
                        &mut runner.part.movement_keys.2,
                        &mut runner.part.left_mouse_down,
                        &runner.state,
                        &runner.settings,
                    )
                }
            }
            Event::Key(KeyEvent {
                code: KeyCode::Char(' '),
                modifiers: _,
                kind,
                state: _,
            }) => {
                if runner.part.players_health > 0 {
                    runner.part.shooting = match kind {
                        KeyEventKind::Press | KeyEventKind::Repeat => true,
                        KeyEventKind::Release => false,
                    };
                }
            }
            Event::Key(KeyEvent {
                code: KeyCode::Char('1'),
                modifiers: _,
                kind: KeyEventKind::Press,
                state: _,
            }) => {
                if runner.part.players_health > 0 {
                    // reset camera
                    let now = Instant::now();
                    runner.part.cam_rot.0.set_target(0.0, now);
                    runner
                        .part
                        .cam_rot
                        .1
                        .set_target(std::f64::consts::FRAC_PI_2, now);
                    runner.part.cam_rot.2.set_target(0.0, now);
                }
            }
            Event::Key(KeyEvent {
                code: KeyCode::Char('2'),
                modifiers: _,
                kind: KeyEventKind::Press,
                state: _,
            }) => {
                if runner.part.players_health > 0 {
                    // trackmania reference; switch to cam2
                    let now = Instant::now();
                    runner.part.cam_rot.0.set_target(0.0, now);
                    runner
                        .part
                        .cam_rot
                        .1
                        .set_target(std::f64::consts::FRAC_PI_6, now);
                    runner.part.cam_rot.2.set_target(0.0, now);
                }
            }
            Event::Mouse(MouseEvent {
                kind,
                column,
                row,
                modifiers: _,
            }) => {
                if runner.part.players_health > 0
                    && matches!(kind, MouseEventKind::Drag(MouseButton::Right))
                    && runner.screen_buf.first().is_some_and(|r| !r.is_empty())
                {
                    // trackmania reference, rmb for freecam
                    let (dx, dy) = (
                        (column as i32 - runner.part.mouse_pos.0 as i32) as f64
                            / runner.screen_buf[0].len() as f64,
                        (row as i32 - runner.part.mouse_pos.1 as i32) as f64
                            / runner.screen_buf.len() as f64,
                    );
                    let now = Instant::now();
                    let cam_rot = (
                        (((runner.part.cam_rot.0.get_value(now) + dx * std::f64::consts::PI)
                            % (2.0 * std::f64::consts::PI)
                            + 3.0 * std::f64::consts::PI)
                            % (2.0 * std::f64::consts::PI))
                            - std::f64::consts::PI,
                        (runner.part.cam_rot.1.get_value(now) + dy * std::f64::consts::FRAC_PI_2)
                            .clamp(std::f64::consts::PI / 30.0, std::f64::consts::FRAC_PI_2),
                    );
                    runner.part.cam_rot.0 =
                        DefaultAnimatorF64Quadratic::new(cam_rot.0, CAM_ROT_ANIMATOR_SPEED);
                    runner.part.cam_rot.1 =
                        DefaultAnimatorF64Quadratic::new(cam_rot.1, CAM_ROT_ANIMATOR_SPEED);
                }
                if runner.part.players_health > 0 {
                    runner.part.mouse_pos = (column, row);
                }

                // lmb in pause menu to select or click on button
                if runner.part.paused.is_some()
                    && matches!(kind, MouseEventKind::Down(MouseButton::Left))
                {
                    if row as usize > runner.screen_buf.len().saturating_sub(2) / 2 {
                        // exit
                        if runner.part.paused_selected_option == 1 {
                            // make obstacles not move even after unpausing
                            for o in &mut runner.part.spawned_obstacles {
                                o.speed = 0.0;
                            }
                            // lose the game
                            runner.part.players_health = 0;
                            runner.part.toggle_pause();
                        } else {
                            runner.part.paused_selected_option = 1;
                        }
                    } else {
                        // continue
                        if runner.part.paused_selected_option == 0 {
                            // unpause
                            runner.part.toggle_pause();
                        } else {
                            runner.part.paused_selected_option = 0;
                        }
                    }
                }

                if runner.part.players_health > 0 {
                    // lmb drag -> mouse control of spaceship
                    if matches!(kind, MouseEventKind::Down(MouseButton::Left)) {
                        runner.part.left_mouse_down = true;
                    }
                    if matches!(kind, MouseEventKind::Up(MouseButton::Left)) {
                        runner.part.left_mouse_down = false;
                    }
                }
            }
            _ => (),
        }
        fn kind2bool(
            kind: KeyEventKind,
            dir: &mut bool,
            opposite: &mut bool,
            mouse: &mut bool,
            state: &GlobalState,
            settings: &Settings,
        ) {
            if matches!(kind, KeyEventKind::Press) {
                *mouse = false;
            }
            match (
                state.config.controls_key_mode(),
                settings.with_keyboard_enhancement,
            ) {
                (ControlsKeyMode::Hold, _)
                | (ControlsKeyMode::ToggleIfUnsupportedTerminal, true) => match kind {
                    KeyEventKind::Press | KeyEventKind::Repeat => *dir = true,
                    KeyEventKind::Release => *dir = false,
                },
                (ControlsKeyMode::Toggle, _)
                | (ControlsKeyMode::ToggleIfUnsupportedTerminal, false) => match kind {
                    KeyEventKind::Press | KeyEventKind::Repeat => {
                        if *dir && !*opposite {
                            // already moving in that direction, stop moving
                            *dir = false;
                        } else if *opposite && !*dir {
                            // moving in the opposite direction, stop moving
                            *opposite = false;
                        } else {
                            // not moving, start moving
                            *dir = true;
                            *opposite = false;
                        }
                    }
                    KeyEventKind::Release => (),
                },
            }
        }
    }
}

impl Display for ObstacleSpawnMode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Waveless => write!(f, "waveless"),
            Self::Lastenheft { wave } => write!(
                f,
                "spec-compliant, {wave} wave{}",
                if *wave == 1 { "" } else { "s" }
            ),
        }
    }
}

impl Obstacle {
    fn new_waveless(
        right: f64,
        up: f64,
        width: f64,
        height: f64,
        speed: f64,
        object: Object,
    ) -> Self {
        Self::new_of_wave(right, up, width, height, speed, object, 0)
    }
    fn new_of_wave(
        right: f64,
        up: f64,
        width: f64,
        height: f64,
        speed: f64,
        object: Object,
        wave: usize,
    ) -> Self {
        Self::new_internal(
            right, up, -0.5, 0.5, width, height, speed, object, wave, 0.33, 1.0,
        )
    }
    // a shot with dimensions notmuchxnotmuchxheight and negative speed (so it moves up)
    fn new_shot(
        right: f64,
        up: f64,
        y: f32,
        height: f64,
        speed: f64,
        object: Object,
        pixel_width_in_3d: f64,
    ) -> Self {
        // how many seconds does it take this shot to travel its own length?
        // -> it should take this long to spawn/despawn for a nice animation
        let t = height / speed;
        Self::new_internal(
            right,
            up,
            y,
            y,
            (pixel_width_in_3d * 1.6).max(0.03),
            height,
            -speed,
            object,
            0,
            t,
            t,
        )
    }
    #[allow(clippy::too_many_arguments)]
    fn new_internal(
        right: f64,
        up: f64,
        ymin: f32,
        ymax: f32,
        width: f64,
        height: f64,
        speed: f64,
        object: Object,
        wave: usize,
        spawn_time: f64,
        destroy_time: f64,
    ) -> Self {
        let mut world = SWorld {
            objects: vec![SObject {
                name: "Obstacle".to_owned(),
                import: vec![],
                parts: vec![],
            }],
            order_by: SortBy::Order,
            order: vec![SObjRef {
                obj: Some("Obstacle".to_owned()),
                id: Some("Obstacle".to_owned()),
                condition: None,
                min: Some((-0.5, ymin, -0.5)),
                max: Some((0.5, ymax, 0.5)),
                rotation: None,
            }],
            conditions: vec![],
        }
        .optimize(&mut Assets::new_empty());
        let obstacle_in_world = world.find_object("Obstacle").unwrap();
        *world.object_mut(obstacle_in_world) = object;
        let elem_in_world = world.find_in_order("Obstacle").unwrap();
        let elem = world.elem_in_order_mut(elem_in_world);
        let height_half = height / 2.0;
        let width_half = width / 2.0;
        let min = ((up - height_half) as _, ymin, (right - width_half) as _);
        let max = ((up + height_half) as _, ymax, (right + width_half) as _);
        elem.min = min;
        elem.max = max;
        Self {
            right,
            up,
            ymin,
            ymax,
            height_half,
            width_half,
            speed,
            world,
            elem_in_world,
            destroyed: -2.0,
            wave,
            spawn_time,
            destroy_time,
        }
    }
    fn world(&self) -> &World {
        &self.world
    }
    fn move_down(&mut self, delta_time: f64) {
        let elem = self.world.elem_in_order_mut(self.elem_in_world);
        self.up -= self.speed * delta_time;
        if self.destroyed != -1.0 {
            if self.destroyed < -1.0 {
                // spawn animation: increase size from 0 to 1 over 0.33s
                self.destroyed = (self.destroyed + delta_time / self.spawn_time).min(-1.0);
                let p = self.destroyed + 2.0;
                elem.min.0 = (self.up - self.height_half * p) as _;
                elem.max.0 = (self.up + self.height_half * p) as _;
                elem.min.1 = self.ymin;
                elem.max.1 = self.ymin + (self.ymax - self.ymin) * p as f32;
                elem.min.2 = (self.right - self.width_half * p) as _;
                elem.max.2 = (self.right + self.width_half * p) as _;
            } else {
                // destroy animation: shrink to size 0x0x0
                self.destroyed += delta_time / self.destroy_time;
                let p = 1.0 - self.destroyed.min(1.0);
                let p = 3.0 * p * p - 2.0 * p * p * p;
                elem.min.0 = (self.up - self.height_half * p) as _;
                elem.max.0 = (self.up + self.height_half * p) as _;
                elem.min.1 = self.ymin;
                elem.max.1 = self.ymin + (self.ymax - self.ymin) * p as f32;
                elem.min.2 = (self.right - self.width_half * p) as _;
                elem.max.2 = (self.right + self.width_half * p) as _;
            }
        } else {
            elem.min.0 = (self.up - self.height_half) as _;
            elem.max.0 = (self.up + self.height_half) as _;
        }
    }
    fn top_edge(&self) -> f64 {
        self.up + self.height_half
    }
    fn bottom_edge(&self) -> f64 {
        self.up - self.height_half
    }
    fn left_edge(&self) -> f64 {
        self.right - self.width_half
    }
    fn right_edge(&self) -> f64 {
        self.right + self.width_half
    }

    fn destroy(&mut self) -> bool {
        if self.destroyed < 0.0 {
            self.destroyed = 0.0;
            true
        } else {
            false
        }
    }
}

#[test]
fn test_shots_do_actually_spawn() {
    use crate::tests::*;
    let check_res = CheckRes {
        size: (80, 30),
        supports_keyboard_enhancement: true,
    };
    let config = default_config();
    GameState::new(&GlobalState::new(&check_res, config), None)
        .test(&check_res)
        .draw()
        .unwrap()
        .tick(Duration::from_secs(5))
        .draw()
        .unwrap()
        .check(|r| assert!(r.part.shots.is_empty(), "there should be no shot yet"))
        .event(Event::Key(KeyEvent {
            code: KeyCode::Char(' '),
            modifiers: KeyModifiers::NONE,
            kind: KeyEventKind::Press,
            state: KeyEventState::empty(),
        }))
        .unwrap()
        .draw()
        .unwrap()
        .tick(Duration::from_millis(100))
        .draw()
        .unwrap()
        .event(Event::Key(KeyEvent {
            code: KeyCode::Char(' '),
            modifiers: KeyModifiers::NONE,
            kind: KeyEventKind::Release,
            state: KeyEventState::empty(),
        }))
        .unwrap()
        .draw()
        .unwrap()
        .check(|r| {
            assert!(
                !r.part.shots.is_empty(),
                "there should be at least one shot now"
            )
        });
}

#[test]
fn test_shots_despawn() {
    use crate::tests::*;
    let check_res = CheckRes {
        size: (80, 30),
        supports_keyboard_enhancement: true,
    };
    let config = default_config();
    GameState::new(&GlobalState::new(&check_res, config), None)
        .test(&check_res)
        .draw()
        .unwrap()
        .tick(Duration::from_secs(5))
        .draw()
        .unwrap()
        .event(Event::Key(KeyEvent {
            code: KeyCode::Char(' '),
            modifiers: KeyModifiers::NONE,
            kind: KeyEventKind::Press,
            state: KeyEventState::empty(),
        }))
        .unwrap()
        .draw()
        .unwrap()
        .tick(Duration::from_millis(100))
        .draw()
        .unwrap()
        .event(Event::Key(KeyEvent {
            code: KeyCode::Char(' '),
            modifiers: KeyModifiers::NONE,
            kind: KeyEventKind::Release,
            state: KeyEventState::empty(),
        }))
        .unwrap()
        .draw()
        .unwrap()
        .check(|r| {
            assert!(
                !r.part.shots.is_empty(),
                "there should be at least one shot now"
            )
        })
        .tick(Duration::from_secs(10))
        .draw()
        .unwrap()
        .tick(Duration::from_secs(2))
        .draw()
        .unwrap()
        .check(|r| {
            assert!(
                r.part.shots.is_empty(),
                "all shots should have despawned, but {} were left",
                r.part.shots.len()
            )
        });
}

mod large_numbers {
    const M: bool = true;
    const L: bool = false;
    pub const LARGE_NUM_0: [[bool; 5]; 9] = [
        [L, M, M, M, L],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [L, M, M, M, L],
    ];
    pub const LARGE_NUM_1: [[bool; 5]; 9] = [
        [L, L, L, M, L],
        [L, L, M, M, L],
        [L, M, L, M, L],
        [M, L, L, M, L],
        [L, L, L, M, L],
        [L, L, L, M, L],
        [L, L, L, M, L],
        [L, L, L, M, L],
        [L, L, L, M, L],
    ];
    pub const LARGE_NUM_2: [[bool; 5]; 9] = [
        [L, M, M, M, L],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [L, L, L, L, M],
        [L, L, L, M, L],
        [L, L, M, L, L],
        [L, M, L, L, L],
        [M, L, L, L, L],
        [M, M, M, M, M],
    ];
    pub const LARGE_NUM_3: [[bool; 5]; 9] = [
        [M, M, M, M, L],
        [L, L, L, L, M],
        [L, L, L, L, M],
        [L, L, L, L, M],
        [M, M, M, M, L],
        [L, L, L, L, M],
        [L, L, L, L, M],
        [L, L, L, L, M],
        [M, M, M, M, L],
    ];
    pub const LARGE_NUM_4: [[bool; 5]; 9] = [
        [L, L, L, L, M],
        [L, L, L, M, L],
        [L, L, M, L, L],
        [L, M, L, L, L],
        [M, L, L, M, L],
        [M, L, L, M, L],
        [M, M, M, M, M],
        [L, L, L, M, L],
        [L, L, L, M, L],
    ];
    pub const LARGE_NUM_5: [[bool; 5]; 9] = [
        [M, M, M, M, M],
        [M, L, L, L, L],
        [M, L, L, L, L],
        [M, M, M, L, L],
        [L, L, L, M, L],
        [L, L, L, L, M],
        [L, L, L, L, M],
        [M, L, L, L, M],
        [L, M, M, M, L],
    ];
    pub const LARGE_NUM_6: [[bool; 5]; 9] = [
        [L, L, M, M, L],
        [L, M, L, L, L],
        [M, L, L, L, L],
        [M, L, L, L, L],
        [M, M, M, M, L],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [L, M, M, M, L],
    ];
    pub const LARGE_NUM_7: [[bool; 5]; 9] = [
        [M, M, M, M, M],
        [L, L, L, L, M],
        [L, L, L, L, M],
        [L, L, L, M, L],
        [L, L, M, L, L],
        [L, L, M, L, L],
        [L, M, L, L, L],
        [M, L, L, L, L],
        [M, L, L, L, L],
    ];
    pub const LARGE_NUM_8: [[bool; 5]; 9] = [
        [L, M, M, M, L],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [L, M, M, M, L],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [L, M, M, M, L],
    ];
    pub const LARGE_NUM_9: [[bool; 5]; 9] = [
        [L, M, M, M, L],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [M, L, L, L, M],
        [L, M, M, M, M],
        [L, L, L, L, M],
        [L, L, L, L, M],
        [M, L, L, L, M],
        [L, M, M, M, L],
    ];
    pub const LARGE_NUM_MINUS: [[bool; 5]; 9] = [
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [M, M, M, M, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
    ];
    pub const LARGE_NUM_DOT: [[bool; 5]; 9] = [
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, L, L, L],
        [L, L, M, L, L],
    ];
}
