use std::time::Duration;

use asciivation_duty_renderer::CheckRes;

use crate::{parts::game_state::GameState, state::GlobalState};

use crate::tests::*;

#[test]
fn spaceship_can_move() {
    let check_res = CheckRes {
        size: (80, 30),
        supports_keyboard_enhancement: true,
    };
    let mut spaceship_pos = (0.0, 0.0);
    let mut spaceship_pos_in_world = (0.0, 0.0, 0.0);
    GameState::new(&GlobalState::new(&check_res, default_config()), None)
        .test(&check_res)
        .tick(Duration::from_secs(5))
        .draw()
        .unwrap()
        .check(|r| {
            spaceship_pos = r.part.spaceship_pos;
            spaceship_pos_in_world = r.part.world.elem_in_order(r.part.spaceship_in_order).min;
        })
        .tick(Duration::from_secs(1))
        .draw()
        .unwrap()
        .check(|r| {
            assert!(
                spaceship_pos == r.part.spaceship_pos
                    && spaceship_pos_in_world
                        == r.part.world.elem_in_order(r.part.spaceship_in_order).min,
                "spaceship pos should not change since no inputs are given"
            )
        })
        .event(Event::Key(KeyEvent {
            code: KeyCode::Up,
            modifiers: KeyModifiers::empty(),
            kind: KeyEventKind::Press,
            state: KeyEventState::empty(),
        }))
        .unwrap()
        .tick(Duration::from_secs(1))
        .draw()
        .unwrap()
        .check(|r| {
            assert!(
                spaceship_pos != r.part.spaceship_pos
                    && spaceship_pos_in_world
                        != r.part.world.elem_in_order(r.part.spaceship_in_order).min,
                "spaceship pos should change since key up was pressed"
            )
        })
        .check(|r| {
            assert!(
                spaceship_pos.0 == r.part.spaceship_pos.0
                    && spaceship_pos_in_world.1
                        == r.part.world.elem_in_order(r.part.spaceship_in_order).min.1
                    && spaceship_pos_in_world.2
                        == r.part.world.elem_in_order(r.part.spaceship_in_order).min.2,
                "spaceship pos should not have changed in 2d's x-direction or 3d's y or z directions"
            )
        })
        .check(|r| {
            assert!(
                r.part.world.elem_in_order(r.part.spaceship_in_order).min.0
                    > spaceship_pos_in_world.0,
                    r#"spaceship pos should have changed in 3d's positive x direction ("up")"#
            )
        });
}

#[test]
fn spaceship_cannot_move_within_the_first_second() {
    let check_res = CheckRes {
        size: (80, 30),
        supports_keyboard_enhancement: true,
    };
    let mut spaceship_pos = (0.0, 0.0);
    let mut spaceship_pos_in_world = (0.0, 0.0, 0.0);
    GameState::new(&GlobalState::new(&check_res, default_config()), None)
        .test(&check_res)
        .draw()
        .unwrap()
        .check(|r| {
            spaceship_pos = r.part.spaceship_pos;
            spaceship_pos_in_world = r.part.world.elem_in_order(r.part.spaceship_in_order).min;
        })
        .draw()
        .unwrap()
        .event(Event::Key(KeyEvent {
            code: KeyCode::Up,
            modifiers: KeyModifiers::empty(),
            kind: KeyEventKind::Press,
            state: KeyEventState::empty(),
        }))
        .unwrap()
        .tick(Duration::from_secs_f32(0.9))
        .draw()
        .unwrap()
        .check(|r| {
            assert!(
                spaceship_pos == r.part.spaceship_pos
                    && spaceship_pos_in_world
                        == r.part.world.elem_in_order(r.part.spaceship_in_order).min,
                "spaceship pos should not change since less than 3 seconds have elapsed since the game screen became vsible"
            )
        });
}

#[test]
fn spaceship_cannot_move_past_border() {
    let check_res = CheckRes {
        size: (80, 30),
        supports_keyboard_enhancement: true,
    };
    let mut test_runner =
        GameState::new(&GlobalState::new(&check_res, default_config()), None).test(&check_res);
    let mut initial_spaceship_pos = (0.0, 0.0);
    test_runner
        .tick(Duration::from_secs(5))
        .draw()
        .unwrap()
        .check(|r| {
            initial_spaceship_pos = r.part.spaceship_pos;
        })
        .tick(Duration::from_secs(1))
        .draw()
        .unwrap()
        .check(|r| {
            assert!(
                initial_spaceship_pos == r.part.spaceship_pos,
                "spaceship pos should not change since no inputs are given"
            )
        })
        .event(Event::Key(KeyEvent {
            code: KeyCode::Up,
            modifiers: KeyModifiers::empty(),
            kind: KeyEventKind::Press,
            state: KeyEventState::empty(),
        }))
        .unwrap();
    let mut prev_spaceship_pos = initial_spaceship_pos;
    let mut stopped_moving = false;
    for _ in 0..100 {
        test_runner.tick(Duration::from_secs(1)).draw().unwrap();
        let new_spaceship_pos = test_runner.part_ref().spaceship_pos;
        assert!(
            new_spaceship_pos.0 == initial_spaceship_pos.0,
            "spaceship should not move sideways"
        );
        assert!(new_spaceship_pos.1 != initial_spaceship_pos.1);
        if prev_spaceship_pos == new_spaceship_pos {
            stopped_moving = true;
            break;
        }
        prev_spaceship_pos = new_spaceship_pos;
    }
    assert!(
        stopped_moving,
        "spaceship should stop moving after some time since the screen's edge must be reached"
    );
}

#[test]
fn spaceship_movement_speed_when_going_diagonal() {
    let check_res = CheckRes {
        size: (80, 30),
        supports_keyboard_enhancement: true,
    };

    let mut test_runner_up =
        GameState::new(&GlobalState::new(&check_res, default_config()), None).test(&check_res);
    let mut test_runner_right =
        GameState::new(&GlobalState::new(&check_res, default_config()), None).test(&check_res);
    let mut test_runner_diagonal =
        GameState::new(&GlobalState::new(&check_res, default_config()), None).test(&check_res);

    type TR<'a> = &'a mut TestRunner<GameState>;
    macro_rules! do_for_all {
        ($e:expr) => {
            #[allow(unused_mut)]
            let mut e = $e;
            e(&mut test_runner_up);
            e(&mut test_runner_right);
            e(&mut test_runner_diagonal);
        };
    }

    do_for_all!(|v: TR| {
        v.tick(Duration::from_secs(5)).draw().unwrap();
    });

    let spaceship_pos_og = test_runner_diagonal.part_ref().spaceship_pos;
    assert!(spaceship_pos_og == test_runner_up.part_ref().spaceship_pos);
    assert!(spaceship_pos_og == test_runner_right.part_ref().spaceship_pos);

    // send keyboard inputs
    test_runner_up
        .event(Event::Key(KeyEvent {
            code: KeyCode::Up,
            modifiers: KeyModifiers::empty(),
            kind: KeyEventKind::Press,
            state: KeyEventState::empty(),
        }))
        .unwrap();
    test_runner_right
        .event(Event::Key(KeyEvent {
            code: KeyCode::Right,
            modifiers: KeyModifiers::empty(),
            kind: KeyEventKind::Press,
            state: KeyEventState::empty(),
        }))
        .unwrap();
    test_runner_diagonal
        .event(Event::Key(KeyEvent {
            code: KeyCode::Up,
            modifiers: KeyModifiers::empty(),
            kind: KeyEventKind::Press,
            state: KeyEventState::empty(),
        }))
        .unwrap()
        .event(Event::Key(KeyEvent {
            code: KeyCode::Right,
            modifiers: KeyModifiers::empty(),
            kind: KeyEventKind::Press,
            state: KeyEventState::empty(),
        }))
        .unwrap();

    do_for_all!(|v: TR| {
        v.tick(Duration::from_millis(100)).draw().unwrap();
    });

    let spaceship_pos_up = test_runner_up.part_ref().spaceship_pos;
    let spaceship_pos_right = test_runner_right.part_ref().spaceship_pos;
    let spaceship_pos_diagonal = test_runner_diagonal.part_ref().spaceship_pos;

    assert!(spaceship_pos_og.0 == spaceship_pos_up.0);
    assert!(spaceship_pos_og.1 != spaceship_pos_up.1);
    assert!(spaceship_pos_og.0 != spaceship_pos_right.0);
    assert!(spaceship_pos_og.1 == spaceship_pos_right.1);

    let spaceship_dist_up = (spaceship_pos_up.1 - spaceship_pos_og.1).abs();
    let spaceship_dist_right = (spaceship_pos_right.0 - spaceship_pos_og.0).abs();
    let spaceship_dist_diagonal = ((spaceship_pos_diagonal.0 - spaceship_pos_og.0)
        * (spaceship_pos_diagonal.0 - spaceship_pos_og.0)
        + (spaceship_pos_diagonal.1 - spaceship_pos_og.1)
            * (spaceship_pos_diagonal.1 - spaceship_pos_og.1))
        .sqrt();

    assert!(
        (spaceship_dist_up - spaceship_dist_right).abs() < 0.01,
        "spaceship movement speed should be the same in both directions"
    );

    assert!(
        (spaceship_dist_diagonal - spaceship_dist_up).abs() < 0.01,
        "spaceship movement speed should be the same even when going diagonally"
    );
}
