#[cfg(test)]
pub mod tests;

use std::time::Instant;

use asciivation_duty_renderer::{
    deps::crossterm::event::{
        Event, KeyCode, KeyEvent, KeyEventKind, MouseButton, MouseEvent, MouseEventKind,
    },
    part::{Part, PartRunner},
    render::{clear_pre_render, render, Color, CLR_WHITE, PX_BLACK},
    world::{Assets, ElemInOrder, ObjectInWorld, SObjRef, SObject, SWorld, SortBy, World},
};
use uianimator::{default_animator_f64_quadratic::DefaultAnimatorF64Quadratic, Animator};

use crate::{
    background,
    parts::{options_screen::OptionsScreen, spaceship_selection::SpaceshipSelectionScreen},
    popups::Popups,
    state::GlobalState,
};

use super::{fov_scale, game_state::GameState, AnyPart};

pub struct MainMenu {
    /// the time when the main menu was created.
    /// used for fade-in and other animations
    initialized: Instant,
    selected_button: Option<usize>,
    buttons: Vec<String>,
    world: World,
    spaceship_in_world: ObjectInWorld,
    spaceship_in_order: ElemInOrder,
    selected_button_float: DefaultAnimatorF64Quadratic,
}

impl MainMenu {
    // new without default because there are multiple other new functions
    #[allow(clippy::new_without_default)]
    pub fn new() -> Self {
        Self::new_internal(None)
    }
    pub fn new_from_spaceship_selection() -> Self {
        Self::new_internal(Some(3))
    }
    pub fn new_from_options_screen() -> Self {
        Self::new_internal(Some(1))
    }
    fn new_internal(selected_button: Option<usize>) -> Self {
        // world with an empty object
        let world = SWorld {
            objects: vec![SObject {
                name: "Spaceship".to_owned(),
                import: vec![],
                parts: vec![],
            }],
            order_by: SortBy::Order,
            order: vec![SObjRef {
                obj: Some("Spaceship".to_owned()),
                id: Some("Spaceship".to_owned()),
                condition: None,
                min: Some((-1.5, -0.5, -0.5)),
                max: Some((-0.5, 0.5, 0.5)),
                rotation: None,
            }],
            conditions: vec![],
        }
        .optimize(&mut Assets::new_empty());
        let spaceship_in_world = world.find_object("Spaceship").unwrap();
        let spaceship_in_order = world.find_in_order("Spaceship").unwrap();
        let buttons = ["Exit", "Options", "Start", "Spaceships"]
            .into_iter()
            .map(|str| str.to_owned())
            .collect::<Vec<_>>();
        let buttons_count = buttons.len();
        Self {
            initialized: Instant::now(),
            selected_button,
            buttons,
            world,
            spaceship_in_world,
            spaceship_in_order,
            selected_button_float: DefaultAnimatorF64Quadratic::new(
                selected_button
                    .map(|v| v as f64)
                    .unwrap_or(buttons_count.saturating_sub(1) as f64 / 2.0),
                2.0,
            ),
        }
    }
}

impl Part for MainMenu {
    type End = Option<AnyPart>;
    type State = GlobalState;
    fn draw(runner: &mut PartRunner<Self>, now: Instant) {
        if runner.draw_reset || runner.state.draw_reset {
            runner.state.draw_reset = false;
            // make star background not transparent
            runner.state.background.set_transparency(0.0);
            // load spaceship
            let spaceship_model = runner
                .state
                .config
                .selected_spaceship()
                .model
                .get(runner.state.spaceship_color)
                .clone();
            *runner.part.world.object_mut(runner.part.spaceship_in_world) = spaceship_model;
        }

        // fill entire screen buffer with black pixels
        clear_pre_render(&mut runner.screen_buf, PX_BLACK);

        let height = runner.screen_buf.len();

        for row in &mut runner.screen_buf[height.saturating_sub(2)..] {
            // clear the last 2 rows (in case the instructor text changes)
            for px in &mut *row {
                px.char = ' ';
            }
        }

        // draw stars in the background
        background::draw(runner, now);

        if !runner.screen_buf.is_empty() {
            let width = runner.screen_buf[0].len();

            // show the spaceship
            let fov_scale = fov_scale(width, height, runner.settings.rel_height);
            let spaceship = runner
                .part
                .world
                .elem_in_order_mut(runner.part.spaceship_in_order);
            let spaceship_down_rel = Self::spaceship_down_rel(now, runner.part.initialized);
            let spaceship_x = Self::spaceship_x(spaceship_down_rel);
            let selected_button_float = runner.part.selected_button_float.get_value(now);
            let spaceship_z = Self::spaceship_z(
                selected_button_float,
                runner.part.buttons.len(),
                width,
                height,
                runner.settings.rel_height,
                fov_scale,
            );
            spaceship.min = (spaceship_x - 0.5, -0.5, spaceship_z - 0.5);
            spaceship.max = (spaceship_x + 0.5, 0.5, spaceship_z + 0.5);
            render(
                &runner.part.world,
                (0.0, 10.0, 0.0),
                (0.0, std::f32::consts::FRAC_PI_2, 0.0),
                runner.settings.rel_height,
                fov_scale,
                1.0,
                &mut runner.screen_buf,
            );

            // show Starfighter Alliance text

            // larger-than-normal text, stored as lines
            let mut logo_text = r#"
 ____  _      _         _____ _  ____ _   _____    ____
/ ___|| |_   / \   _ __|  ___(_)/ ___| |_|_   _|__|  _ \
\___ \| __| / _ \ | '__| |_  | | |  _| '_ \| |/ _ \ |_) |
 ___) | |_ / ___ \| |  |  _| | | |_| | | | | |  __/  _ <
|____/ \__/_/   \_\_|  |_|   |_|\____|_| |_|_|\___|_| \_\

        _    _ _     _    _           ____
       / \  | | |   (_)  / \   _ __  / ___|___
      / _ \ | | |   | | / _ \ | '_ \| |   / _ \
     / ___ \| | |___| |/ ___ \| | | | |__|  __/
    /_/   \_\_|_____|_/_/   \_\_| |_|\____\___|  /s
"#
            .trim_matches(['\n', '\r'])
            .lines()
            .collect::<Vec<_>>();

            let mut logo_width = logo_text
                .iter()
                .map(|line| line.chars().count())
                .max()
                .unwrap_or(0);

            // the lines in which the logo may be drawn, max being exclusive
            let min_height = 1;
            let max_height = height - height.div_ceil(2);
            let avail_height = max_height - min_height;

            // if the logo would fill most of the available space, use smaller text instead
            if logo_width / 2 > width / 3 || logo_text.len() / 2 > avail_height / 3 {
                logo_text = vec!["Starfighter", "  Alliance"];
                logo_width = logo_text
                    .iter()
                    .map(|line| line.chars().count())
                    .max()
                    .unwrap_or(0);
            }

            // used to center the logo horizontally
            let padding = width.saturating_sub(logo_width) / 2;
            for (y, (row, line)) in runner
                .screen_buf
                .iter_mut()
                // center the logo vertically
                .skip(min_height + avail_height.saturating_sub(logo_text.len()) / 2)
                // don't draw too far (logo is less important than other ui elements, as it is only a nice-to-have)
                .take(avail_height)
                .zip(logo_text.iter())
                .enumerate()
            {
                for (x, (px, char)) in row.iter_mut().skip(padding).zip(line.chars()).enumerate() {
                    // color the text yellow-white-ish, calculate color
                    // for each char to create a glint effect over the whole text
                    let elapsed_seconds = (now - runner.part.initialized).as_secs_f64();
                    let x = (x
                        + ((elapsed_seconds / 4.5 % 1.0) * logo_width as f64).round() as usize)
                        % logo_width;
                    let up_to_130 = 130
                        - (130 * (x.abs_diff(logo_width / 2) + y))
                            / (logo_text.len() + logo_width.div_ceil(2));
                    px.fg = Color::rgb(255, 235, 80 + (up_to_130 * up_to_130 / 100) as u8);
                    // fade-in effect: if less than 5 seconds have elapsed, multiply each color channel by a value < 1
                    if elapsed_seconds < 5.0 {
                        let p = 1.0 - elapsed_seconds / 5.0;
                        let p = 1.0 - p * p;
                        px.fg = Color::rgb(
                            (px.fg.r as f64 * p) as u8,
                            (px.fg.g as f64 * p) as u8,
                            (px.fg.b as f64 * p) as u8,
                        );
                    }
                    // show text
                    px.char = char;
                }
            }

            // show instructor texts at bottom of screen
            // define the text that should be displayed
            if !runner.state.config.instructor_disabled() {
                let instructor_texts = [
                    if runner.state.popups.has_popup() {
                        "A popup is open."
                    } else {
                        match runner.part.selected_button {
                            Some(0) => "Press enter or left mouse button to exit the game.",
                            Some(1) => "Press enter or left mouse button to open the game options.",
                            Some(2) => "Press enter or left mouse button to start the game.",
                            Some(3) => {
                                "Press enter or left mouse button to choose a different spaceship."
                            }
                            _ => "Press left/right arrow keys or use the mouse to select a button,",
                        }
                    },
                    if runner.state.popups.has_popup() {
                        "Press esc or click outside the box to close it."
                    } else if runner.part.selected_button.is_some() {
                        "Press esc to deselect the button."
                    } else {
                        "Press esc to exit the game."
                    },
                ];
                let height = runner.screen_buf.len();
                for (row, ascii_text) in instructor_texts.into_iter().enumerate() {
                    if let Some(row) = runner
                        .screen_buf
                        .get_mut(height - instructor_texts.len() + row)
                    {
                        // draw instructor text so that each row is centered horizontally
                        let width = row.len();
                        for (ch, px) in ascii_text
                            .chars()
                            .zip(&mut row[(width.saturating_sub(ascii_text.len())) / 2..])
                        {
                            px.fg = Color::rgb(100, 100, 100);
                            px.char = ch;
                        }
                    }
                }
            }

            // draw buttons
            if let Some(row) = runner.screen_buf.get_mut(height - height / 4) {
                let width = row.len();
                for (i, button_text) in runner.part.buttons.iter().enumerate() {
                    let button_text_len = button_text.len();
                    let left = width * i / runner.part.buttons.len();
                    let right = width * (i + 1) / runner.part.buttons.len();
                    let row = &mut row[left..right];
                    let padding = row.len().saturating_sub(button_text_len) / 2;
                    for (px, char) in row.iter_mut().skip(padding).zip(button_text.chars()) {
                        // white if selected, gray otherwise
                        px.fg = if runner.part.selected_button.is_some_and(|sel| sel == i) {
                            CLR_WHITE
                        } else {
                            Color::rgb(127, 127, 127)
                        };
                        px.char = char;
                    }
                }
            }
        }

        Popups::draw(&mut runner.screen_buf, &mut runner.state, now);
    }
    fn event(runner: &mut PartRunner<Self>, e: Event) {
        if super::event(runner, &e) || runner.state.popups.event(&e) {
            return;
        }
        match e {
            // pressing Esc deselectes a selected button or exits the game if no button is selected
            Event::Key(KeyEvent {
                code: KeyCode::Esc,
                kind: KeyEventKind::Press,
                ..
            }) => {
                if runner.part.selected_button.is_none() {
                    runner.end = Some(None);
                } else {
                    runner.part.set_selected_button(None);
                }
            }
            // pressing Enter triggers the selected button's event, if a button is selected
            Event::Key(KeyEvent {
                code: KeyCode::Enter,
                kind: KeyEventKind::Press,
                ..
            }) => {
                press_selected_button(runner);
            }
            // right arrow key pressed -> select button
            Event::Key(KeyEvent {
                code: KeyCode::Right,
                kind: KeyEventKind::Press,
                ..
            }) => {
                runner.part.set_selected_button(Some(
                    runner
                        .part
                        .selected_button
                        .unwrap_or((runner.part.buttons.len() - 1) / 2)
                        .saturating_add(1)
                        .min(runner.part.buttons.len().saturating_sub(1)),
                ));
            }
            // left arrow key pressed -> select button
            Event::Key(KeyEvent {
                code: KeyCode::Left,
                kind: KeyEventKind::Press,
                ..
            }) => {
                runner.part.set_selected_button(Some(
                    runner
                        .part
                        .selected_button
                        .unwrap_or((runner.part.buttons.len() - 1).div_ceil(2))
                        .saturating_sub(1)
                        .max(0),
                ));
            }
            // any mouse event -> select or deselect button
            Event::Mouse(MouseEvent {
                kind,
                column: x,
                row: y,
                ..
            }) => {
                runner.part.set_selected_button(
                    if y as usize > runner.screen_buf.len() / 2 && !runner.screen_buf.is_empty() {
                        Some(
                            (x as usize * runner.part.buttons.len() / runner.screen_buf[0].len())
                                .min(runner.part.buttons.len().saturating_sub(1)),
                        )
                    } else {
                        None
                    },
                );
                if let MouseEventKind::Up(MouseButton::Left) = kind {
                    press_selected_button(runner)
                }
            }
            _ => (),
        }
        fn press_selected_button(runner: &mut PartRunner<MainMenu>) {
            if let Some(btn) = runner.part.selected_button {
                'end: {
                    runner.end = Some(match btn {
                        0 => None,
                        1 => Some(AnyPart::OptionsScreen(OptionsScreen::new(
                            &runner.state,
                            &runner.settings,
                        ))),
                        2 => {
                            let mut selected_button_float = std::mem::replace(
                                &mut runner.part.selected_button_float,
                                DefaultAnimatorF64Quadratic::new(0.0, 1.0),
                            );
                            let buttons = runner.part.buttons.len();
                            selected_button_float
                                .set_target((buttons - 1) as f64 / 2.0, Instant::now());
                            let initialized = runner.part.initialized;
                            Some(AnyPart::GameState(GameState::new(
                                &runner.state,
                                Some(Box::new(
                                    move |_spaceship,
                                          progress,
                                          now,
                                          target_x,
                                          _target_z,
                                          width,
                                          height,
                                          rel_height,
                                          fov_scale| {
                                        let p = (progress * 4.0).min(1.0);
                                        let p = (3.0 * p * p - 2.0 * p * p * p) as f32;
                                        (
                                            ((MainMenu::spaceship_x(
                                                MainMenu::spaceship_down_rel(now, initialized),
                                            ) - 0.5)
                                                * (1.0 - p)
                                                + p * target_x),
                                            MainMenu::spaceship_z(
                                                selected_button_float.get_value(now),
                                                buttons,
                                                width,
                                                height,
                                                rel_height,
                                                fov_scale,
                                            ) - 0.5,
                                        )
                                    },
                                )),
                            )))
                        }
                        3 => Some(AnyPart::SpaceshipSelectionScreen(
                            SpaceshipSelectionScreen::new(),
                        )),
                        _ => break 'end,
                    });
                }
            }
        }
        if runner.end.as_ref().is_some_and(|end| end.is_none()) {
            // exiting
            runner
                .state
                .config
                .set_rel_height(runner.settings.rel_height);
            runner.state.config.save(&mut runner.state.popups);
            runner
                .state
                .highscores
                .save(&runner.state.config, &mut runner.state.popups);
            if runner.state.popups.has_popup() {
                runner.end = None;
            }
        }
    }
}

impl MainMenu {
    fn set_selected_button(&mut self, btn: Option<usize>) {
        if let Some(btn) = btn {
            self.selected_button = Some(btn);
            self.selected_button_float
                .set_target(btn as _, Instant::now());
        } else {
            self.selected_button = None;
            self.selected_button_float.set_target(
                self.buttons.len().saturating_sub(1) as f64 / 2.0,
                Instant::now(),
            );
        }
    }

    fn spaceship_down_rel(now: Instant, initialized: Instant) -> f64 {
        1.0 - ((now - initialized).as_secs_f64() / 3.0).min(1.0)
    }
    fn spaceship_x(spaceship_down_rel: f64) -> f32 {
        (-0.5 - 3.0 * spaceship_down_rel * spaceship_down_rel * spaceship_down_rel) as f32
    }
    fn spaceship_z(
        selected_button_float: f64,
        buttons: usize,
        width: usize,
        height: usize,
        rel_height: f32,
        fov_scale: f32,
    ) -> f32 {
        asciivation_duty_renderer::render::pos_from_2d(
            (
                (selected_button_float as f32 + 0.5) * width as f32 / buttons as f32,
                0.5,
            ),
            (width as _, height as _),
            rel_height,
            fov_scale,
        )
        .0 * 10.5
    }
}
