use std::{fmt::Debug, time::Duration};

use asciivation_duty_renderer::{
    deps::crossterm::event::{Event, KeyCode, KeyEvent, KeyEventKind},
    part::{Part, PartRunner},
};
use end_screen::EndScreen;
use game_state::GameState;
use main_menu::MainMenu;
use options_screen::OptionsScreen;
use spaceship_selection::SpaceshipSelectionScreen;

use crate::state::GlobalState;

pub mod end_screen;
pub mod game_state;
pub mod main_menu;
pub mod options_screen;
pub mod spaceship_selection;

#[allow(clippy::large_enum_variant)]
pub enum AnyPart {
    MainMenu(MainMenu),
    GameState(GameState),
    EndScreen(EndScreen),
    SpaceshipSelectionScreen(SpaceshipSelectionScreen),
    OptionsScreen(OptionsScreen),
}
#[allow(clippy::large_enum_variant)]
pub enum AnyPartRunner {
    MainMenu(PartRunner<MainMenu>),
    GameState(PartRunner<GameState>),
    EndScreen(PartRunner<EndScreen>),
    SpaceshipSelectionScreen(PartRunner<SpaceshipSelectionScreen>),
    OptionsScreen(PartRunner<OptionsScreen>),
}
impl AnyPartRunner {
    pub fn new<P: Part<State = GlobalState>>(part: AnyPart, runner: PartRunner<P>) -> Self {
        match part {
            AnyPart::MainMenu(part) => Self::MainMenu(runner.to(part)),
            AnyPart::GameState(part) => Self::GameState(runner.to(part)),
            AnyPart::EndScreen(part) => Self::EndScreen(runner.to(part)),
            AnyPart::SpaceshipSelectionScreen(part) => {
                Self::SpaceshipSelectionScreen(runner.to(part))
            }
            AnyPart::OptionsScreen(part) => Self::OptionsScreen(runner.to(part)),
        }
    }
}

pub(crate) fn event<P: Part>(runner: &mut PartRunner<P>, e: &Event) -> bool {
    match e {
        Event::Key(KeyEvent {
            code: KeyCode::F(5),
            modifiers: _,
            kind: KeyEventKind::Press,
            state: _,
        }) => runner.settings.frame_time = Duration::from_secs_f64(1.0 / 30.0),
        Event::Key(KeyEvent {
            code: KeyCode::F(6),
            modifiers: _,
            kind: KeyEventKind::Press,
            state: _,
        }) => runner.settings.frame_time = Duration::from_secs_f64(1.0 / 60.0),
        Event::Key(KeyEvent {
            code: KeyCode::F(7),
            modifiers: _,
            kind: KeyEventKind::Press,
            state: _,
        }) => runner.settings.frame_time = Duration::from_secs_f64(1.0 / 120.0),
        Event::Key(KeyEvent {
            code: KeyCode::F(8),
            modifiers: _,
            kind: KeyEventKind::Press,
            state: _,
        }) => runner.settings.frame_time = Duration::ZERO,
        Event::Key(KeyEvent {
            code: KeyCode::F(9),
            modifiers: _,
            kind: KeyEventKind::Press,
            state: _,
        }) => runner.settings.debug ^= 1,
        Event::Key(KeyEvent {
            code: KeyCode::F(10),
            modifiers: _,
            kind: KeyEventKind::Press,
            state: _,
        }) => runner.settings.debug ^= 2,
        Event::Key(KeyEvent {
            code: KeyCode::F(11),
            modifiers: _,
            kind: KeyEventKind::Press | KeyEventKind::Repeat,
            state: _,
        }) => runner.settings.rel_height = (runner.settings.rel_height - 0.1).max(0.1),
        Event::Key(KeyEvent {
            code: KeyCode::F(12),
            modifiers: _,
            kind: KeyEventKind::Press | KeyEventKind::Repeat,
            state: _,
        }) => runner.settings.rel_height += 0.1,
        _ => return false,
    }
    true
}

pub(crate) fn fov_scale(width: usize, height: usize, rel_height: f32) -> f32 {
    1.2 * (width as f32 / (height as f32 * rel_height)).min(4.0 / 3.0)
}

impl Debug for AnyPart {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::MainMenu(_) => write!(f, "AnyPart::MainMenu"),
            Self::GameState(_) => write!(f, "AnyPart::GameState"),
            Self::EndScreen(_) => write!(f, "AnyPart::EndScreen"),
            Self::SpaceshipSelectionScreen(_) => write!(f, "AnyPart::SpaceshipSelectionScreen"),
            Self::OptionsScreen(_) => write!(f, "AnyPart::OptionsScreen"),
        }
    }
}
