use std::{
    path::PathBuf,
    sync::Arc,
    time::{Duration, Instant},
};

use asciivation_duty_renderer::{
    deps::crossterm::event::{
        Event, KeyCode, KeyEvent, KeyEventKind, KeyModifiers, MouseEvent, MouseEventKind,
    },
    part::{Part, PartRunner, Settings},
    render::{clear_pre_render, Color, Pixel, CLR_WHITE, PX_BLACK},
};
use uianimator::{default_animator_f64_quadratic::DefaultAnimatorF64Quadratic, Animator};

use crate::{
    background,
    config::Config,
    config::{ControlsKeyMode, DamageEffect},
    popups::Popups,
    state::GlobalState,
    websocket::WsColor,
};

use super::main_menu::MainMenu;

pub struct OptionsScreen {
    options: Vec<OptionType>,
    selected_option: usize,
    scroll: usize,
    scrollf: DefaultAnimatorF64Quadratic,
}
pub enum OptionType {
    DropdownMenu(DropdownMenu),
    FloatSlider(FloatSlider),
}
pub struct DropdownMenu {
    text: String,
    options: Vec<DropdownOption>,
    selected_option: usize,
    expanded: bool,
}
pub struct FloatSlider {
    text: String,
    min: f32,
    max: f32,
    value: f32,
    #[allow(clippy::type_complexity)]
    on_change: Arc<dyn Fn(&mut FloatSlider, &mut GlobalState, &mut Settings)>,
}
pub struct DropdownOption {
    text: String,
    #[allow(clippy::type_complexity)]
    on_select:
        Arc<dyn Fn(&mut DropdownMenu, &mut GlobalState, &mut Settings, &mut Option<MainMenu>)>,
}

impl OptionsScreen {
    pub fn new(state: &GlobalState, settings: &Settings) -> Self {
        let mut s = Self {
            options: vec![],
            selected_option: 0,
            scroll: 0,
            scrollf: DefaultAnimatorF64Quadratic::new(0.0, 30.0),
        };
        s.reset_options(state, settings);
        s
    }

    const OPT_REL_HEIGHT: usize = 2;

    fn reset_options(&mut self, state: &GlobalState, settings: &Settings) {
        let colors = [
            WsColor::Red,
            WsColor::Blue,
            WsColor::Yellow,
            WsColor::Orange,
            WsColor::Green,
            WsColor::Violet,
            WsColor::White,
            WsColor::Black,
            WsColor::Transparent,
        ];
        let selected_color = colors
            .iter()
            .position(|color| *color == state.spaceship_color);
        self.options = vec![
            OptionType::DropdownMenu(DropdownMenu {
                text: "back to main menu".to_owned(),
                options: vec![
                    DropdownOption {
                        text: "and save config (esc)".to_owned(),
                        on_select: Arc::new(|_, _, _, end| {
                            *end = Some(MainMenu::new_from_options_screen());
                        }),
                    },
                    DropdownOption {
                        text: "nevermind, stay here".to_owned(),
                        on_select: Arc::new(|_, _, _, _| {}),
                    },
                ],
                selected_option: 0,
                expanded: false,
            }),
            OptionType::DropdownMenu(DropdownMenu {
                text: "Controls: key mode".to_owned(),
                options: vec![
                    DropdownOption {
                        text: "hold to move".to_owned(),
                        on_select: Arc::new(|_, state, _, _| {
                            state.config.set_controls_key_mode(ControlsKeyMode::Hold)
                        }),
                    },
                    DropdownOption {
                        text: "toggle if unsupported terminal, hold otherwise".to_owned(),
                        on_select: Arc::new(|_, state, _, _| {
                            state
                                .config
                                .set_controls_key_mode(ControlsKeyMode::ToggleIfUnsupportedTerminal)
                        }),
                    },
                    DropdownOption {
                        text: "toggle moving".to_owned(),
                        on_select: Arc::new(|_, state, _, _| {
                            state.config.set_controls_key_mode(ControlsKeyMode::Toggle)
                        }),
                    },
                ],
                selected_option: match state.config.controls_key_mode() {
                    ControlsKeyMode::Hold => 0,
                    ControlsKeyMode::ToggleIfUnsupportedTerminal => 1,
                    ControlsKeyMode::Toggle => 2,
                },
                expanded: false,
            }),
            OptionType::FloatSlider(FloatSlider {
                text: format!("Char Height ({:.2})", settings.rel_height),
                min: 0.1,
                max: 3.0,
                value: settings.rel_height,
                on_change: Arc::new(|slider, _, settings| {
                    settings.rel_height = slider.value;
                    slider.text = format!("Char Height ({:.2})", settings.rel_height)
                }),
            }),
            OptionType::DropdownMenu(DropdownMenu {
                text: "Spaceship Color".to_owned(),
                options: colors
                    .into_iter()
                    .map(|color| DropdownOption {
                        text: format!("{color}"),
                        on_select: Arc::new(move |_, state, _, _| {
                            state.spaceship_color = color;
                            state.spaceship_color_ignore_websocket = true;
                        }),
                    })
                    .chain(
                        [if state.config.websocket().is_some() {
                            Some(DropdownOption {
                                text: "from websocket".to_owned(),
                                on_select: Arc::new(move |_, state, _, _| {
                                    if let Some(address) = state.config.websocket() {
                                        crate::websocket::start_thread(address.to_owned(), state);
                                    }
                                }),
                            })
                        } else {
                            None
                        }]
                        .into_iter()
                        .flatten(),
                    )
                    .collect::<Vec<_>>(),
                selected_option: selected_color.unwrap_or(0),
                expanded: false,
            }),
            OptionType::DropdownMenu(DropdownMenu {
                text: "obstacle spawning".to_owned(),
                options: vec![
                    DropdownOption {
                        text: "waveless".to_owned(),
                        on_select: Arc::new(move |_, state, _, _| {
                            state.better_obstacle_spawning = true;
                        }),
                    },
                    DropdownOption {
                        text: "spec-compliant (lastenheft)".to_owned(),
                        on_select: Arc::new(move |_, state, _, _| {
                            state.better_obstacle_spawning = false;
                        }),
                    },
                ],
                selected_option: if state.better_obstacle_spawning { 0 } else { 1 },
                expanded: false,
            }),
            OptionType::DropdownMenu(DropdownMenu {
                text: "damage effect".to_owned(),
                options: vec![
                    DropdownOption {
                        text: "none".to_owned(),
                        on_select: Arc::new(move |_, state, _, _| {
                            state.config.set_damage_effect(DamageEffect::None);
                        }),
                    },
                    DropdownOption {
                        text: "only flash health text".to_owned(),
                        on_select: Arc::new(move |_, state, _, _| {
                            state.config.set_damage_effect(DamageEffect::TextOnly);
                        }),
                    },
                    DropdownOption {
                        text: "dark flash".to_owned(),
                        on_select: Arc::new(move |_, state, _, _| {
                            state.config.set_damage_effect(DamageEffect::FlashBlack);
                        }),
                    },
                    DropdownOption {
                        text: "red flash".to_owned(),
                        on_select: Arc::new(move |_, state, _, _| {
                            state.config.set_damage_effect(DamageEffect::FlashRed);
                        }),
                    },
                ],
                selected_option: match state.config.damage_effect() {
                    DamageEffect::None => 0,
                    DamageEffect::TextOnly => 1,
                    DamageEffect::FlashBlack => 2,
                    DamageEffect::FlashRed => 3,
                },
                expanded: false,
            }),
            OptionType::DropdownMenu(DropdownMenu {
                text: "instructor".to_owned(),
                options: vec![
                    DropdownOption {
                        text: "shown".to_owned(),
                        on_select: Arc::new(move |_, state, _, _| {
                            state.config.set_instructor_disabled(false);
                        }),
                    },
                    DropdownOption {
                        text: "hidden".to_owned(),
                        on_select: Arc::new(move |_, state, _, _| {
                            state.config.set_instructor_disabled(true);
                        }),
                    },
                ],
                selected_option: if state.config.instructor_disabled() {
                    1
                } else {
                    0
                },
                expanded: false,
            }),
            OptionType::FloatSlider({
                let fps = (1.0 / settings.frame_time.as_secs_f32()).round();
                FloatSlider {
                    text: if fps < 500.5 {
                        format!("FPS limit ({fps:.0})")
                    } else {
                        "FPS limit (none)".to_owned()
                    },
                    min: 5.0,
                    max: 505.0,
                    value: if fps > 500.0 {
                        // display as "no limit"
                        505.0
                    } else {
                        fps
                    },
                    on_change: Arc::new(|slider, state, settings| {
                        let fps = slider.value.round();
                        let frame_time = 1.0 / fps;
                        if fps < 500.5 {
                            settings.frame_time = Duration::from_secs_f32(frame_time);
                            state.config.set_fps_limit(fps);
                            slider.text = format!("FPS limit ({fps:.0})");
                        } else {
                            settings.frame_time = Duration::ZERO;
                            state.config.set_fps_limit(0.0);
                            slider.text = "FPS limit (none)".to_owned();
                        }
                    }),
                }
            }),
            OptionType::DropdownMenu(DropdownMenu {
                text: "debugging overlay".to_owned(),
                options: vec![
                    DropdownOption {
                        text: "hidden".to_owned(),
                        on_select: Arc::new(move |_, _, settings, _| {
                            settings.debug = 0;
                        }),
                    },
                    DropdownOption {
                        text: "fps".to_owned(),
                        on_select: Arc::new(move |_, _, settings, _| {
                            settings.debug = 1;
                        }),
                    },
                    DropdownOption {
                        text: "render- / frametime".to_owned(),
                        on_select: Arc::new(move |_, _, settings, _| {
                            settings.debug = 2;
                        }),
                    },
                    DropdownOption {
                        text: "full (0xFF)".to_owned(),
                        on_select: Arc::new(move |_, _, settings, _| {
                            settings.debug = 0xFF;
                        }),
                    },
                ],
                selected_option: settings.debug.min(3) as _,
                expanded: false,
            }),
            OptionType::DropdownMenu(DropdownMenu {
                text: "selected config file".to_owned(),
                options: {
                    [
                        {
                            let path = state.config.path.clone();
                            DropdownOption {
                                text: format!("keep using {:?}", path),
                                on_select: Arc::new(move |_, state, _, _| {
                                    state.config.path = path.clone();
                                }),
                            }
                        },
                        DropdownOption {
                            text: "–––– no config / don't save ––––".to_owned(),
                            on_select: Arc::new(|_, state, _, _| {
                                state.config.path = PathBuf::new();
                            }),
                        },
                    ]
                    .into_iter()
                    .chain(std::fs::read_dir(".").ok().into_iter().flat_map(|dir| {
                        dir.into_iter().filter_map(|entry| {
                            let path: PathBuf = entry.ok()?.file_name().into();
                            if path.extension().is_some_and(|ext| {
                                ext.to_str().is_some_and(|ext| ext.to_lowercase() == "json")
                            }) {
                                Some(DropdownOption {
                                    text: format!("load {:?}", path),
                                    on_select: Arc::new(move |menu, state, _, _| {
                                        state.highscores.save(&state.config, &mut state.popups);
                                        let new_config = Config::from_json(&path);
                                        if let Ok(new_config) = new_config {
                                            state.config = new_config;
                                            state.highscores = crate::scores::Highscores::try_load(
                                                &state.config,
                                                &mut state.popups,
                                            );
                                            state.draw_reset = true;
                                        } else {
                                            state.config.path = PathBuf::new();
                                            menu.selected_option = 1;
                                        }
                                    }),
                                })
                            } else {
                                None
                            }
                        })
                    }))
                    .collect()
                },
                selected_option: 0,
                expanded: false,
            }),
        ];
    }
}

impl Part for OptionsScreen {
    type End = MainMenu;
    type State = GlobalState;
    fn draw(runner: &mut PartRunner<Self>, now: Instant) {
        if runner.state.draw_reset {
            runner.state.draw_reset = false;
            runner.part.reset_options(&runner.state, &runner.settings);
        }

        if let OptionType::FloatSlider(float_slider) =
            &mut runner.part.options[Self::OPT_REL_HEIGHT]
        {
            if float_slider.value != runner.settings.rel_height {
                float_slider.value = runner.settings.rel_height;
                Arc::clone(&float_slider.on_change)(
                    float_slider,
                    &mut runner.state,
                    &mut runner.settings,
                );
            }
        }

        if runner.draw_reset {
            // make star background not transparent
            runner.state.background.set_transparency(0.0);
        }

        runner
            .part
            .scrollf
            .set_target(runner.part.scroll as f64, now);
        let scroll = runner.part.scrollf.get_value(now).round() as usize;

        // fill entire screen buffer with black pixels
        clear_pre_render(&mut runner.screen_buf, PX_BLACK);

        // draw stars in the background
        background::draw(runner, now);

        if !runner.screen_buf.is_empty() {
            // draw options
            let mut y = 0;
            for (i, opt) in runner.part.options.iter_mut().enumerate() {
                if i == runner.part.selected_option {
                    if let Some(row) = runner.screen_buf.get_mut(y) {
                        for px in row {
                            px.fg = CLR_WHITE;
                            px.char = '–';
                        }
                    }
                }
                let y2 = y + opt.height();
                if i == runner.part.selected_option {
                    if let Some(row) = runner.screen_buf.get_mut(y2) {
                        for px in row {
                            px.fg = CLR_WHITE;
                            px.char = '–';
                        }
                    }
                }
                if y2 >= scroll {
                    let y2 = y2 - scroll;
                    opt.draw(&mut runner.screen_buf, y2);
                }
                y = y2;
            }
        }

        Popups::draw(&mut runner.screen_buf, &mut runner.state, now);
    }
    fn event(runner: &mut PartRunner<Self>, e: Event) {
        if super::event(runner, &e) || runner.state.popups.event(&e) {
            return;
        }
        match e {
            // pressing Esc deselectes a selected button or exits the game if no button is selected
            Event::Key(KeyEvent {
                code: KeyCode::Esc,
                kind: KeyEventKind::Press,
                ..
            }) => {
                runner
                    .state
                    .config
                    .set_rel_height(runner.settings.rel_height);
                runner.state.config.save(&mut runner.state.popups);
                runner.end = Some(MainMenu::new_from_options_screen());
            }
            // pressing Enter allows changing the option
            Event::Key(KeyEvent {
                code: KeyCode::Enter,
                kind: KeyEventKind::Press,
                ..
            }) => runner.part.options[runner.part.selected_option].click(
                None,
                &mut runner.state,
                &mut runner.settings,
                &mut runner.end,
            ),
            // select up/down
            Event::Key(KeyEvent {
                code: KeyCode::Down,
                kind: KeyEventKind::Press,
                ..
            }) => scroll1(&mut runner.part, false, runner.screen_buf.len()),
            Event::Key(KeyEvent {
                code: KeyCode::Up,
                kind: KeyEventKind::Press,
                ..
            }) => scroll1(&mut runner.part, true, runner.screen_buf.len()),
            // move slider on left/right arrow pressed
            Event::Key(KeyEvent {
                code,
                kind: KeyEventKind::Press | KeyEventKind::Repeat,
                modifiers,
                ..
            }) if matches!(code, KeyCode::Right | KeyCode::Left) => {
                // set float slider value if float slider is selected
                if let OptionType::FloatSlider(float_slider) =
                    &mut runner.part.options[runner.part.selected_option]
                {
                    let diff = float_slider.max - float_slider.min;
                    let mut step = if modifiers & KeyModifiers::SHIFT == KeyModifiers::SHIFT {
                        diff / 10.0
                    } else {
                        diff / 100.0
                    };
                    if matches!(code, KeyCode::Left) {
                        step *= -1.0;
                    }
                    float_slider.value = (float_slider.value + step)
                        .max(float_slider.min)
                        .min(float_slider.max);
                    Arc::clone(&float_slider.on_change)(
                        float_slider,
                        &mut runner.state,
                        &mut runner.settings,
                    );
                }
            }
            Event::Mouse(MouseEvent {
                kind: MouseEventKind::Drag(_),
                column: x,
                ..
            }) => {
                // set float slider value if float slider is selected
                if let OptionType::FloatSlider(float_slider) =
                    &mut runner.part.options[runner.part.selected_option]
                {
                    let text_width = float_slider.text.chars().count() + 5;
                    if x as usize >= text_width {
                        let slider_width =
                            runner.screen_buf[0].len().saturating_sub(text_width + 1);
                        if slider_width > 1 {
                            float_slider.value = (float_slider.min
                                + (float_slider.max - float_slider.min)
                                    * (x as usize - text_width) as f32
                                    / (slider_width - 1) as f32)
                                .max(float_slider.min)
                                .min(float_slider.max);
                            Arc::clone(&float_slider.on_change)(
                                float_slider,
                                &mut runner.state,
                                &mut runner.settings,
                            );
                        }
                    }
                }
            }
            // mouse down -> select option
            Event::Mouse(MouseEvent {
                kind: MouseEventKind::Down(_),
                column: x,
                row: y,
                ..
            }) => {
                'select: {
                    // select option
                    let y =
                        y as usize + runner.part.scrollf.get_value(Instant::now()).round() as usize;
                    let mut option_start = 0;
                    for (i, opt) in runner.part.options.iter().enumerate() {
                        let option_end = option_start + opt.height();
                        if option_start <= y && y < option_end {
                            if runner.part.selected_option != i {
                                runner.part.selected_option = i;
                                // skip setting float slider's value, because selected option has changed
                                break 'select;
                            } else {
                                // allow setting float slider's value
                                break;
                            }
                        }
                        option_start = option_end;
                    }

                    // set float slider value if float slider was selected and is still selected
                    if let OptionType::FloatSlider(float_slider) =
                        &mut runner.part.options[runner.part.selected_option]
                    {
                        let text_width = float_slider.text.chars().count() + 5;
                        if x as usize >= text_width {
                            let slider_width =
                                runner.screen_buf[0].len().saturating_sub(text_width + 1);
                            if slider_width > 1 {
                                float_slider.value = (float_slider.min
                                    + (float_slider.max - float_slider.min)
                                        * (x as usize - text_width) as f32
                                        / (slider_width - 1) as f32)
                                    .max(float_slider.min)
                                    .min(float_slider.max);
                                Arc::clone(&float_slider.on_change)(
                                    float_slider,
                                    &mut runner.state,
                                    &mut runner.settings,
                                );
                            }
                        }
                    }

                    // click event
                    let y = y + runner.part.scrollf.get_value(Instant::now()).round() as usize;
                    let mut option_start = 0;
                    for (i, opt) in runner.part.options.iter().enumerate() {
                        let option_end = option_start + opt.height();
                        if option_start < y && y < option_end {
                            if runner.part.selected_option == i {
                                runner.part.options[i].click(
                                    Some((x as usize, y - option_start - 1)),
                                    &mut runner.state,
                                    &mut runner.settings,
                                    &mut runner.end,
                                );
                            }
                            break;
                        }
                        option_start = option_end;
                    }
                }
            }

            _ => (),
        }
        fn scroll1(s: &mut OptionsScreen, up: bool, height: usize) {
            'scroll: {
                match &mut s.options[s.selected_option] {
                    OptionType::DropdownMenu(dropdown_menu) => {
                        if dropdown_menu.expanded {
                            if up && dropdown_menu.selected_option > 0 {
                                dropdown_menu.selected_option -= 1;
                            } else if !up
                                && dropdown_menu.selected_option + 1 < dropdown_menu.options.len()
                            {
                                dropdown_menu.selected_option += 1;
                            }
                            break 'scroll;
                        }
                    }
                    OptionType::FloatSlider(_) => (),
                }
                if up && s.selected_option > 0 {
                    s.selected_option -= 1;
                    let option_top = s
                        .options
                        .iter()
                        .take(s.selected_option)
                        .map(|v| v.height())
                        .sum::<usize>();
                    let scroll = s.scrollf.get_value(Instant::now()).round() as usize;
                    if option_top < scroll {
                        s.scroll = option_top;
                    }
                } else if !up && s.selected_option + 1 < s.options.len() {
                    s.selected_option += 1;
                    let option_bottom = s
                        .options
                        .iter()
                        .take(s.selected_option + 1)
                        .map(|v| v.height())
                        .sum::<usize>();
                    let scroll = s.scrollf.get_value(Instant::now()).round() as usize;
                    if option_bottom > height + scroll {
                        s.scroll = option_bottom - height;
                    }
                }
            }
        }
    }
}

impl OptionType {
    fn height(&self) -> usize {
        1 + match self {
            Self::DropdownMenu(v) => v.height(),
            Self::FloatSlider(v) => v.height(),
        }
    }
    fn click(
        &mut self,
        mouse_pos: Option<(usize, usize)>,
        state: &mut GlobalState,
        settings: &mut Settings,
        end: &mut Option<MainMenu>,
    ) {
        match self {
            Self::DropdownMenu(dropdown_menu) => {
                if dropdown_menu.expanded {
                    if let Some((_, y)) = mouse_pos {
                        dropdown_menu.selected_option = y.min(dropdown_menu.options.len() - 1);
                    }
                    Arc::clone(&dropdown_menu.options[dropdown_menu.selected_option].on_select)(
                        dropdown_menu,
                        state,
                        settings,
                        end,
                    );
                    dropdown_menu.expanded = false;
                } else {
                    dropdown_menu.expanded = true;
                }
            }
            Self::FloatSlider(_) => (),
        }
    }
    fn draw(&mut self, screen_buf: &mut Vec<Vec<Pixel>>, y2: usize) {
        match self {
            Self::DropdownMenu(dropdown_menu) => dropdown_menu.draw(screen_buf, y2),
            Self::FloatSlider(float_slider) => float_slider.draw(screen_buf, y2),
        }
    }
}

impl DropdownMenu {
    fn height(&self) -> usize {
        if self.expanded {
            1.max(self.options.len())
        } else {
            1
        }
    }
    fn draw(&mut self, #[allow(clippy::ptr_arg)] screen_buf: &mut Vec<Vec<Pixel>>, y2: usize) {
        if !self.expanded {
            if let Some(row) = y2
                .checked_sub(self.height())
                .filter(|y| *y > 0)
                .and_then(|y| screen_buf.get_mut(y))
            {
                for (ch, px) in self.text.chars().zip(row.iter_mut()) {
                    px.fg = CLR_WHITE;
                    px.char = ch;
                }
                if let Some(px) = row.get_mut(self.text.chars().count() + 2) {
                    px.fg = Color::rgb(128, 128, 128);
                    px.char = '|';
                }
                for (ch, px) in self.options[self.selected_option]
                    .text
                    .chars()
                    .zip(row.iter_mut().skip(self.text.chars().count() + 5))
                {
                    px.fg = CLR_WHITE;
                    px.char = ch;
                }
            }
        } else {
            let text_width = screen_buf[0]
                .len()
                .saturating_sub(
                    self.options
                        .iter()
                        .map(|v| v.text.chars().count())
                        .max()
                        .unwrap_or(0),
                )
                .min(self.text.chars().count() + 5);
            if let Some(row) = y2
                .checked_sub(self.height())
                .filter(|y| *y > 0)
                .and_then(|y| screen_buf.get_mut(y))
            {
                for (ch, px) in self
                    .text
                    .chars()
                    .zip(row.iter_mut())
                    .take(text_width.saturating_sub(5))
                {
                    px.fg = CLR_WHITE;
                    px.char = ch;
                }
            }
            for (i, opt) in self.options.iter().enumerate() {
                if let Some(row) = y2
                    .checked_sub(self.height() - i)
                    .filter(|y| *y > 0)
                    .and_then(|y| screen_buf.get_mut(y))
                {
                    if let Some(px) = row.get_mut(text_width.saturating_sub(3)) {
                        px.fg = Color::rgb(128, 128, 128);
                        px.char = '|';
                    }
                    for (ch, px) in opt.text.chars().zip(row.iter_mut().skip(text_width)) {
                        px.fg = if i == self.selected_option {
                            CLR_WHITE
                        } else {
                            Color::rgb(192, 192, 192)
                        };
                        px.char = ch;
                    }
                }
            }
        }
    }
}

impl FloatSlider {
    fn height(&self) -> usize {
        1
    }
    fn draw(&mut self, #[allow(clippy::ptr_arg)] screen_buf: &mut Vec<Vec<Pixel>>, y2: usize) {
        let text_width = self.text.chars().count() + 5;
        if let Some(row) = y2
            .checked_sub(self.height())
            .filter(|y| *y > 0)
            .and_then(|y| screen_buf.get_mut(y))
        {
            let slider_width = row.len().saturating_sub(text_width + 1);
            for (ch, px) in self
                .text
                .chars()
                .zip(row.iter_mut())
                .take(text_width.saturating_sub(5))
            {
                px.fg = CLR_WHITE;
                px.char = ch;
            }
            if let Some(px) = row.get_mut(text_width.saturating_sub(3)) {
                px.fg = Color::rgb(128, 128, 128);
                px.char = '|';
            }
            for i in 0..slider_width {
                let as_float_left =
                    self.min + (self.max - self.min) * i as f32 / slider_width as f32;
                let as_float_right =
                    self.min + (self.max - self.min) * (i + 1) as f32 / slider_width as f32;
                let px = &mut row[text_width + i];
                px.fg = CLR_WHITE;
                px.bg = Color::rgb(64, 64, 64);
                px.char = if as_float_left <= self.value {
                    [' ', '▏', '▎', '▍', '▌', '▋', '▊', '▉', '█'][(8.0
                        * ((self.value - as_float_left) / (as_float_right - as_float_left))
                            .clamp(0.0, 1.0))
                    .round()
                        as usize]
                } else {
                    ' '
                };
            }
        }
    }
}
