use std::time::Instant;

use asciivation_duty_renderer::{
    part::Settings,
    render::{render, Color, Pixel, CLR_WHITE, PX_BLACK},
};
use uianimator::{default_animator_f64_quadratic::DefaultAnimatorF64Quadratic, Animator};

use crate::state::GlobalState;

pub struct SpaceshipsList {
    start: Instant,
    width: usize,
    pub(crate) scroll: usize,
    pub(crate) scrollf: DefaultAnimatorF64Quadratic,
    pub(crate) spaceship_height: usize,
}
impl SpaceshipsList {
    #[allow(clippy::new_without_default)]
    pub fn new() -> Self {
        Self {
            start: Instant::now(),
            width: 0,
            scroll: 0,
            scrollf: DefaultAnimatorF64Quadratic::new(0.0, 30.0),
            spaceship_height: 1,
        }
    }
    pub fn set_width(&mut self, width: usize) {
        self.width = width;
    }
    pub fn width(&self) -> usize {
        self.width
    }
}
impl SpaceshipsList {
    pub fn draw(
        &mut self,
        screen_buf: &mut Vec<Vec<Pixel>>,
        settings: &Settings,
        state: &mut GlobalState,
        next_spaceship_index: Option<usize>,
        now: Instant,
    ) {
        let preview_height = (24.0 / settings.rel_height).round() as usize;
        let spaceship_height = 3 + preview_height;
        self.spaceship_height = spaceship_height;
        let scroll_factor = 3 + spaceship_height;
        self.scrollf
            .set_target((self.scroll * scroll_factor) as _, now);
        let scroll = self.scrollf.get_value(now).round() as usize;

        // draw every spaceship as a (24+stats)x(3+24/rel_height) rectangle,
        // if there is at least a width of 24, with a
        // vertical padding between spaceships of 3 lines.
        //
        //          ------------------------  <- reference 24 char width
        //
        //                   Title
        //          max-24-char catch-phrase
        //
        // <stats>    +------------------+
        // <stats>    |    Spaceship     |
        // <stats>    |     Preview      |
        // <stats>    |                  |
        // <stats>    |     (20x20)      |
        // <stats>    +------------------+

        if screen_buf.len() > 5 && self.width >= 5 {
            // width - 28, or 0 if it is only 24 or less
            let left = self.width.saturating_sub(28);
            let leftplus2 = self.width.saturating_sub(26);

            let selected_spaceship_index = state.config.selected_spaceship_index();
            // slowly rotate all spaceships
            for (i, spaceship) in state.config.spaceships_mut().enumerate() {
                spaceship
                    .preview_world
                    .elem_in_order_mut(spaceship.model_in_preview_world_order)
                    .rotation = Some((
                    (((now - self.start).as_secs_f32() - i as f32) / 10.0 * std::f32::consts::PI),
                    0.0,
                    0.0,
                ));
            }
            // show spaceships in list
            for (i, spaceship) in state.config.spaceships().enumerate() {
                let y1 = 3 + i * (spaceship_height + 3);

                // show title
                if let Some(row) = y1.checked_sub(scroll).and_then(|y| screen_buf.get_mut(y)) {
                    let row = &mut row[left..self.width];
                    let len = spaceship.name.chars().count();
                    let skip = row.len().saturating_sub(len) / 2;
                    for (px, ch) in row.iter_mut().skip(skip).zip(spaceship.name.chars()) {
                        px.fg = CLR_WHITE;
                        px.char = ch;
                    }
                }

                // show phrase
                if let Some(row) = (y1 + 1)
                    .checked_sub(scroll)
                    .and_then(|y| screen_buf.get_mut(y))
                {
                    let row = &mut row[left..self.width];
                    let len = spaceship.phrase.chars().count();
                    let skip = row.len().saturating_sub(len) / 2;
                    for (px, ch) in row.iter_mut().skip(skip).zip(spaceship.phrase.chars()) {
                        px.fg = Color::rgb(127, 127, 127);
                        px.char = ch;
                    }
                }

                // show 3d model
                if self.width >= 24 {
                    let spaceship_image_width = 24usize;
                    let spaceship_image_height = preview_height;
                    let spaceship_image_pos = (leftplus2, y1 + 3);
                    if spaceship_image_height > 1 {
                        // copy pixels of that specific area
                        let mut spaceship_image_area = Vec::with_capacity(spaceship_image_height);
                        for y in 0..spaceship_image_height {
                            if let Some(row) = (y1 + 3 + y)
                                .checked_sub(scroll)
                                .and_then(|y| screen_buf.get_mut(y))
                            {
                                let mut copied_row = Vec::with_capacity(spaceship_image_width);
                                for px in row
                                    .iter()
                                    .skip(spaceship_image_pos.0)
                                    .take(spaceship_image_width)
                                {
                                    copied_row.push(*px);
                                }
                                spaceship_image_area.push(copied_row);
                            } else {
                                spaceship_image_area.push(vec![PX_BLACK; spaceship_image_width]);
                            }
                        }
                        while spaceship_image_area.len() < spaceship_image_height {
                            spaceship_image_area.push(vec![PX_BLACK; spaceship_image_width]);
                        }
                        // render to those pixels
                        if i == selected_spaceship_index
                            || next_spaceship_index.is_some_and(|n| i == n)
                        {
                            for (x, y, ch) in [
                                (0, 0, '▛'),
                                (0, spaceship_image_height - 1, '▙'),
                                (spaceship_image_width - 1, 0, '▜'),
                                (spaceship_image_width - 1, spaceship_image_height - 1, '▟'),
                            ] {
                                spaceship_image_area[y][x].fg = if i == selected_spaceship_index {
                                    Color::rgb(64, 64, 64)
                                } else {
                                    CLR_WHITE
                                };
                                spaceship_image_area[y][x].char = ch;
                            }
                        }
                        render(
                            &spaceship.preview_world,
                            (0.0, 3.0, 0.0),
                            (0.0, std::f32::consts::FRAC_PI_2, 0.0),
                            settings.rel_height,
                            2.0,
                            1.0,
                            &mut spaceship_image_area,
                        );
                        // put the pixels back on screen
                        for y in 0..spaceship_image_height {
                            if let Some(row) = (y1 + 3 + y)
                                .checked_sub(scroll)
                                .and_then(|y| screen_buf.get_mut(y))
                            {
                                for (px, px_new) in row
                                    .iter_mut()
                                    .skip(spaceship_image_pos.0)
                                    .zip(spaceship_image_area[y].iter())
                                {
                                    *px = *px_new;
                                }
                            } else {
                                spaceship_image_area.push(vec![PX_BLACK; spaceship_image_width]);
                            }
                        }
                    }

                    // show bar chart of stats
                    if leftplus2 > 0 && spaceship_image_height > 1 {
                        let text_row = spaceship_image_pos.1 + spaceship_image_height - 1;
                        let bar_base = text_row - 1;
                        let bar_height = spaceship_image_height - 1;
                        if bar_height > 1 {
                            let stats_width = leftplus2;
                            let selected = state.config.selected_spaceship();
                            let mut stats = vec![
                                (
                                    super::CHAR_HEALTH,
                                    super::CLR_HEALTH,
                                    spaceship.health,
                                    state.config.minmax_health,
                                    selected.health,
                                ),
                                (
                                    super::CHAR_SHOT_POWER,
                                    super::CLR_SHOT_POWER,
                                    spaceship.shot_power,
                                    state.config.minmax_shot_power,
                                    selected.shot_power,
                                ),
                                (
                                    super::CHAR_SHOT_RATE,
                                    super::CLR_SHOT_RATE,
                                    spaceship.shot_rate,
                                    state.config.minmax_shot_rate,
                                    selected.shot_rate,
                                ),
                                (
                                    super::CHAR_SPEED,
                                    super::CLR_SPEED,
                                    spaceship.speed,
                                    state.config.minmax_speed,
                                    selected.speed,
                                ),
                            ];
                            stats.truncate(stats_width);
                            if !stats.is_empty() {
                                let spacing = ((stats_width + 1) / (stats.len() + 1)).min(4);
                                for (i, (stat_char, stat_char_color, val, (min, max), cmp)) in
                                    stats.into_iter().rev().enumerate()
                                {
                                    let col = stats_width - (i + 1) * spacing;
                                    if let Some(text_row) = text_row.checked_sub(scroll) {
                                        if text_row < screen_buf.len() {
                                            screen_buf[text_row][col].fg = stat_char_color;
                                            screen_buf[text_row][col].char = stat_char;
                                        }
                                    }
                                    let bar_top = bar_base
                                        .saturating_sub(scroll + bar_height - 1)
                                        .min(screen_buf.len() - 1);
                                    if let Some(bar_bottom) = bar_base
                                        .checked_sub(scroll)
                                        .map(|v| v.min(screen_buf.len() - 1))
                                    {
                                        let bar_height = bar_bottom + 1 - bar_top;
                                        if bar_height > 0 {
                                            super::draw_stats_bar(
                                                screen_buf,
                                                col,
                                                bar_bottom,
                                                bar_height,
                                                val,
                                                min,
                                                max,
                                                Some(cmp),
                                            );
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}
