use std::time::{Duration, Instant};

use asciivation_duty_renderer::{
    deps::crossterm::event::{Event, KeyCode, KeyEvent, KeyEventKind, MouseEvent, MouseEventKind},
    part::{Part, PartRunner},
    render::{clear_pre_render, PX_BLACK},
};
use uianimator::Animator;

use crate::{background, config::Config, popups::Popups, state::GlobalState};

use super::{
    super::{main_menu::MainMenu, AnyPart},
    list::SpaceshipsList,
    ship::SpaceshipDetails,
};

pub struct SpaceshipSelectionScreen {
    spaceships_list: SpaceshipsList,
    next_details: Option<SpaceshipDetails>,
    current_details: SpaceshipDetails,
    /// if  None , there is no transition happening right now.
    /// if `Some((_, None))`, the `next_details` panel is opening
    /// if `Some((_, Some(false)))`, the `next_details` panel is closing
    /// if `Some((_, Some(true)))`, the `current_details` panel is closing and will be replaced by next_details once it is closed
    /// the `bool` is whether to switch to the next_details spaceship (`true`) or to keep the current one (`false`),
    /// the `Instant` is when the animation started.
    transition: Option<(Instant, Option<bool>)>,
    transition_duration: Duration,
    /// what to do once the transition finishes.
    /// also used to start a transition or trigger an action.
    post_transition: PostTransition,
}
pub(crate) enum PostTransition {
    DoNothing,
    GoToMainMenu,
    SelectNextSpaceship(usize),
    CancelNextSpaceship,
    ConfirmNextSpaceship,
}

impl SpaceshipSelectionScreen {
    #[allow(clippy::new_without_default)]
    pub fn new() -> Self {
        let transition_duration = Duration::from_millis(300);
        Self {
            spaceships_list: SpaceshipsList::new(),
            next_details: None,
            current_details: SpaceshipDetails::new_current(transition_duration, None),
            transition: None,
            transition_duration,
            post_transition: PostTransition::DoNothing,
        }
    }
}

impl Part for SpaceshipSelectionScreen {
    type End = AnyPart;
    type State = GlobalState;
    fn draw(runner: &mut PartRunner<Self>, now: Instant) {
        let width = if runner.screen_buf.is_empty() {
            0
        } else {
            runner.screen_buf[0].len()
        };

        if runner.draw_reset {
            // make star background opaque
            runner.state.background.set_transparency(0.0);
            // set x position of the 2-3 parts
            runner.part.set_splits(
                if runner.part.next_details.is_some() {
                    0.0
                } else {
                    1.0
                },
                false,
                width,
            );
        }

        // transitioning, calculate positions for animation
        if let Some((transition_started, transition)) = runner.part.transition {
            let progress = (now - transition_started)
                .div_duration_f64(runner.part.transition_duration)
                .min(1.0);
            let p = (1.0 - progress) * (1.0 - progress);
            match transition {
                None | Some(false) => {
                    runner.part.set_splits(
                        if transition.is_some() { 1.0 - p } else { p },
                        false,
                        width,
                    );
                }
                Some(true) => {
                    runner.part.set_splits(p, true, width);
                }
            }
            if progress == 1.0 {
                match transition {
                    None => (),
                    Some(false) => runner.part.next_details = None,
                    Some(true) => {
                        if let Some(next) = runner.part.next_details.take() {
                            runner.part.current_details = next;
                            runner.state.config.select_spaceship(
                                runner
                                    .part
                                    .current_details
                                    .spaceship()
                                    .unwrap_or(runner.state.config.selected_spaceship_index()),
                            );
                            runner.part.current_details.set_as_default_instantly();
                            runner.part.current_details.comparing_to(None);
                            runner.part.set_splits(1.0, false, width);
                        }
                    }
                }
                runner.part.transition = None;
            }
        }

        // not transitioning, check `post_transition`
        if runner.part.transition.is_none() {
            match runner.part.post_transition {
                PostTransition::DoNothing => (),
                PostTransition::GoToMainMenu => {
                    runner
                        .state
                        .config
                        .set_rel_height(runner.settings.rel_height);
                    runner.state.config.save(&mut runner.state.popups);
                    runner.part.post_transition = PostTransition::DoNothing;
                    runner.end = Some(AnyPart::MainMenu(MainMenu::new_from_spaceship_selection()));
                }
                PostTransition::SelectNextSpaceship(spaceship) => {
                    runner.part.post_transition = PostTransition::DoNothing;
                    runner.part.current_details.comparing_to(Some(spaceship));
                    if let Some(next_details) = &mut runner.part.next_details {
                        next_details.change_spaceship(Some(spaceship), now);
                    } else {
                        runner.part.next_details = Some(SpaceshipDetails::new_next(
                            spaceship,
                            runner.part.transition_duration,
                            Some(runner.state.config.selected_spaceship_index()),
                        ));
                        runner.part.transition = Some((now, None));
                    }
                }
                PostTransition::CancelNextSpaceship => {
                    runner.part.current_details.comparing_to(None);
                    runner.part.post_transition = PostTransition::DoNothing;
                    if runner.part.next_details.is_some() {
                        runner.part.transition = Some((now, Some(false)));
                    }
                }
                PostTransition::ConfirmNextSpaceship => {
                    // ship switching is handeled post-transition
                    runner.part.post_transition = PostTransition::DoNothing;
                    if runner.part.next_details.is_some() {
                        runner.part.transition = Some((now, Some(true)));
                    }
                }
            }
        }

        if !runner.screen_buf.is_empty() {
            // fill entire screen buffer with black pixels
            clear_pre_render(&mut runner.screen_buf, PX_BLACK);

            // draw background stars
            background::draw(runner, now);

            // show the spaceship list (left)
            runner.part.spaceships_list.draw(
                &mut runner.screen_buf,
                &runner.settings,
                &mut runner.state,
                runner
                    .part
                    .next_details
                    .as_ref()
                    .filter(|_| {
                        runner
                            .part
                            .transition
                            .is_none_or(|(_, t)| t.is_none_or(|v| v))
                    })
                    .and_then(|v| v.spaceship()),
                now,
            );
            // show the next spaceship, if one is selected
            if let Some(next_details) = &mut runner.part.next_details {
                next_details.draw(&mut runner.screen_buf, &runner.settings, &runner.state, now);
            }
            // show the current spaceship
            runner.part.current_details.draw(
                &mut runner.screen_buf,
                &runner.settings,
                &runner.state,
                now,
            );
        }

        Popups::draw(&mut runner.screen_buf, &mut runner.state, now);
    }
    fn event(runner: &mut PartRunner<Self>, e: Event) {
        if super::super::event(runner, &e) || runner.state.popups.event(&e) {
            return;
        }
        match e {
            // enter "clicks" on the first spaceship
            Event::Key(KeyEvent {
                code: KeyCode::Enter,
                kind: KeyEventKind::Press,
                ..
            }) => {
                runner
                    .part
                    .select_spaceship(runner.part.spaceships_list.scroll, &runner.state.config);
            }
            // scrolling in list with MWheel or Up/Down Arrow
            Event::Key(KeyEvent {
                code: KeyCode::Up,
                modifiers: _,
                kind: KeyEventKind::Press | KeyEventKind::Repeat,
                state: _,
            })
            | Event::Mouse(MouseEvent {
                kind: MouseEventKind::ScrollUp,
                ..
            }) => {
                runner.part.spaceships_list.scroll =
                    runner.part.spaceships_list.scroll.saturating_sub(1);
            }
            Event::Key(KeyEvent {
                code: KeyCode::Down,
                modifiers: _,
                kind: KeyEventKind::Press | KeyEventKind::Repeat,
                state: _,
            })
            | Event::Mouse(MouseEvent {
                kind: MouseEventKind::ScrollDown,
                ..
            }) => {
                runner.part.spaceships_list.scroll = runner
                    .part
                    .spaceships_list
                    .scroll
                    .saturating_add(1)
                    .min(runner.state.config.spaceships_count().saturating_sub(1));
            }
            Event::Mouse(MouseEvent {
                kind: MouseEventKind::Up(_),
                column,
                row,
                modifiers: _,
            }) => {
                if (column as usize) < runner.part.spaceships_list.width() {
                    // clicking in list
                    let scroll = runner
                        .part
                        .spaceships_list
                        .scrollf
                        .get_value(Instant::now())
                        .round() as usize;
                    let y = scroll + row as usize;
                    if y % (runner.part.spaceships_list.spaceship_height + 3) >= 3 {
                        // didn't click between two ships
                        let ship = y / (runner.part.spaceships_list.spaceship_height + 3);
                        runner.part.select_spaceship(ship, &runner.state.config);
                    }
                } else if let Some(next_details) = &runner.part.next_details {
                    if next_details.left() <= (column as usize)
                        && (column as usize) < next_details.left() + next_details.width()
                    {
                        // clicking on the "next spaceship" view -> confirm that spaceship
                        runner.part.post_transition = PostTransition::ConfirmNextSpaceship;
                    } else if runner.part.current_details.left() <= (column as usize)
                        && (column as usize)
                            < runner.part.current_details.left()
                                + runner.part.current_details.width()
                    {
                        // clicking on the "previous spaceship" view -> cancel the next spaceship, keep using the previous one
                        runner.part.post_transition = PostTransition::CancelNextSpaceship;
                    }
                } else if runner.part.current_details.left() <= (column as usize)
                    && (column as usize)
                        < runner.part.current_details.left() + runner.part.current_details.width()
                {
                    // clicking on the "current spaceship" view (no next spaceship selected right now) -> okay, keep that spaceship -> go to main menu
                    runner.part.post_transition = PostTransition::GoToMainMenu;
                }
            }
            // `Esc` always goes back to the main menu
            Event::Key(KeyEvent {
                code: KeyCode::Esc,
                kind: KeyEventKind::Press,
                ..
            }) => runner.part.post_transition = PostTransition::GoToMainMenu,
            _ => (),
        }
    }
}

impl SpaceshipSelectionScreen {
    fn select_spaceship(&mut self, index: usize, config: &Config) {
        if index == config.selected_spaceship_index() {
            if self.next_details.is_some() {
                self.post_transition = PostTransition::CancelNextSpaceship;
            } else {
                self.post_transition = PostTransition::GoToMainMenu;
            }
        } else if let Some(next_details) = &self.next_details {
            if next_details.spaceship().is_some_and(|i| i == index) {
                self.post_transition = PostTransition::ConfirmNextSpaceship;
            } else {
                self.post_transition = PostTransition::SelectNextSpaceship(index);
            }
        } else {
            self.post_transition = PostTransition::SelectNextSpaceship(index);
        }
    }

    fn split12(&self, width: usize) -> f64 {
        ((width as f64) * 0.20)
            .max(24.0 /* min width of list element */ + 2.0 /* right-of-element padding */ + 4.0 /* stats */)
            .min((width / 2) as f64)
    }
    fn split13(&self, width: usize) -> f64 {
        width.min(24 + 2 + 4 /* stats */) as f64
    }
    fn set_splits(&mut self, rev: f64, close_current: bool, width: usize) {
        if !close_current {
            // `next` panel is opening / closing
            let split1 = rev * self.split12(width) + (1.0 - rev) * self.split13(width);
            let split2 = rev * split1 + (1.0 - rev) * ((split1 + width as f64) / 2.0);
            let split1 = split1.round() as usize;
            let split2 = split2.round() as usize;
            self.spaceships_list.set_width(split1);
            if let Some(next_details) = &mut self.next_details {
                next_details.set_pos(split1, split2 - split1);
            }
            self.current_details.set_pos(split2, width - split2);
        } else {
            // `current` panel is closing
            let split1 = rev * self.split13(width) + (1.0 - rev) * self.split12(width);
            let split2 = rev * ((split1 + width as f64) / 2.0) + (1.0 - rev) * width as f64;
            let split1 = split1.round() as usize;
            let split2 = split2.round() as usize;
            self.spaceships_list.set_width(split1);
            if let Some(next_details) = &mut self.next_details {
                next_details.set_pos(split1, split2 - split1);
            }
            self.current_details.set_pos(split2, width - split2);
        }
    }
}
