use std::time::{Duration, Instant};

use asciivation_duty_renderer::{
    part::Settings,
    render::{render, Color, Pixel, CLR_WHITE},
};

use crate::{
    config::{Config, Spaceship},
    state::GlobalState,
};

pub struct SpaceshipDetails {
    start: Instant,
    /// if none, use the default, if `Some(_)`, this `SpaceshipDetails` is the `next`
    /// part of the spaceship selection screen and not the `current` part.
    spaceship_index: Option<usize>,
    comparing_to: Option<usize>,
    left: usize,
    width: usize,
    transition: Option<Instant>,
    transition_duration: Duration,
}
impl SpaceshipDetails {
    /// shows the currently selected spaceship
    pub fn new_current(transition_duration: Duration, comparing_to: Option<usize>) -> Self {
        Self::new(None, transition_duration, comparing_to)
    }
    /// shows the given spaceship
    pub fn new_next(
        spaceship_index: usize,
        transition_duration: Duration,
        comparing_to: Option<usize>,
    ) -> Self {
        Self::new(Some(spaceship_index), transition_duration, comparing_to)
    }
    fn new(
        spaceship_index: Option<usize>,
        transition_duration: Duration,
        comparing_to: Option<usize>,
    ) -> Self {
        Self {
            start: Instant::now(),
            spaceship_index,
            comparing_to,
            left: 0,
            width: 0,
            transition: None,
            transition_duration,
        }
    }
    pub fn comparing_to(&mut self, comparing_to: Option<usize>) {
        self.comparing_to = comparing_to;
    }
    pub fn set_pos(&mut self, left: usize, width: usize) {
        self.left = left;
        self.width = width;
    }
    pub fn left(&self) -> usize {
        self.left
    }
    pub fn width(&self) -> usize {
        self.width
    }
    /// `None` -> show currently selected spaceship (`new_current(..)`),
    /// `Some(i)` -> show the given spaceship (`new_next(i, ..)`)
    pub fn change_spaceship(&mut self, next_spaceship_index: Option<usize>, now: Instant) {
        if self.spaceship_index != next_spaceship_index {
            self.change_spaceship_force(next_spaceship_index, now);
        }
    }
    /// like `change_spaceship`, but the animation plays even if the spaceship index has not changed.
    pub fn change_spaceship_force(&mut self, next_spaceship_index: Option<usize>, now: Instant) {
        self.start = Instant::now();
        self.spaceship_index = next_spaceship_index;
        self.transition = Some(now);
    }
    pub fn spaceship(&self) -> Option<usize> {
        self.spaceship_index
    }
    pub fn spaceship_from<'a>(&self, config: &'a Config) -> &'a Spaceship {
        self.spaceship_index
            .map(|i| config.spaceship(i))
            .unwrap_or_else(|| config.selected_spaceship())
    }
    /// without playing an animation, display the default spaceship instead of a specific one from now on.
    /// this is used directly after `SpaceshipSelectionScreen` sets the default spaceship to the one the
    /// previous index pointed to, so changing from that index to the default is actually no change.
    /// -> no animation is played -> user does not notice this action
    pub fn set_as_default_instantly(&mut self) {
        self.spaceship_index = None;
    }
}
impl SpaceshipDetails {
    pub fn draw(
        &mut self,
        screen_buf: &mut Vec<Vec<Pixel>>,
        settings: &Settings,
        state: &GlobalState,
        now: Instant,
    ) {
        let (width, height) = (self.width, screen_buf.len());

        if width < 5 || height < 5 {
            return;
        }

        let spaceship = self.spaceship_from(&state.config);

        let spaceship_image_height =
            (width as f32 * 2.0 / 3.0 / settings.rel_height).min((height - 5) as f32 / 2.0);
        let spaceship_image_width = (spaceship_image_height * settings.rel_height).round() as usize;
        let spaceship_image_height = spaceship_image_height.round() as usize;
        let spaceship_image_pos = (
            self.left + (width - spaceship_image_width) / 2,
            5 + (height / 2).saturating_sub(spaceship_image_height) / 2,
        );

        // show spaceship name
        let name_len = spaceship.name.chars().count();
        for (px, ch) in screen_buf[1][self.left + self.width.saturating_sub(name_len) / 2..]
            .iter_mut()
            .zip(spaceship.name.chars())
            .take(self.width)
        {
            px.fg = CLR_WHITE;
            px.char = ch;
        }
        // show spaceship description
        let description_len = spaceship.description.chars().count();
        let max_width_description = self.width * 4 / 5;
        if description_len > max_width_description {
            let mut j = 0;
            let mut c = 0;
            while let Some(i) = spaceship.description[j..].find([' ']) {
                let nc = spaceship.description[j..][..i].chars().count() + 1;
                if c + nc > max_width_description {
                    break;
                }
                j += i + 1;
                c += nc;
            }
            let line1 = &spaceship.description[0..j.saturating_sub(1)];
            let line2 = &spaceship.description[j..];
            for (px, ch) in screen_buf[2]
                [self.left + self.width.saturating_sub(line1.chars().count()) / 2..]
                .iter_mut()
                .zip(line1.chars())
                .take(self.width)
            {
                px.fg = Color::rgb(150, 150, 150);
                px.char = ch;
            }
            for (px, ch) in screen_buf[3]
                [self.left + self.width.saturating_sub(line2.chars().count()) / 2..]
                .iter_mut()
                .zip(line2.chars())
                .take(self.width)
            {
                px.fg = Color::rgb(150, 150, 150);
                px.char = ch;
            }
        } else {
            for (px, ch) in screen_buf[3]
                [self.left + self.width.saturating_sub(description_len) / 2..]
                .iter_mut()
                .zip(spaceship.description.chars())
                .take(self.width)
            {
                px.fg = Color::rgb(150, 150, 150);
                px.char = ch;
            }
        }

        // render spaceship
        if spaceship_image_height > 1 && spaceship_image_width > 1 {
            // copy pixels of that specific area
            let mut spaceship_image_area = Vec::with_capacity(spaceship_image_height);
            for row in screen_buf
                .iter()
                .skip(spaceship_image_pos.1)
                .take(spaceship_image_height)
            {
                let mut copied_row = Vec::with_capacity(spaceship_image_width);
                for px in row
                    .iter()
                    .skip(spaceship_image_pos.0)
                    .take(spaceship_image_width)
                {
                    copied_row.push(*px);
                }
                spaceship_image_area.push(copied_row);
            }
            // render to those pixels
            for (x, y, ch) in [
                (0, 0, '▛'),
                (0, spaceship_image_height - 1, '▙'),
                (spaceship_image_width - 1, 0, '▜'),
                (spaceship_image_width - 1, spaceship_image_height - 1, '▟'),
            ] {
                spaceship_image_area[y][x].fg = Color::rgb(64, 64, 64);
                spaceship_image_area[y][x].char = ch;
            }
            let cam_angle = 3.0 * std::f32::consts::FRAC_PI_8
                + std::f32::consts::FRAC_PI_8
                    * ((now - self.start).as_secs_f32() / 23.23 * std::f32::consts::PI).sin();
            render(
                &self.spaceship_from(&state.config).preview_world,
                (-3.0 * cam_angle.cos(), 3.0 * cam_angle.sin(), 0.0),
                (0.0, cam_angle, 0.0),
                settings.rel_height,
                2.0,
                1.0,
                &mut spaceship_image_area,
            );
            // put the pixels back on screen
            for (row, new_row) in screen_buf
                .iter_mut()
                .skip(spaceship_image_pos.1)
                .take(spaceship_image_height)
                .zip(spaceship_image_area.into_iter())
            {
                for (px, new_px) in row
                    .iter_mut()
                    .skip(spaceship_image_pos.0)
                    .take(spaceship_image_width)
                    .zip(new_row)
                {
                    *px = new_px;
                }
            }
        }

        // below spaceship
        let below_spaceship = spaceship_image_pos.1 + spaceship_image_height + 1;
        if below_spaceship < screen_buf.len() {
            let stats_height = screen_buf.len() - below_spaceship;
            if stats_height > 2 {
                let comparing_to = self.comparing_to.map(|v| state.config.spaceship(v));
                let mut stats = vec![
                    (
                        "Health",
                        super::CHAR_HEALTH,
                        super::CLR_HEALTH,
                        spaceship.health,
                        state.config.minmax_health,
                        comparing_to.map(|s| s.health),
                    ),
                    (
                        "Power : of shots",
                        super::CHAR_SHOT_POWER,
                        super::CLR_SHOT_POWER,
                        spaceship.shot_power,
                        state.config.minmax_shot_power,
                        comparing_to.map(|s| s.shot_power),
                    ),
                    (
                        "Rate : of fire",
                        super::CHAR_SHOT_RATE,
                        super::CLR_SHOT_RATE,
                        spaceship.shot_rate,
                        state.config.minmax_shot_rate,
                        comparing_to.map(|s| s.shot_rate),
                    ),
                    (
                        "Speed",
                        super::CHAR_SPEED,
                        super::CLR_SPEED,
                        spaceship.speed,
                        state.config.minmax_speed,
                        comparing_to.map(|s| s.speed),
                    ),
                ];
                stats.truncate(self.width / 4);
                if !stats.is_empty() {
                    let left = self.left + (self.width - stats.len() * 3) / 2;

                    let stats_bars_height = match stats_height.saturating_sub(
                        stats
                            .iter()
                            .map(|v| {
                                v.0.split_once(" : ")
                                    .map(|(a, b)| a.chars().count().max(b.chars().count()))
                                    .unwrap_or_else(|| v.0.chars().count())
                            })
                            .max()
                            .unwrap(),
                    ) {
                        0..=3 => stats_height - 2,
                        v => v - 3,
                    };
                    let stats_bars_base = below_spaceship + stats_bars_height - 1;
                    // draw stats
                    for (i, (stat, stat_char, stat_char_color, val, (min, max), cmp)) in
                        stats.into_iter().enumerate()
                    {
                        let col1 = left + 4 * i + 1;
                        let col2 = col1 + 1;

                        // draw `▂` just above the max value
                        screen_buf[below_spaceship - 1][col1].fg = Color::rgb(64, 64, 64);
                        screen_buf[below_spaceship - 1][col1].char = '▂';

                        // draw columns of stats diagram
                        super::draw_stats_bar(
                            screen_buf,
                            col1,
                            stats_bars_base,
                            stats_bars_height,
                            val,
                            min,
                            max,
                            cmp,
                        );
                        // show stat char
                        if let Some(row) =
                            screen_buf.get_mut(below_spaceship + stats_bars_height + 1)
                        {
                            row[col1].fg = stat_char_color;
                            row[col1].char = stat_char;
                        }
                        // show stat label
                        let (text, text2) = stat
                            .split_once(" : ")
                            .map(|(a, b)| (a, Some(b)))
                            .unwrap_or((stat, None));
                        let label_top = stats_bars_base + 4;
                        let label_bottom = below_spaceship + stats_height;
                        if label_top < label_bottom {
                            for (y, ch) in (label_top..label_bottom).zip(text.chars()) {
                                screen_buf[y][col1].fg = CLR_WHITE;
                                screen_buf[y][col1].char = ch;
                            }
                        }
                        if let Some(text) = text2 {
                            if label_top < label_bottom {
                                for (y, ch) in (label_top..label_bottom).zip(text.chars()) {
                                    screen_buf[y][col2].fg = Color::rgb(100, 100, 100);
                                    screen_buf[y][col2].char = ch;
                                }
                            }
                        }
                    }
                }
            }
        }

        // if we are transitioning to a different spaceship, fade from black
        if let Some(transition) = self.transition {
            let p = (now - transition)
                .div_duration_f64(self.transition_duration)
                .min(1.0);
            let p1 = 1.0 - p;
            let p2 = p1 * p1;
            let clr = Color::rgba(0, 0, 0, (255.0 * p2).round() as _);
            for row in &mut *screen_buf {
                for px in row.iter_mut().skip(self.left).take(self.width) {
                    px.bg = px.bg.merge(clr);
                    px.fg = px.fg.merge(clr);
                }
            }
            if p == 1.0 {
                self.transition = None;
            }
        }
    }
}
