use std::{
    collections::VecDeque,
    time::{Duration, Instant},
};

use asciivation_duty_renderer::{
    deps::crossterm::event::{Event, KeyCode, KeyEvent, KeyEventKind, MouseEvent, MouseEventKind},
    render::{Pixel, CLR_BLACK, CLR_WHITE, PX_BLACK},
};

use crate::{state::GlobalState, websocket::WsColor};

/// Contains and manages the currently active popups
#[derive(Debug)]
pub struct Popups {
    last_thread_check: Instant,
    thread_check_interval: Duration,
    popup_sender: PopupSender,
    popup_receiver: std::sync::mpsc::Receiver<AnyPopup>,
    popups_warn: VecDeque<Popup>,
    popups_err: VecDeque<Popup>,
    popup_shown_since: Option<Instant>,
    text_rotation_since: Option<Instant>,
    size: (usize, usize),
    body_available_size: (usize, usize),
    body: Vec<String>,
    scroll: usize,
    popup_changed: bool,
}
/// Used to send information from other threads to the main thread,
/// including (but not limited to) warning/error popups.
#[derive(Clone, Debug)]
pub struct PopupSender(std::sync::mpsc::Sender<AnyPopup>);
/// The type of data that can be sent from other threads to the main thread.
/// `Silent(_)` represents messages that should not cause a normal popup to appear.
/// in other ways.
pub enum AnyPopup {
    Warning(Popup),
    Error(Popup),
    Silent(PMessage),
}
/// A Message that can be sent from other threads to the main thread
/// to change the application's global state.
#[derive(Debug)]
pub enum PMessage {
    SetSpaceshipColor(WsColor),
}
impl Popups {
    pub fn show_warning(&mut self, popup: Popup) {
        if self.popups_err.is_empty() && self.popups_warn.is_empty() {
            self.popup_changed = true;
        }
        self.popups_warn.push_back(popup);
    }
    pub fn show_error(&mut self, popup: Popup) {
        if self.popups_err.is_empty() {
            self.popup_changed = true;
        }
        self.popups_err.push_back(popup);
    }
    #[allow(clippy::new_without_default)]
    pub fn new() -> Self {
        let (popup_sender, popup_receiver) = std::sync::mpsc::channel();
        Self {
            last_thread_check: Instant::now(),
            thread_check_interval: Duration::from_millis(330),
            popup_sender: PopupSender(popup_sender),
            popup_receiver,
            popups_warn: VecDeque::new(),
            popups_err: VecDeque::new(),
            popup_shown_since: None,
            text_rotation_since: None,
            size: (0, 0),
            body_available_size: (0, 0),
            body: vec![],
            scroll: 0,
            popup_changed: false,
        }
    }
    pub fn has_popup(&self) -> bool {
        !(self.popups_warn.is_empty() && self.popups_err.is_empty())
    }
    pub fn draw(
        #[allow(clippy::ptr_arg)] screen_buf: &mut Vec<Vec<Pixel>>,
        state: &mut GlobalState,
        now: Instant,
    ) {
        if (now - state.popups.last_thread_check) > state.popups.thread_check_interval {
            state.popups.last_thread_check = now;
            while let Ok(popup) = state.popups.popup_receiver.try_recv() {
                match popup {
                    AnyPopup::Warning(popup) => state.popups.show_warning(popup),
                    AnyPopup::Error(popup) => state.popups.show_error(popup),
                    AnyPopup::Silent(msg) => match msg {
                        PMessage::SetSpaceshipColor(c) => {
                            if !state.spaceship_color_ignore_websocket {
                                state.spaceship_color = c;
                                state.draw_reset = true;
                            }
                        }
                    },
                }
            }
        }
        if screen_buf.is_empty() {
            return;
        }
        let size = (screen_buf[0].len(), screen_buf.len());
        if size.0 == 0 {
            return;
        }
        if state.popups.popup_changed {
            state.popups.popup_shown_since = Some(now);
        }
        if size != state.popups.size || state.popups.popup_changed {
            state.popups.size = size;
            state.popups.relayout();
        }
        state.popups.popup_changed = false;
        if state.popups.body_available_size.0 > 0 && state.popups.body_available_size.1 > 0 {
            if let Some(popup) = state
                .popups
                .popups_err
                .front()
                .or(state.popups.popups_warn.front())
            {
                let top = (state.popups.size.1 - state.popups.body_available_size.1) / 2;
                let left = (state.popups.size.0 - state.popups.body_available_size.0) / 2;

                // black background
                for y in 0..state.popups.body_available_size.1 + 2 {
                    let row = &mut screen_buf[top + y - 2];
                    for x in 0..state.popups.body_available_size.0 {
                        row[left + x] = PX_BLACK;
                    }
                }

                // outline
                let bg = CLR_BLACK;
                let fg = CLR_WHITE;
                // draw corners
                screen_buf[top - 3][left - 1] = Pixel {
                    bg, fg, char: '┌'
                };
                screen_buf[top - 3][left + state.popups.body_available_size.0] =
                    Pixel {
                        bg, fg, char: '┐'
                    };
                screen_buf[top + state.popups.body_available_size.1][left - 1] =
                    Pixel {
                        bg, fg, char: '└'
                    };
                screen_buf[top + state.popups.body_available_size.1]
                    [left + state.popups.body_available_size.0] = Pixel {
                    bg, fg, char: '┘'
                };
                // draw vertical lines
                for row in &mut screen_buf[top - 2..][..state.popups.body_available_size.1 + 2] {
                    row[left - 1] = Pixel {
                        bg, fg, char: '│'
                    };
                }
                for row in &mut screen_buf[top - 2..][..state.popups.body_available_size.1 + 2] {
                    row[left + state.popups.body_available_size.0] = Pixel {
                        bg, fg, char: '│'
                    };
                }
                screen_buf[top - 1][left - 1] = Pixel {
                    bg, fg, char: '├'
                };
                screen_buf[top - 1][left + state.popups.body_available_size.0] =
                    Pixel {
                        bg, fg, char: '┤'
                    };
                // draw horizontal lines
                for px in &mut screen_buf[top - 3][left..][..state.popups.body_available_size.0] {
                    *px = Pixel {
                        bg, fg, char: '─'
                    };
                }
                for px in &mut screen_buf[top - 1][left..][..state.popups.body_available_size.0] {
                    *px = Pixel {
                        bg, fg, char: '─'
                    };
                }
                for px in &mut screen_buf[top + state.popups.body_available_size.1][left..]
                    [..state.popups.body_available_size.0]
                {
                    *px = Pixel {
                        bg, fg, char: '─'
                    };
                }

                // draw title
                let max_skip = popup
                    .title_chars
                    .saturating_sub(state.popups.body_available_size.0);
                let center_skip = state
                    .popups
                    .body_available_size
                    .0
                    .saturating_sub(popup.title_chars)
                    / 2;
                if max_skip == 0 {
                    state.popups.text_rotation_since = None;
                }
                let padding = "        ";
                for (i, ch) in " "
                    .repeat(center_skip)
                    .chars()
                    .chain(popup.title.chars().skip(if max_skip > 0 {
                        (((now - *state.popups.text_rotation_since.get_or_insert(now))
                            .as_secs_f32()
                            * 2.0)
                            % (max_skip as f32 + 6.0)
                            - 3.0)
                            .max(0.0)
                            .min(max_skip as f32)
                            .round() as usize
                    } else {
                        0
                    }))
                    .chain(padding.chars())
                    .chain(
                        if max_skip > 0 {
                            popup.title.as_str()
                        } else {
                            ""
                        }
                        .chars(),
                    )
                    .take(state.popups.body_available_size.0)
                    .enumerate()
                {
                    screen_buf[top - 2][left + i] = Pixel {
                        bg: CLR_BLACK,
                        fg: CLR_WHITE,
                        char: ch,
                    }
                }

                // draw body
                for (j, row) in state
                    .popups
                    .body
                    .iter()
                    .skip(state.popups.scroll)
                    .take(state.popups.body_available_size.1)
                    .enumerate()
                {
                    for (i, ch) in row
                        .chars()
                        .take(state.popups.body_available_size.0)
                        .enumerate()
                    {
                        screen_buf[top + j][left + i] = Pixel {
                            bg: CLR_BLACK,
                            fg: CLR_WHITE,
                            char: ch,
                        }
                    }
                }
            }
        }
    }
    pub fn event(&mut self, e: &Event) -> bool {
        if self.popups_warn.is_empty() && self.popups_err.is_empty() {
            return false;
        }
        let (skip, close) = match e {
            Event::Key(KeyEvent {
                code: KeyCode::Esc,
                modifiers: _,
                kind: KeyEventKind::Press,
                state: _,
            }) => (true, true),
            Event::Mouse(MouseEvent {
                kind: MouseEventKind::Up(_),
                column,
                row,
                modifiers: _,
            }) => {
                if self.size.0 > 6 && self.size.1 > 6 {
                    let top = (self.size.1 - self.body_available_size.1) / 2;
                    let left = (self.size.0 - self.body_available_size.0) / 2;
                    (
                        true,
                        (*column as usize) < left - 1
                            || (*column as usize) > left + self.body_available_size.0
                            || (*row as usize) < top - 3
                            || (*row as usize) > top + self.body_available_size.1,
                    )
                } else {
                    (true, false)
                }
            }
            Event::Mouse(MouseEvent {
                kind: MouseEventKind::Down(_) | MouseEventKind::Drag(_),
                ..
            }) => (true, false),
            Event::Key(KeyEvent {
                code: KeyCode::Up,
                modifiers: _,
                kind: KeyEventKind::Press | KeyEventKind::Repeat,
                state: _,
            })
            | Event::Mouse(MouseEvent {
                kind: MouseEventKind::ScrollUp,
                ..
            }) => {
                self.scroll = self.scroll.saturating_sub(1);
                (true, false)
            }
            Event::Key(KeyEvent {
                code: KeyCode::Down,
                modifiers: _,
                kind: KeyEventKind::Press | KeyEventKind::Repeat,
                state: _,
            })
            | Event::Mouse(MouseEvent {
                kind: MouseEventKind::ScrollDown,
                ..
            }) => {
                self.scroll = self
                    .scroll
                    .saturating_add(1)
                    .min(self.body.len().saturating_sub(2));
                (true, false)
            }
            Event::Key(_) => (true, false),
            _ => (false, false),
        };
        if close
            && self
                .popup_shown_since
                .is_some_and(|t| t.elapsed().as_secs_f32() > 0.5)
        {
            if self.popups_err.pop_front().is_none() {
                let _ = self.popups_warn.pop_front().is_some();
            }
            self.relayout();
        }
        skip
    }
    pub fn relayout(&mut self) {
        if self.size.0 > 5 && self.size.1 > 5 {
            self.body_available_size.0 = self.size.0 / 3;
            self.body_available_size.1 = self.size.1 / 2;
        } else {
            self.body_available_size = (0, 0);
        }
        self.body = vec![];
        if self.body_available_size.0 == 0 || self.body_available_size.1 == 0 {
            return;
        }
        if let Some(popup) = self.popups_err.front().or(self.popups_warn.front()) {
            let mut line = String::new();
            let mut line_chars = 0;
            let mut line_end_whitespace = 0;
            let mut word = String::new();
            let mut word_chars = 0;
            for ch in popup.body.chars().filter_map(char_filter) {
                if ch == '\n' {
                    push_word(
                        &mut self.body,
                        &mut line,
                        &mut line_chars,
                        &mut line_end_whitespace,
                        &mut word,
                        &mut word_chars,
                        false,
                        self.body_available_size.0,
                    );
                    push_line(
                        &mut self.body,
                        &mut line,
                        &mut line_chars,
                        &mut line_end_whitespace,
                    );
                    continue;
                }
                word.push(ch);
                word_chars += 1;
                // line full from single word or word-terminating characters found
                if ch.is_whitespace() {
                    line_end_whitespace += 1;
                } else {
                    line_end_whitespace = 0;
                }
                if matches!(ch, ' ' | '-') {
                    // word was terminated, push it
                    push_word(
                        &mut self.body,
                        &mut line,
                        &mut line_chars,
                        &mut line_end_whitespace,
                        &mut word,
                        &mut word_chars,
                        ch.is_whitespace(),
                        self.body_available_size.0,
                    );
                } else if word_chars > self.body_available_size.0 {
                    // word is longer than an entire line, just push it directly
                    let mut word_iter = word.chars();
                    'outer: loop {
                        while line_chars < self.body_available_size.0 {
                            if let Some(ch) = word_iter.next() {
                                line.push(ch);
                                line_chars += 1;
                            } else {
                                break 'outer;
                            }
                        }
                        push_line(
                            &mut self.body,
                            &mut line,
                            &mut line_chars,
                            &mut line_end_whitespace,
                        );
                    }
                    word = String::new();
                    word_chars = 0;
                }
            }
            push_word(
                &mut self.body,
                &mut line,
                &mut line_chars,
                &mut line_end_whitespace,
                &mut word,
                &mut word_chars,
                false,
                self.body_available_size.0,
            );
            push_line(
                &mut self.body,
                &mut line,
                &mut line_chars,
                &mut line_end_whitespace,
            );
        } else {
            self.scroll = 0;
        }

        #[allow(clippy::too_many_arguments)]
        fn push_word(
            body: &mut Vec<String>,
            line: &mut String,
            line_chars: &mut usize,
            line_end_whitespace: &mut usize,
            word: &mut String,
            word_chars: &mut usize,
            word_end_whitespace: bool,
            body_available_width: usize,
        ) {
            if *line_chars + *word_chars - if word_end_whitespace { 1 } else { 0 }
                > body_available_width
            {
                push_line(body, line, line_chars, line_end_whitespace);
            }
            line.push_str(word);
            *line_chars += *word_chars;
            *word = String::new();
            *word_chars = 0;
        }
        fn push_line(
            body: &mut Vec<String>,
            line: &mut String,
            line_chars: &mut usize,
            line_end_whitespace: &mut usize,
        ) {
            // remove trailing whitespace
            for _ in 0..*line_end_whitespace {
                line.pop();
            }
            body.push(std::mem::take(line));
            *line_chars = 0;
            *line_end_whitespace = 0;
        }
    }
    pub fn popup_sender(&self) -> PopupSender {
        self.popup_sender.clone()
    }
}
impl PopupSender {
    pub fn show_warning(&self, popup: Popup) {
        self.show_popup(AnyPopup::Warning(popup));
    }
    pub fn show_error(&self, popup: Popup) {
        self.show_popup(AnyPopup::Error(popup));
    }
    pub fn show_popup(&self, popup: AnyPopup) {
        let _ = self.0.send(popup);
    }
    pub fn message(&self, msg: PMessage) {
        self.show_popup(AnyPopup::Silent(msg))
    }
}

/// A popup. Can be used as a warning or an error.
#[derive(Debug)]
pub struct Popup {
    title: String,
    title_chars: usize,
    body: String,
}
impl Popup {
    pub fn new(title: impl Into<String>, body: impl Into<String>) -> Self {
        let title: String = title.into();
        let title_chars = title.chars().count();
        Self {
            title,
            title_chars,
            body: body.into(),
        }
    }
}

fn char_filter(ch: char) -> Option<char> {
    Some(match ch {
        '\t' => ' ',
        '\r' => return None,
        ch => ch,
    })
}
