use std::ffi::OsStr;

use rand::Rng;
use serde::{Deserialize, Serialize};

use crate::{
    config::{Config, SpaceshipStats},
    parts::game_state::ObstacleSpawnMode,
    popups::{Popup, Popups},
};

/// Contains up to 10.000 of the best scores achieved while playing the game.
#[derive(Clone, Serialize, Deserialize, PartialEq)]
pub struct Highscores {
    scores: Vec<Score>,
}
impl Highscores {
    pub fn new_empty() -> Self {
        Self { scores: vec![] }
    }
    pub fn is_empty(&self) -> bool {
        self.scores.is_empty()
    }
    pub fn new_test(config: &Config) -> Self {
        fn rand_spaceship(config: &Config) -> SpaceshipStats {
            config.spaceships_stats(rand::thread_rng().gen_range(0..config.spaceships_count()))
        }
        Self {
            scores: vec![
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 12,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 11,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 10,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 9,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 8,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 7,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 6,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 5,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 4,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 3,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
                Score {
                    player_name: "hackerman".to_owned(),
                    points: 2,
                    obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
                    spaceship: rand_spaceship(config),
                    shots_fired: 4,
                    shots_hit: 2,
                    shots_missed: 1,
                },
            ],
        }
    }
    pub fn add(&mut self, score: Score) -> usize {
        for i in 0..self.scores.len() {
            if score.points > self.scores[i].points {
                // remove 100. score if it exists
                if self.scores.len() >= 100 {
                    self.scores.pop();
                }
                self.scores.insert(i, score);
                return i;
            }
        }
        let i = self.scores.len();
        self.scores.push(score);
        i
    }
    pub fn scores(&self) -> &[Score] {
        &self.scores
    }
    pub fn score_unchecked(&self, i: usize) -> &Score {
        &self.scores[i]
    }
    pub fn score_mut_unchecked(&mut self, i: usize) -> &mut Score {
        &mut self.scores[i]
    }
    pub fn remove_score_unchecked(&mut self, i: usize) {
        self.scores.remove(i);
    }

    pub fn try_load(config: &Config, popups: &mut Popups) -> Self {
        if !config.path.as_os_str().is_empty() {
            let mut file_name = config.path.file_name().unwrap_or(OsStr::new("")).to_owned();
            file_name.push(OsStr::new(".highscores.toml"));
            let path = config.path.with_file_name(file_name);
            if path.try_exists().is_ok_and(|v| !v) {
                // file does not exist, use empty highscores
                Self::new_empty()
            } else {
                // file does exist, try to load it
                match std::fs::read_to_string(&path) {
                    Ok(as_str) => match toml::from_str(&as_str) {
                        Ok(v) => v,
                        Err(e) => {
                            popups.show_error(Popup::new(
                            "failed to parse highscores",
                            format!("the highscores file {path:?} contains an error.\nIf you\n1. do not fix it and\n2. set a highscore in this session,\nyour previous highscores will be lost!\n\nError:\n{e}"),
                        ));
                            Self::new_empty()
                        }
                    },
                    Err(e) => {
                        popups.show_error(Popup::new(
                            "failed to load highscores",
                            format!("the highscores file {path:?} could not be read.\nIf you\n1. do not fix it and\n2. set a highscore in this session,\nyour previous highscores will be lost!\n\nError:\n{e}"),
                        ));
                        Self::new_empty()
                    }
                }
            }
        } else {
            Self::new_test(config)
        }
    }

    pub fn save(&mut self, config: &Config, popups: &mut Popups) {
        self.scores.truncate(10000);
        if !config.path.as_os_str().is_empty() && !self.is_empty() {
            let mut file_name = config.path.file_name().unwrap_or(OsStr::new("")).to_owned();
            file_name.push(OsStr::new(".highscores.toml"));
            let path = config.path.with_file_name(file_name);
            match std::fs::write(&path, toml::to_string(self).unwrap()) {
                Ok(()) => (),
                Err(e) => popups.show_error(Popup::new(
                    "failed to save highscores",
                    format!("the highscores could not be saved to {path:?}:\n{e}"),
                )),
            }
        }
    }
}

/// A score achieved by a player.
/// Contains the player's name, how many points they got,
/// and information about the game they played so the score
/// can later be compared to scores achieved under similar conditions.
#[derive(Clone, Serialize, Deserialize, PartialEq)]
pub struct Score {
    pub player_name: String,
    pub points: usize,

    pub obstacle_spawn_mode: ObstacleSpawnMode,
    pub spaceship: SpaceshipStats,

    pub shots_fired: u32,
    pub shots_hit: u32,
    pub shots_missed: u32,
}
impl Score {
    pub fn new(
        points: usize,
        obstacle_spawn_mode: ObstacleSpawnMode,
        spaceship: SpaceshipStats,
        shots_fired: u32,
        shots_hit: u32,
        shots_missed: u32,
    ) -> Self {
        Self {
            player_name: String::new(),
            points,
            obstacle_spawn_mode,
            spaceship,
            shots_fired,
            shots_hit,
            shots_missed,
        }
    }
}

#[test]
fn test_highscores_serializing_and_deserializing() {
    use crate::tests::*;
    let scores = Highscores::new_test(&default_config());
    let scores_saved = toml::to_string(&scores).unwrap();
    eprintln!("{}", scores_saved);
    let scores_loaded: Highscores = toml::from_str(&scores_saved).unwrap();
    assert!(scores == scores_loaded);
    let scores_saved_again = toml::to_string(&scores_loaded).unwrap();
    assert!(scores_saved.trim() == scores_saved_again.trim());
}

#[test]
fn test_highscores_saving_after_settings_new_score() {
    use crate::tests::*;
    let scores_og = Highscores::new_test(&default_config());
    let mut scores = scores_og.clone();
    let _new_score = scores.add(Score {
        player_name: "djas".to_owned(),
        points: 42,
        obstacle_spawn_mode: ObstacleSpawnMode::Waveless,
        spaceship: SpaceshipStats {
            name: "aaa".to_owned(),
            health: 69,
            shot_power: 4.2,
            shot_rate: 0.42,
            speed: 6.9,
            shot_spawn_pos: vec![],
            bounds: crate::collision::CollisionBounds::new(vec![]),
        },
        shots_fired: 14,
        shots_hit: 6,
        shots_missed: 5,
    });
    assert!(
        toml::from_str::<Highscores>(&toml::to_string(&scores).unwrap()).unwrap() != scores_og,
        "saving and loading highscores caused them to be equal even though a change was made"
    );
    assert!(
        toml::from_str::<Highscores>(&toml::to_string(&scores).unwrap()).unwrap() == scores,
        "saving and loading highscores caused them to be not equal"
    );
}
