use std::thread::JoinHandle;

use asciivation_duty_renderer::CheckRes;

use crate::{
    background::Background,
    config::{Config, ControlsKeyMode},
    popups::{Popup, Popups},
    scores::Highscores,
    websocket::WsColor,
};

/// The application's global state, which is shared by all screens
pub struct GlobalState {
    pub config: Config,
    pub background: Background,
    pub popups: Popups,
    pub better_obstacle_spawning: bool,
    pub spaceship_color: WsColor,
    spaceship_color_websocket_thread: Option<JoinHandle<()>>,
    pub spaceship_color_ignore_websocket: bool,
    pub draw_reset: bool,
    pub highscores: Highscores,
}

impl GlobalState {
    pub fn new(res: &CheckRes, config: Config) -> Self {
        let mut popups = Popups::new();
        if !res.supports_keyboard_enhancement
            && !matches!(config.controls_key_mode(), ControlsKeyMode::Toggle)
        {
            popups.show_warning(Popup::new("unsupported terminal", format!("your terminal does not support the keyboard enhancement protocol, which is required for the game to work properly.\nSupported terminals are Alacritty, Kitty, or others that supports the protocol.\n\n{}", match config.controls_key_mode() {
                ControlsKeyMode::Hold => "You can also enable the Toggle mode for 'Controls: key mode' in the Options menu.",
                ControlsKeyMode::ToggleIfUnsupportedTerminal => "Because of this, the Toggle mode from 'Controls: key mode' will automatically be used instead of the Hold mode.\nThis warning is hidden if you switch to the Toggle mode in the Options menu.",
                ControlsKeyMode::Toggle => "",
            })));
        }
        let highscores = Highscores::try_load(&config, &mut popups);
        Self {
            config,
            background: Background::new(
                150.0,
                0.025,
                0.5,
                0.33,
                (res.size.0 as _, res.size.1 as _),
            ),
            popups,
            better_obstacle_spawning: true,
            spaceship_color: WsColor::rand(),
            spaceship_color_websocket_thread: None,
            spaceship_color_ignore_websocket: false,
            draw_reset: false,
            highscores,
        }
    }
    pub fn set_spaceship_color_websocket_thread(&mut self, thread: JoinHandle<()>) {
        self.spaceship_color_websocket_thread = Some(thread);
    }
    pub fn spaceship_color_websocket_thread_running(&mut self) -> bool {
        if self
            .spaceship_color_websocket_thread
            .as_ref()
            .is_some_and(|t| !t.is_finished())
        {
            true
        } else {
            self.spaceship_color_websocket_thread = None;
            false
        }
    }
}
