use std::{
    fmt::Debug,
    time::{Duration, Instant},
};

use asciivation_duty_renderer::{
    deps::crossterm::event::Event,
    part::{DrawMode, Part, PartRunner, Settings},
    render::PX_BLACK,
    CheckRes,
};

use crate::{config::Config, state::GlobalState};

pub struct TestRunner<P: Part<State = GlobalState>> {
    part_runner: PartRunner<P>,
    now: Instant,
}
pub trait TestablePart: Part<State = GlobalState> {
    /// creates a new `TestRunner` with a screen size of `check_res.size`.
    /// use `.resize()` to change the size, `.tick()` to simulate time passing,
    /// `.draw()` to render a frame, or `.event()` to simulate an event.
    fn test(self, check_res: &CheckRes) -> TestRunner<Self> {
        TestRunner::new(self, check_res)
    }
    fn test_from<P2: Part<State = GlobalState>>(self, other: TestRunner<P2>) -> TestRunner<Self> {
        TestRunner::new_from(self, other)
    }
}
impl<T: Part<State = GlobalState>> TestablePart for T {}

pub const DEFAULT_CONFIG_STR: &str = include_str!("../../starfighter_alliance_config_default.json");

pub fn default_config() -> Config {
    Config::from_json_str(DEFAULT_CONFIG_STR).expect("Failed to load default config for test")
}

#[allow(unused)]
impl<P: Part<State = GlobalState>> TestRunner<P> {
    fn new(part: P, check_res: &CheckRes) -> Self {
        Self::new_with_config(part, check_res, default_config())
    }
    fn new_with_config(part: P, check_res: &CheckRes, config: Config) -> Self {
        Self::new_internal(part, check_res, config)
    }
    fn new_from<P2: Part<State = GlobalState>>(part: P, other: TestRunner<P2>) -> Self {
        Self {
            part_runner: other.part_runner.to(part),
            now: other.now,
        }
    }
    fn new_internal(part: P, check_res: &CheckRes, config: Config) -> Self {
        let mut s = Self {
            part_runner: PartRunner {
                part,
                end: None,
                screen_buf: vec![],
                screen_second: None,
                state: GlobalState::new(check_res, config),
                settings: Settings::new(check_res, DrawMode::DiffOnly),
                draw_reset: true,
            },
            now: Instant::now(),
        };
        s.resize_silent(check_res.size);
        s
    }

    pub fn tick(&mut self, elapsed_duration: Duration) -> &mut Self {
        self.now += elapsed_duration;
        self
    }
    // resizes the screen buffer and sends an event
    pub fn resize(&mut self, new_size: (u16, u16)) -> Result<&mut Self, P::End> {
        self.resize_silent(new_size);
        self.event(Event::Resize(new_size.0, new_size.1))
    }
    /// like resize, but does not send an event
    pub fn resize_silent(&mut self, new_size: (u16, u16)) -> &mut Self {
        self.part_runner.screen_buf = (0..new_size.1)
            .map(|_| vec![PX_BLACK; new_size.0 as _])
            .collect();
        self.part_runner.screen_second = None;
        self
    }
    pub fn draw(&mut self) -> Result<&mut Self, P::End> {
        Part::draw(&mut self.part_runner, self.now);
        self.err_if_end()
    }
    pub fn event(&mut self, e: Event) -> Result<&mut Self, P::End> {
        Part::event(&mut self.part_runner, e);
        self.err_if_end()
    }

    /// can be used to `assert!` on values contained in the part runner
    pub fn check(&mut self, func: impl FnOnce(&mut PartRunner<P>)) -> &mut Self {
        func(&mut self.part_runner);
        self
    }

    /// generates a check_res based on the current screen size and values in `part_runner.settings`
    pub fn check_res(&self) -> CheckRes {
        CheckRes {
            size: (
                self.part_runner
                    .screen_buf
                    .first()
                    .map(|row| row.len())
                    .unwrap_or(0) as _,
                self.part_runner.screen_buf.len() as _,
            ),
            supports_keyboard_enhancement: self.part_runner.settings.with_keyboard_enhancement,
        }
    }
    /// drops the runner and returns the contained part
    pub fn part(self) -> P {
        self.part_runner.part
    }
    pub fn part_ref(&self) -> &P {
        &self.part_runner.part
    }
    pub fn part_mut(&mut self) -> &mut P {
        &mut self.part_runner.part
    }

    fn err_if_end(&mut self) -> Result<&mut Self, P::End> {
        if let Some(end) = self.part_runner.end.take() {
            Err(end)
        } else {
            Ok(self)
        }
    }
}

impl<P: Part<State = GlobalState>> Debug for TestRunner<P> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "TestRunner[end={}]",
            if self.part_runner.end.is_some() {
                "Some"
            } else {
                "None"
            }
        )
    }
}

#[test]
fn test_load_config() {
    // config is loaded by this function, uses starfighter_alliance_config_default.json
    default_config();
}

#[test]
fn test_load_invalid_config() {
    if Config::from_json_str(&DEFAULT_CONFIG_STR.replace(r#""name""#, r#""nmae""#)).is_ok() {
        panic!("name changed to nmae -> should error");
    }
}
