#[cfg(feature = "websocket")]
use std::error::Error;
use std::fmt::Debug;
use std::fmt::Display;

use rand::seq::SliceRandom;
#[cfg(feature = "websocket")]
use serde::Deserialize;
#[cfg(feature = "websocket")]
use websocket::{
    url::ParseError, ClientBuilder as Websocket, Message, OwnedMessage, WebSocketError,
};

use crate::{
    popups::{self, Popup},
    state::GlobalState,
};

pub fn start_thread(address: String, global_state: &mut GlobalState) {
    #[cfg(not(feature = "websocket"))]
        global_state.popups.show_warning(Popup::new("no websocket support", "The config specifies a websocket address, but the websocket feature was disabled at build-time. You can ignore this warning, remove the websocket address from your config, or build the application with the websocket feature enabled."));
    #[cfg(feature = "websocket")]
    if !global_state.spaceship_color_websocket_thread_running() {
        let popups = global_state.popups.popup_sender();
        global_state.spaceship_color_ignore_websocket = false;
        global_state.set_spaceship_color_websocket_thread(std::thread::spawn(move || {
            match get_websocket_color(&address) {
                Ok(color) => popups.message(popups::PMessage::SetSpaceshipColor(color)),
                Err(e) => popups.show_warning(Popup::new(
                    "websocket error",
                    format!("Websocket at {address} caused an error:\n{e}"),
                )),
            }
        }));
    }
}

#[cfg(feature = "websocket")]
pub fn get_websocket_color(address: &str) -> Result<WsColor, GetWebsocketColorError> {
    let mut client = Websocket::new(address)
        .map_err(|e| GetWebsocketColorError::BadAddress(address.to_owned(), e))?
        .connect(None)?;
    client.send_message(&Message::text(
        r#"{
    "messageType": "COLOR_REQUEST"
}"#,
    ))?;
    let response = client.recv_message()?;
    Ok(match response {
        OwnedMessage::Text(response) => match serde_json::from_str::<ExpectedResponse>(&response) {
            Ok(res) => {
                if res.code != 200 {
                    return Err(GetWebsocketColorError::ExpectedResponseWithUnexpectedCode(
                        res.code,
                    ));
                }
                let color = res.color.to_lowercase();
                match color.trim() {
                    "red" => WsColor::Red,
                    "blue" => WsColor::Blue,
                    "yellow" => WsColor::Yellow,
                    "orange" => WsColor::Orange,
                    "green" => WsColor::Green,
                    "violet" => WsColor::Violet,
                    "white" => WsColor::White,
                    "black" => WsColor::Black,
                    "transparent" => WsColor::Transparent,
                    _ => return Err(GetWebsocketColorError::UnknownColor(res.color)),
                }
            }
            Err(_) => match serde_json::from_str::<GenericResponse>(&response) {
                Ok(res) => {
                    return Err(GetWebsocketColorError::UnexpectedResponseWithCode(res.code))
                }
                Err(_) => return Err(GetWebsocketColorError::UnexpectedResponse(response)),
            },
        },
        other => return Err(GetWebsocketColorError::NonTextResponse(other)),
    })
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum WsColor {
    Red,
    Blue,
    Yellow,
    Orange,
    Green,
    Violet,
    White,
    Black,
    Transparent,
}
impl WsColor {
    pub fn rand() -> Self {
        *[
            Self::Red,
            Self::Blue,
            Self::Yellow,
            Self::Orange,
            Self::Green,
            Self::Violet,
            Self::White,
            Self::Black,
            Self::Transparent,
        ]
        .choose(&mut rand::thread_rng())
        .unwrap()
    }
}
impl Display for WsColor {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::Red => "Red",
                Self::Blue => "Blue",
                Self::Yellow => "Yellow",
                Self::Orange => "Orange",
                Self::Green => "Green",
                Self::Violet => "Violet",
                Self::White => "White",
                Self::Black => "Black",
                Self::Transparent => "Transparent",
            }
        )
    }
}

#[cfg(feature = "websocket")]
#[derive(Deserialize)]
pub struct ExpectedResponse {
    code: i32,
    color: String,
}
#[cfg(feature = "websocket")]
#[derive(Deserialize)]
pub struct GenericResponse {
    code: i32,
}

#[cfg(feature = "websocket")]
pub enum GetWebsocketColorError {
    BadAddress(String, ParseError),
    WebsocketError(WebSocketError),
    NonTextResponse(OwnedMessage),
    ExpectedResponseWithUnexpectedCode(i32),
    UnexpectedResponseWithCode(i32),
    UnexpectedResponse(String),
    UnknownColor(String),
}

#[cfg(feature = "websocket")]
impl From<WebSocketError> for GetWebsocketColorError {
    fn from(value: WebSocketError) -> Self {
        Self::WebsocketError(value)
    }
}

#[cfg(feature = "websocket")]
impl Error for GetWebsocketColorError {}
#[cfg(feature = "websocket")]
impl Display for GetWebsocketColorError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::BadAddress(addr, e) => write!(f, "invalid websocket address {addr}: {e}"),
            Self::WebsocketError(e) => write!(f, "{e}"),
            Self::NonTextResponse(res) => {
                write!(
                    f,
                    "got non-text response (type: {})",
                    match res {
                        OwnedMessage::Text(_) => "text",
                        OwnedMessage::Binary(_) => "binary",
                        OwnedMessage::Close(_) => "close",
                        OwnedMessage::Ping(_) => "ping",
                        OwnedMessage::Pong(_) => "pong",
                    }
                )
            }
            Self::ExpectedResponseWithUnexpectedCode(code)
            | Self::UnexpectedResponseWithCode(code) => {
                write!(f, "got code {code} from server, which is not 200")
            }
            Self::UnexpectedResponse(res) => write!(
                f,
                "got unexpected response from server: {}",
                res.replace(['\n', '\r', '\t'], " ")
            ),
            Self::UnknownColor(color) => write!(
                f,
                "got unknown color from server: {}",
                color.replace(['\n', '\r', '\t'], " ")
            ),
        }
    }
}
#[cfg(feature = "websocket")]
impl Debug for GetWebsocketColorError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{self}")
    }
}
